/***************************************************************************

    M.A.M.E.32  -  Multiple Arcade Machine Emulator for Win32
    Win32 Portions Copyright (C) 1997-98 Michael Soderstrom and Chris Kirmse

    This file is part of MAME32, and may only be used, modified and
    distributed under the terms of the MAME license, in "readme.txt".
    By continuing to use, modify or distribute this file you indicate
    that you have read the license and understand and accept it fully.

 ***************************************************************************/

/***************************************************************************
    directinput.c

    Direct Input routines.

 ***************************************************************************/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#define DIRECTINPUT_VERSION 0x0700
#include <dinput.h>
#include "DirectInput.h"
#include "dxdecode.h"
#include "M32Util.h"
#include "win32ui.h"
#include "win32text.h"

/***************************************************************************
    function prototypes
 ***************************************************************************/

/***************************************************************************
    External variables
 ***************************************************************************/

LPDIRECTINPUT7 di = NULL;

/***************************************************************************
    Internal structures
 ***************************************************************************/

/***************************************************************************
    Internal variables
 ***************************************************************************/

static HANDLE hDLL = NULL;

/***************************************************************************
    External functions
 ***************************************************************************/

/****************************************************************************
 *      DirectInputInitialize
 *
 *      Initialize the DirectInput variables.
 *
 *      This entails the following functions:
 *
 *          DirectInputCreate
 *
 ****************************************************************************/

typedef HRESULT (WINAPI *dicex_proc)(HINSTANCE hinst,
                                     DWORD dwVersion,
                                     REFIID riidltf,
                                     LPVOID* ppvOut,
                                     LPUNKNOWN punkOuter);

BOOL DirectInputInitialize(void)
{
    HRESULT hResult;
    UINT error_mode;
    dicex_proc dicex;

    if (hDLL != NULL)
        return TRUE;

    hDLL = NULL;

    /* Turn off error dialog for this call */
    error_mode = SetErrorMode(0);
    hDLL = LoadLibrary("dinput.dll");
    SetErrorMode(error_mode);

    if (hDLL == NULL)
    {
        ErrorMsg(win32_getstring(WIN32_dinput_dllnotfound));
        return FALSE;
    }

    dicex = (dicex_proc)GetProcAddress(hDLL, "DirectInputCreateEx");
    if (dicex == NULL)
    {
        ErrorMsg(win32_getstring(WIN32_dinput_dicexnotfound));
        return FALSE;
    }

    hResult = dicex((HINSTANCE)GetWindowLong(GetMainWindow(), GWL_HINSTANCE),
                    DIRECTINPUT_VERSION,
                    &IID_IDirectInput7,
                    (void **)&di,
                    NULL);
    if (FAILED(hResult))
    {
        ErrorMsg("DirectInputCreateEx %s%s",
				 win32_getstring(WIN32_errorcode),
                 DirectXDecodeError(hResult));
        di = NULL;
        return FALSE;
    }

    return TRUE;
}

/****************************************************************************
 *
 *      DirectInputClose
 *
 *      Terminate our usage of DirectInput.
 *
 ****************************************************************************/

void DirectInputClose(void)
{
    /*
        Destroy any lingering IDirectInput object.
    */
    if (di)
    {
        IDirectInput7_Release(di);
        di = NULL;
    }
}


BOOL CALLBACK inputEnumDeviceProc(LPDIDEVICEINSTANCE pdidi, LPVOID pv)
{
    GUID *pguidDevice;

    /* report back the instance guid of the device we enumerated */
    if (pv)
    {
        pguidDevice  = (GUID *)pv;
        *pguidDevice = pdidi->guidInstance;
    }

    /* BUGBUG for now, stop after the first device has been found */
    return DIENUM_STOP;
}


HRESULT SetDIDwordProperty(LPDIRECTINPUTDEVICE7 pdev, REFGUID guidProperty,
               DWORD dwObject, DWORD dwHow, DWORD dwValue)
{
    DIPROPDWORD dipdw;

    dipdw.diph.dwSize       = sizeof(dipdw);
    dipdw.diph.dwHeaderSize = sizeof(dipdw.diph);
    dipdw.diph.dwObj        = dwObject;
    dipdw.diph.dwHow        = dwHow;
    dipdw.dwData            = dwValue;

    return IDirectInputDevice7_SetProperty(pdev, guidProperty, &dipdw.diph);
}


/***************************************************************************
    Internal functions
 ***************************************************************************/
