/***************************************************************************

  M.A.M.E.32  -  Multiple Arcade Machine Emulator for Win32
  Win32 Portions Copyright (C) 1997-2001 Michael Soderstrom and Chris Kirmse

  This file is part of MAME32, and may only be used, modified and
  distributed under the terms of the MAME license, in "readme.txt".
  By continuing to use, modify or distribute this file you indicate
  that you have read the license and understand and accept it fully.

 ***************************************************************************/

/***************************************************************************

  KailleraChat.c

  Fullscreen chat for Kaillera.

  Chat Display Functions Created by japmame.

  Japanese IMM Functions Created 15/5/2001  by NJ.

  Fixed Form & Chat Position Move Functions Created 4/6/2001  by RYO.

***************************************************************************/

#include <driver.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <stdio.h>
#include "mame32.h"
#include "resource.h"
#include "usrintrf.h"
#include "kailleraclient.h"
#ifdef __GNUC__
#include "m32imm.h"
#else
#include "imm.h"
#endif

/***************************************************************************
    Macros and definitions
 ***************************************************************************/

#define KC_BUFFER_MAX    128
#define KC_SCROLL_SPD    100
#define KC_BLINK_SPD     20

#define KC_CHATLINE_MAX  8
#define KC_EDITLINE      1
#define KC_CONVSTR       1
#define KC_CURSOR        1
#define KC_IME           1
#define KC_TERMINATE     1

#define DT_MAX           (KC_CHATLINE_MAX + KC_EDITLINE + KC_CONVSTR + KC_CURSOR + KC_IME + KC_TERMINATE)

#define IME_CLOSE         0
#define IME_OPEN          1
#define IME_CONVERSION    2

#define MAX_FIXEDFORM    10

#define KAILLERA_TITLE_LEN        128
#define KAILLERA_CHAT_TITLE       "Kaillera v"

/* Control ID (KailleraClient.dll) */
#define IDC_KAILLERA_CHAT_EDIT     1005
#define IDC_KAILLERA_CHAT_LOG      1006
#define IDC_KAILLERA_PALY_BUTTON   1013
#define IDC_KAILLERA_CANCEL_BUTTON 2

#define ISSJIS(c) (((c) >= 0x81 && (c) <= 0x9f) || ((c) >= 0xe0 && (c) <= 0xfc))

typedef BOOL (WINAPI *ime_proc)(HWND hWnd, BOOL bflag);

/***************************************************************************
    Function prototypes
 ***************************************************************************/

static void             fixedform_open(void);
static LRESULT CALLBACK EditProc(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam);
static void             KailleraChatCloseChat(void);
static void             KailleraChatImeGetConversionStatus(void);
static BOOL             MSIME2K_GetReg(void);
static void             GetWinVersion(BOOL *bIsNT, int *nMajor, int *nMinor, int *nBuildNumber);

static HWND             GetKailleraClientDialog(void);
static HWND             GetKailleraClientControl(HWND hDlg, int nID);

/***************************************************************************
    External variables
 ***************************************************************************/

int nChatDrawMode;

/***************************************************************************
    Internal variables
 ***************************************************************************/

static BOOL            bChatActive;
static BOOL            bUseIME;
static BOOL            bUseFF;
static BOOL            bInputEnable;
static BOOL            bShowLog;
static int             nIMEState;
static int             nPrevIMEState;
static int             nScroll;
static int             nIndex;

static int             nEditMaxLen;
static int             nCursorPos;
static int             nBlinkCounter;
static BOOL            bShowCursor;
static int	           bChatPosition;

static HANDLE          hUser32DLL;

static unsigned char   szRecvBuf[KC_BUFFER_MAX * 4];
static unsigned char   szChatText[KC_CHATLINE_MAX][KC_BUFFER_MAX];

static unsigned char   szInputBuf[KC_BUFFER_MAX];
static unsigned char   szConvStrBuf[KC_BUFFER_MAX];
static int             pConvClauseBuf[KC_BUFFER_MAX];

static BOOL            bCloseCompositionWindow;
static unsigned char   szIMEName[64];
static unsigned char   szIMEStateStr[16];

static ime_proc        _WINNLSEnableIME;
static WNDPROC         DefaultWindowProc;

static int             loaded_fixedform;
static char            fixedform_desc[MAX_FIXEDFORM][48];
static int             fixedform_flag;

static HWND            hKailleraDialog;
static HWND            hKailleraPlayButton;
static HWND            hKailleraCancelButton;

char *fixedform_filename = "fixedfrm.dat";

/***************************************************************************
    External functions
 ***************************************************************************/

void KailleraChatInit(void)
{
    int  i;
    HIMC hImc;
    HKL  hKeyboardLayout;
    UINT error_mode;


    bChatActive   = FALSE;
    bInputEnable  = FALSE;
    nPrevIMEState = IME_CLOSE;
    nIMEState     = IME_CLOSE;
    nIndex        = 0;
    nScroll       = 0;
    bUseIME       = GetUseImeInChat();
    bUseFF        = GetUseFixedForm();
	bChatPosition = 0;

    for (i = 0; i < KC_CHATLINE_MAX; i++)
        ZeroMemory(szChatText[i], KC_BUFFER_MAX);

    ZeroMemory(szRecvBuf, KC_BUFFER_MAX * 4);
    ZeroMemory(szInputBuf, KC_BUFFER_MAX);
    ZeroMemory(szConvStrBuf, KC_BUFFER_MAX);

    hImc = ImmGetContext(MAME32App.m_hWnd);

    if (!ImmGetOpenStatus(hImc))
        ImmSetOpenStatus(hImc, TRUE);

    hKeyboardLayout = GetKeyboardLayout(0);

	if (bUseFF)
		fixedform_open();

    if (ImmIsIME(hKeyboardLayout))
    {
        BOOL bIsNT;
        int  nMajor, nMinor, nBuildNumber;

        ImmGetDescription(hKeyboardLayout, szIMEName, 64);

        GetWinVersion(&bIsNT, &nMajor, &nMinor, &nBuildNumber);

#if FORCE_CLOSE_COMPOSITION_WINDOW
        bCloseCompositionWindow = TRUE;
#else
        if (bIsNT == FALSE && nMajor >= 5)
        {
            /* Windows 2000̏ꍇ */
            if (!strcmp(szIMEName, "Microsoft IME 2000 (Japanese)"))
            {
                /*
                   MS-IME 2000̏ꍇ́ARgEChEȂ
                   ʂɉ\Ȃ
                */

                if (MSIME2K_GetReg())
                    bCloseCompositionWindow = FALSE;
                else
                    bCloseCompositionWindow = TRUE;
            }
            else
            if (!strcmp(szIMEName, "MSIME95 4.0")
            ||  !strcmp(szIMEName, "MS-IME97 {ͼ")
            ||  !strcmp(szIMEName, "Microsoft IME 98 {̓VXe"))
            {
                /* MS-IME 95/97/98̏ꍇ́AȂĂOK */
                bCloseCompositionWindow = FALSE;
            }
            else
            {
                /*
                 ̑̓{IMȄꍇ͋IɕB
                 IIME̕ϊ̕ȂȂB
                */
                bCloseCompositionWindow = TRUE;
            }
        }
        else
        {
            /*
              Win9xłIME̎ނɂ炸ϊ̕
              \Ă܂߁AIɕB
              IIME̕ϊ̕ȂȂ邪A
              KailleraClient̃OENbNׂđI
              łȂIMEA(NMAME X̌f)B

              NT4͕s(ۏ؊O炢Ǝv)B
            */
            bCloseCompositionWindow = TRUE;
        }
#endif

        /*
          WINNLSEnableIMÉAĂƎgpłȂ̂
          (user32.dll̂ɑ݂Ȃ͗l)A݂邩
          ǂ`FbNB
          {ImmDisableIME()gpƂ낾
          Win95ł͎gpłȂB
        */
        error_mode = SetErrorMode(0);
        hUser32DLL = LoadLibrary("user32.dll");
        SetErrorMode(error_mode);

        if (hUser32DLL != NULL)
            _WINNLSEnableIME = (ime_proc)GetProcAddress(hUser32DLL, "WINNLSEnableIME");

        if (_WINNLSEnableIME == NULL)
        {
            FreeLibrary(hUser32DLL);
            hUser32DLL = NULL;
        }
    }
    else
    {
        bUseIME = FALSE;
        hUser32DLL = NULL;
        _WINNLSEnableIME = NULL;
        bCloseCompositionWindow = FALSE;
    }

    if (ImmGetOpenStatus(hImc))
        ImmSetOpenStatus(hImc, FALSE);

    ImmReleaseContext(MAME32App.m_hWnd, hImc);

    if (bUseIME)
    {
        KailleraChatImeGetConversionStatus();
        nEditMaxLen   = (Machine->uiwidth - (5 + strlen(szIMEStateStr)) * Machine->uifontwidth) / Machine->uifontwidth - 1;
    }
    else
    {
        nEditMaxLen   = (Machine->uiwidth - 5 * Machine->uifontwidth) / Machine->uifontwidth - 1;
    }

    if (_WINNLSEnableIME)
        _WINNLSEnableIME(MAME32App.m_hWnd, FALSE);

    DefaultWindowProc = (WNDPROC)GetWindowLong(MAME32App.m_hWnd, GWL_WNDPROC);
    SetWindowLong(MAME32App.m_hWnd, GWL_WNDPROC, (LONG)EditProc);
}


void KailleraChatExit(void)
{
	SetWindowLong(MAME32App.m_hWnd, GWL_WNDPROC, (LONG)DefaultWindowProc);

    if (_WINNLSEnableIME)
        _WINNLSEnableIME(MAME32App.m_hWnd, TRUE);

    if (hUser32DLL)
    {
        FreeLibrary(hUser32DLL);
        hUser32DLL = NULL;
    }
}

void KailleraChatUpdate(struct osd_bitmap *pBitmap)
{
    struct DisplayText dtx[DT_MAX];
    unsigned char szChatLine[KC_BUFFER_MAX + 6];
    unsigned char szCursor[KC_BUFFER_MAX];
    int i, j, k;

    if (bCloseCompositionWindow)
    {
	    if (bUseIME && bChatActive)
	    {
	        /* IMẼftHg̃EChEL킳 */
	        HWND hIME = ImmGetDefaultIMEWnd(MAME32App.m_hWnd);
	        if (hIME != NULL)
	            SendMessage(hIME, WM_CLOSE, 0, 0);
	    }
	}

    /* keyboard input */
    if (!bChatActive)
    {
        if (input_ui_pressed(IPT_UI_KAILLERA_CHAT))
        {
            if (bUseIME)
            {
                if (_WINNLSEnableIME)
                    _WINNLSEnableIME(MAME32App.m_hWnd, TRUE);

                if (nPrevIMEState == IME_OPEN)
                {
                    HIMC hImc = ImmGetContext(MAME32App.m_hWnd);
                    ImmSetOpenStatus(hImc, TRUE);
                    ImmReleaseContext(MAME32App.m_hWnd, hImc);
                    nIMEState = IME_OPEN;
                }
                else
                    nIMEState = IME_CLOSE;
            }

            ZeroMemory(szInputBuf, KC_BUFFER_MAX);
            ZeroMemory(szConvStrBuf, KC_BUFFER_MAX);

            nCursorPos   = 0;
            bInputEnable = TRUE;
            bChatActive  = TRUE;
        }
    }

	if (input_ui_pressed(IPT_UI_CHAT_MOVE))
	{
		bChatPosition++;
		if (bChatPosition > 3)	bChatPosition = 0;
	}

    /* copy szRecvBuf to szChatText if line available */
    if (szRecvBuf[0])
    {
        strcpy(szChatText[nIndex], szRecvBuf);
        nIndex++;
        nIndex &= KC_CHATLINE_MAX - 1;
        szChatText[nIndex][0] = '\0';
        szRecvBuf[0] = '\0';
        nScroll = 0;
    }

    /* auto scroll out */
    nScroll++;
    if (nScroll > KC_SCROLL_SPD)
    {
        nScroll = 0;
        nIndex++;
        nIndex &= KC_CHATLINE_MAX - 1;
        szChatText[nIndex][0] = '\0';
    }

    /* display chat text */
    j = 0;
	k = 0;

	for (i = 0; i < KC_CHATLINE_MAX; i++)
	{
        if (szChatText[(nIndex + i) % KC_CHATLINE_MAX][0])
	        j++;
	}

    for (i = 0; i < KC_CHATLINE_MAX; i++)
    {
        dtx[i].text  = szChatText[(nIndex + i) % KC_CHATLINE_MAX];
        dtx[i].color = UI_COLOR_NORMAL;

		if (bChatPosition & 0x02)
		{
			dtx[i].x = Machine->uiwidth - strlen(szChatText[(nIndex + i) % KC_CHATLINE_MAX]) * Machine->uifontwidth;
		}
		else
		{
			dtx[i].x = 0;
		}

		if (bChatPosition & 0x01)
		{
	        if (szChatText[(nIndex + i) % KC_CHATLINE_MAX][0])
			{
				dtx[i].y = Machine->uiheight - (j - k + 1) * Machine->uifontheight;
				k++;
			}
			else
				dtx[i].y = 0;
		}
		else
		{
	        if (szChatText[(nIndex + i) % KC_CHATLINE_MAX][0])
			{
				dtx[i].y = k * Machine->uifontheight;
				k++;
			}
			else
				dtx[i].y = 0;
		}
    }

    /* display input buffer */
    if (bChatActive)
    {
        int x, y, width, height;
        int nCursorDispPos;
        int nCursorColor;

        x      = Machine->uixmin;
        y      = Machine->uiymin + (Machine->uiheight - Machine->uifontheight);
        width  = Machine->uiwidth;
        height = Machine->uifontheight;
        plot_box(pBitmap, x, y, width, height, Machine->uifont->colortable[0], 0);

        if (strlen(szInputBuf) > nEditMaxLen)
        {
            int nLimit;

            bInputEnable = FALSE;
            if (ISSJIS(szInputBuf[nEditMaxLen - 1]))
                nLimit = nEditMaxLen - 1;
            else
                nLimit = nEditMaxLen;

            szInputBuf[nLimit] = '\0';
            if (nCursorPos > nLimit)
                nCursorPos = nLimit;
        }

        sprintf(szChatLine, "Chat:%s", szInputBuf);

        dtx[i].text  = szChatLine;
        dtx[i].color = UI_COLOR_NORMAL;
        dtx[i].x     = 0;
        dtx[i++].y   = (Machine->uiheight - Machine->uifontheight);

        /* cursor blink */
        nBlinkCounter++;
        if (nBlinkCounter > KC_BLINK_SPD)
        {
            nBlinkCounter = 0;
            if (bShowCursor)
                bShowCursor = FALSE;
            else
                bShowCursor = TRUE;
        }

        ZeroMemory(szCursor, KC_BUFFER_MAX);
        nCursorDispPos = 5 + nCursorPos;
        nCursorColor   = bShowCursor ? UI_COLOR_INVERSE : UI_COLOR_NORMAL;
        if (szInputBuf[nCursorPos])
        {
            unsigned char *p = &szInputBuf[nCursorPos];

            szCursor[0] = *p;
            if (ISSJIS(*p))
                szCursor[1] = *(p + 1);
        }
        else
        {
            szCursor[0] = ' ';
        }

        if (bUseIME)
        {
            if (szIMEStateStr[0])
            {
                dtx[i].text  = szIMEStateStr;
                dtx[i].color = UI_COLOR_NORMAL;
                dtx[i].x     = Machine->uiwidth  - strlen(szIMEStateStr) * Machine->uifontwidth;
                dtx[i++].y   = Machine->uiheight - Machine->uifontheight;
            }
		}

	    dtx[i].text  = szCursor;
	    dtx[i].color = nCursorColor;
	    dtx[i].x     = nCursorDispPos * Machine->uifontwidth;
	    dtx[i++].y   = (Machine->uiheight - Machine->uifontheight);

		if (bShowCursor)
			plot_box(pBitmap, nCursorDispPos * Machine->uifontwidth + 7, Machine->uiheight + 4, 1, Machine->uifontheight - 1, CURSOR_COLOR, 0);
	}

    dtx[i].text = NULL; /* terminate array */

    /* display */
	displaytext(pBitmap, dtx, FONT_NORMAL);

	if (bChatActive)
	{
        int nCursorDispPos;
        int nCursorColor;
		i = 0;
        nCursorDispPos = 5 + nCursorPos;
        nCursorColor   = bShowCursor ? UI_COLOR_INVERSE : UI_COLOR_NORMAL;

		if (bUseIME)
        {
            if (nIMEState & IME_CONVERSION)
            {
                if (szConvStrBuf[0] == '\0')
                {
                    nIMEState &= ~IME_CONVERSION;
                }
                else
                {
                    int nStartPos;
                    int nMaxLen;
                    int nStrLen;

                    nStartPos = nCursorDispPos;
                    nStrLen   = strlen(szConvStrBuf);
                    nMaxLen   = 5 + nEditMaxLen;

                    while (nStartPos + nStrLen > nMaxLen)
                        nStartPos--;

                    dtx[i].text  = szConvStrBuf;
                    dtx[i].color = UI_COLOR_INVERSE;
                    dtx[i].x     = nStartPos * Machine->uifontwidth;
                    dtx[i++].y   = Machine->uiheight - Machine->uifontheight;
					plot_box(pBitmap, nStartPos * Machine->uifontwidth + 7, Machine->uiheight + 4, nStrLen * Machine->uifontwidth, 12/*Machine->uifontheight - 1*/, Machine->uifont->colortable[0], 0);

                    nCursorDispPos = nStartPos + nStrLen;
                    nCursorColor   = UI_COLOR_NORMAL;
                }
            }
        }
    }

    dtx[i].text = NULL; /* terminate array */

    /* display */
	displaytext(pBitmap, dtx, FONT_NORMAL);
}


void KailleraChateReceive(char *szText)
{
    if (szRecvBuf[0])
        strcat(szRecvBuf,"\n");
    strcat(szRecvBuf, szText);
}


int KailleraChatIsActive(void)
{
    return bChatActive;
}


/***************************************************************************
    Internal functions
 ***************************************************************************/

static void fixedform_open(void)
{
	FILE *fp;
	char name[256];
	char tmpbuf[256];
	int i;

	sprintf(name, fixedform_filename);
	fp = fopen(name, "r");
	if (fp == NULL)
		return;

	loaded_fixedform = 0;

	/* Ex.: ͂߂܂āI (ő12(F12)܂) */
	while (1)
	{
		if (fgets(tmpbuf, 255, fp) == NULL) break;

		if (tmpbuf[0] == ';' || tmpbuf[0] == '\n') continue;

		i = 0;

		while (tmpbuf[i] != '\n')
		{
			fixedform_desc[loaded_fixedform][i] = tmpbuf[i];
			i++;
		}

		fixedform_desc[loaded_fixedform][++i] = '\0';

		loaded_fixedform++;
		if (loaded_fixedform == MAX_FIXEDFORM)
			break;
	}

	fclose(fp);

	fixedform_flag = 0;
}


static LRESULT CALLBACK EditProc(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
    /* IMẼftHgUI𖳌ɂ */
    /* EEE͂Ȃ񂾂ǁASɖɂȂ킯ł͂Ȃ͗l */
    /* Ƃ肠A悤ɂĂ                     */
    if (Msg == WM_IME_SETCONTEXT)
    {
        lParam &= ~ISC_SHOWUIALL;
        return CallWindowProc(DefaultWindowProc, hWnd, Msg, wParam, lParam);
    }

    /* `bgANeBû݁AMAME32̃EChEtbN */
    if (bChatActive)
    {
        HIMC  hImc;

        switch (Msg)
        {
        case WM_CHAR:
            switch (wParam)
            {
            case VK_RETURN:
            case VK_ESCAPE:
                return TRUE; /* disable beep */

            default:
                if (bChatActive && bInputEnable)
                {
                    int i, nLen;
                    unsigned char *p = &szInputBuf[nCursorPos];

                    nLen = strlen(szInputBuf);

					if (wParam >= 0x30 && wParam <= 0x39)
					{
						if (fixedform_flag)
						{
							fixedform_flag = 0;
							return TRUE;
						}
					}
                    
					if ((wParam & 0xff) > 0x1f)
                    {
                        if (strlen(szInputBuf) < KC_BUFFER_MAX - 1)
                        {
                            for (i = nLen; i >= nCursorPos; i--)
                                szInputBuf[i + 1] = szInputBuf[i];

                            *p = (unsigned char)wParam;
                            nCursorPos++;
                        }
                    }
                }
            }
            break;

        case WM_KEYDOWN:
            if (!(nIMEState & IME_CONVERSION))
            {
                unsigned char *p = &szInputBuf[nCursorPos];

                switch (wParam)
                {
                case VK_RETURN:
                    if (szInputBuf[0] == '\0')
                        return TRUE;
                    kailleraChatSend(szInputBuf);
                    KailleraChatCloseChat();
                    return TRUE;

                case VK_ESCAPE:
                    KailleraChatCloseChat();
                    input_ui_pressed(IPT_UI_CANCEL); /* IPT_UI_CANCEL̉tONA */
                    return TRUE;

				case VK_NUMPAD0:
				case VK_NUMPAD1:
				case VK_NUMPAD2:
				case VK_NUMPAD3:
				case VK_NUMPAD4:
				case VK_NUMPAD5:
				case VK_NUMPAD6:
				case VK_NUMPAD7:
				case VK_NUMPAD8:
				case VK_NUMPAD9:
					if (bUseFF)
					{
						int n = wParam - VK_NUMPAD0;

						if (n > loaded_fixedform - 1)
							break;
						else
						{
							fixedform_flag = 1;
							ZeroMemory(szInputBuf, KC_BUFFER_MAX);
							strcpy(szInputBuf, fixedform_desc[n]);
							nCursorPos = strlen(fixedform_desc[n]);

//							keyboard_pressed_memory(n + KEYCODE_0_PAD);

							return TRUE;
						}
						break;
					}
					break;

                case VK_BACK:
                    if (nCursorPos >= 2)
                    {
                        if (ISSJIS(*(p - 2)))
                        {
                            while (*(p - 1))
                            {
                                *(p - 2) = *p;
                                p++;
                            }
                            nCursorPos -= 2;
                            bInputEnable = TRUE;
                            return TRUE;
                        }
                    }
                    if (nCursorPos != 0)
                    {
                        while (*(p - 1))
                        {
                            *(p - 1) = *p;
                            p++;
                        }
                        nCursorPos--;
                        bInputEnable = TRUE;
                        return TRUE;
                    }
                    break;

                case VK_LEFT:
                    if (nCursorPos >= 2)
                    {
                        if (ISSJIS(*(p - 2)))
                        {
                            nCursorPos -= 2;
                            return TRUE;
                        }
                    }
                    if (nCursorPos != 0)
                    {
                        nCursorPos--;
                        return TRUE;
                    }
                    break;

                case VK_RIGHT:
                    {
                        int len = strlen(szInputBuf);

                        if (nCursorPos < len)
                        {
                            if (ISSJIS(*p))
                                nCursorPos++;
                            nCursorPos++;
                            return TRUE;
                        }
                    }
                    break;

                case VK_DELETE:
                    if (ISSJIS(*p))
                    {
                        while (*p)
                        {
                            *p = *(p + 2);
                            p++;
                        }
                        bInputEnable = TRUE;
                        return TRUE;
                    }
                    while (*p)
                    {
                        *p = *(p + 1);
                        p++;
                    }
                    bInputEnable = TRUE;
                    return TRUE;
                }
                break;
            }
            break;

        case WM_IME_STARTCOMPOSITION:
            /* R|WVEBhEJꂽ */
            if (!bCloseCompositionWindow)
            {
                COMPOSITIONFORM CompForm;

                /* Windows 2000 + MS-IMȄꍇAʊO  */
                /* ړ邱ƂŃR|WbgEChE */
                /* \ɂł */

                hImc = ImmGetContext(hWnd);
                ImmGetCompositionWindow(hImc, &CompForm);
                CompForm.dwStyle = CFS_FORCE_POSITION;
                CompForm.ptCurrentPos.x = 65535;
                CompForm.ptCurrentPos.y = 65535;
                ImmSetCompositionWindow(hImc, &CompForm);
                ImmReleaseContext(hWnd, hImc);
            }
            break;

        case WM_IME_COMPOSITION:
            hImc = ImmGetContext(hWnd);
            nIMEState |= IME_CONVERSION;

            if (lParam & GCS_RESULTSTR)
            {
                /*
                 ϊʂ̎擾
                 ϊʂԂ܂AUOMAME32jłWM_CHARŏEĂ
                 ̂ŁAϊIǂ̔f݂̂ɎgpĂ܂B
                 gp̂ł΁Aȉ̂悤Ȋł傤B

                 unsigned char szTemp[KC_BUFFER_MAX];

                 ZeroMemory(szTemp, KC_BUFFER_MAX);
                 ImmGetCompositionString(hImc, GCS_RESULTSTR, szConvStrBuf, KC_BUFFER_MAX - 1);
                 strcat(szInputBuf, szTemp);
                */

                ZeroMemory(szConvStrBuf, KC_BUFFER_MAX);
            }
            if (lParam & GCS_COMPSTR)
            {
                /*
                 ῗ̎̕擾
                 ῗ̂̂̕Ԃ܂B
                */

                ZeroMemory(szConvStrBuf, KC_BUFFER_MAX);
                ImmGetCompositionString(hImc, GCS_COMPSTR, szConvStrBuf, KC_BUFFER_MAX - 1);
            }

            if (lParam & GCS_COMPCLAUSE)
            {
                int i, j, len;
                int pTemp[KC_BUFFER_MAX];

                /* 
                 ߏ̎擾
                 擾f[^ Ƃ΁A|ǂ|VC|ł ł
                 00000006|0000000a|0000000d|000000012 ̂悤ɁA32bitl
                 ߂̏ÏʒuĂ܂B
                 Q[ł͖Ď擾Kv͂ȂB
                */
                len = strlen(szConvStrBuf);
                ImmGetCompositionString(hImc, GCS_COMPCLAUSE, (unsigned char *)&pTemp, len * sizeof(int));

                for (i = 0, j = 0; i < len; i++)
                    if (pTemp[i])
                        pConvClauseBuf[j++] = pTemp[i];
            }
            ImmReleaseContext(hWnd, hImc);
            break;

        case WM_IME_ENDCOMPOSITION:
            /* R|WVEBhEꂽ */
            ZeroMemory(szConvStrBuf, KC_BUFFER_MAX);
            break;

        case WM_IME_NOTIFY:
            switch (wParam)
            {
            case IMN_SETOPENSTATUS:     /* IME I/It */
                hImc = ImmGetContext(hWnd);
                if (ImmGetOpenStatus(hImc))
                    nIMEState |= IME_OPEN;
                else
                    nIMEState = IME_CLOSE;
                ImmReleaseContext(hWnd, hImc); /* breakȂŁA̓[hXV */

            case IMN_SETCONVERSIONMODE: /* ̓[hύX     */
            case IMN_SETSENTENCEMODE:   /* ߕϊ[hύX */
                 KailleraChatImeGetConversionStatus();
                 break;
            }
            break;
        }
    }

    return CallWindowProc(DefaultWindowProc, hWnd, Msg, wParam, lParam);
}


static void KailleraChatCloseChat(void)
{
    if (bUseIME && !(nIMEState & IME_CONVERSION))
    {
        nPrevIMEState = nIMEState;

        if (nIMEState != IME_CLOSE)
        {
            HIMC hImc = ImmGetContext(MAME32App.m_hWnd);
            ImmSetOpenStatus(hImc, FALSE);
            ImmReleaseContext(MAME32App.m_hWnd, hImc);
        }
        nIMEState = IME_CLOSE;
    }

    if (_WINNLSEnableIME)
        _WINNLSEnableIME(MAME32App.m_hWnd, FALSE);

    ZeroMemory(szInputBuf, KC_BUFFER_MAX);
    ZeroMemory(szConvStrBuf, KC_BUFFER_MAX);
    bInputEnable = FALSE;
    bChatActive  = FALSE;
}


static void KailleraChatImeGetConversionStatus(void)
{
    HIMC  hImc;
    BOOL  bOpen;
    DWORD dwConversion;
    DWORD dwSentence;

    szIMEStateStr[0] = '\0';

    hImc = ImmGetContext(MAME32App.m_hWnd);
    bOpen = ImmGetOpenStatus(hImc);
    if (bOpen)
        ImmGetConversionStatus(hImc, &dwConversion, &dwSentence);
    ImmReleaseContext(MAME32App.m_hWnd, hImc);

    if (bOpen == TRUE)
    {
#if 0
        /* Sp/p */
        if (dwConversion & IME_CMODE_FULLSHAPE)
            strcpy(szIMEStateStr, "S");
        else
            strcpy(szIMEStateStr, "");
#endif

        /* ̓R[h */
        switch (dwConversion & 0x000f)
        {
        case IME_CMODE_CHARCODE:
            strcat(szIMEStateStr, "");
            break;

        case IME_CMODE_NATIVE | IME_CMODE_FULLSHAPE:
            strcat(szIMEStateStr, "");
            break;

        case IME_CMODE_NATIVE | IME_CMODE_FULLSHAPE | IME_CMODE_KATAKANA:
            strcat(szIMEStateStr, "J");
            break;

        case IME_CMODE_NATIVE | IME_CMODE_KATAKANA:
            strcat(szIMEStateStr, " ");
            break;

        case IME_CMODE_FULLSHAPE:
            strcat(szIMEStateStr, "`");
            break;

        default:
            strcat(szIMEStateStr, "A ");
            break;
        }

#if 0
        /* ϊ[h */
        if (dwSentence & IME_SMODE_PLAURALCLAUSE)
            strcat(szIMEStateStr, "");
        else if (dwSentence & IME_SMODE_AUTOMATIC)
            strcat(szIMEStateStr, "");
        else if (dwSentence & IME_SMODE_PHRASEPREDICT)
            strcat(szIMEStateStr, "A");
        else if (dwSentence & IME_SMODE_CONVERSATION)
            strcat(szIMEStateStr, "b");
        else if (!(dwConversion & IME_CMODE_CHARCODE))
            strcat(szIMEStateStr, "");

        /* [}/Ji */
        if (dwConversion & IME_CMODE_ROMAN)
            strcat(szIMEStateStr, "q");
        else
            strcat(szIMEStateStr, "");
#endif
    }
    else
    {
#if 0
        strcat(szIMEStateStr, "");
#endif
        strcat(szIMEStateStr, "A ");
#if 0
        strcat(szIMEStateStr, "");
        strcat(szIMEStateStr, "");
#endif
    }
}

static BOOL MSIME2K_GetReg(void)
{
    HKEY  hKey;
    LONG  lResult;

    lResult = RegOpenKeyEx(HKEY_CURRENT_USER,
                           "Software\\Microsoft\\Windows\\CurrentVersion\\Ime\\Japan\\IMEJP\\MSIME",
                           0, KEY_QUERY_VALUE, &hKey);

    if (lResult == ERROR_SUCCESS)
    {
        DWORD dwType;
        DWORD dwSize;
        DWORD dwValue;

        dwType = REG_DWORD;
        dwSize = 4;
        RegQueryValueEx(hKey, "HideComment", 0, &dwType, (LPBYTE)&dwValue, &dwSize);
        RegCloseKey(hKey);
        return (dwValue != 0);
    }
    return FALSE;
}

static void GetWinVersion(BOOL *bIsNT, int *nMajor, int *nMinor, int *nBuildNumber)
{
    OSVERSIONINFO VerInfo;

    VerInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

    GetVersionEx(&VerInfo);

    switch (VerInfo.dwPlatformId)
    {
    case VER_PLATFORM_WIN32_WINDOWS:
        *bIsNT        = FALSE;
        *nMajor       = HIWORD(VerInfo.dwBuildNumber) >> 8;
        *nMinor       = HIWORD(VerInfo.dwBuildNumber) & 0xff;
        *nBuildNumber = LOWORD(VerInfo.dwBuildNumber);
        break;

    case VER_PLATFORM_WIN32_NT:
        *bIsNT        = TRUE;
        *nMajor       = VerInfo.dwMajorVersion;
        *nMinor       = VerInfo.dwMinorVersion;
        *nBuildNumber = VerInfo.dwBuildNumber;
        break;

    case VER_PLATFORM_WIN32s:
    default:
        *bIsNT        = FALSE;
        *nMajor       = 0;
        *nMinor       = 0;
        *nBuildNumber = 0;
        break;
    }
}

