/***************************************************************************

    M.A.M.E.32  -  Multiple Arcade Machine Emulator for Win32
    Win32 Portions Copyright (C) 1997-98 Michael Soderstrom and Chris Kirmse

    This file is part of MAME32, and may only be used, modified and
    distributed under the terms of the MAME license, in "readme.txt".
    By continuing to use, modify or distribute this file you indicate
    that you have read the license and understand and accept it fully.

 ***************************************************************************/

 /***************************************************************************

  audit32.c

  Audit dialog

***************************************************************************/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <stdio.h>

#include "resource.h"

#include "driver.h"
#include "audit.h"

#include "win32ui.h"
#include "audit32.h"
#include "Properties.h"
#include "unzip.h"
#include "win32text.h"

/***************************************************************************
    function prototypes
 ***************************************************************************/

static INT_PTR CALLBACK AuditWindowProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam);
static void ProcessNextRom(void);
static void AddDetailsText(char *text);
static void __cdecl DetailsPrintf(char *fmt, ...);
static char * StatusString(int iStatus);

/***************************************************************************
    Internal variables
 ***************************************************************************/

HWND hAudit;
static int rom_index;
static int roms_correct;
static int roms_incorrect;

/***************************************************************************
    External functions
 ***************************************************************************/

void AuditDialog(HWND hParent)
{
    rom_index      = 0;
    roms_correct   = 0;
    roms_incorrect = 0;

    DialogBox(GetModuleHandle(NULL),
              MAKEINTRESOURCE(IDD_AUDIT),
              hParent,
              AuditWindowProc);
}

void InitGameAudit(int gameIndex)
{
    rom_index = gameIndex;
}

/***************************************************************************
    Internal functions
 ***************************************************************************/

static INT_PTR CALLBACK AuditWindowProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam)
{
    char buffer[200];
    static bPaused = FALSE;

    switch (Msg)
    {
    case WM_INITDIALOG:
        hAudit = hDlg;
        SendDlgItemMessage(hDlg,IDC_ROMS_PROGRESS,PBM_SETRANGE,0,MAKELPARAM(0,GetNumGames()));
        bPaused = FALSE;
        rom_index = 0;

        SetTimer(hDlg,0,1,NULL);
        return 1;

    case WM_TIMER:
        if (rom_index != -1)
        {
            sprintf(buffer, "%s %s - %s",
				win32_getstring(WIN32_audit_checkgame),
                drivers[rom_index]->name, drivers[rom_index]->description);
            SetWindowText(hDlg, buffer);
            ProcessNextRom();
        }
        else
        {
            KillTimer(hDlg,0);
            SetWindowText(hDlg, win32_getstring(WIN32_audit_fileaudit));
            EnableWindow(GetDlgItem(hDlg,ID_PAUSE), FALSE);
        }
        break;

    case WM_COMMAND:
        switch(LOWORD(wParam))
        {
        case IDCANCEL:
            EndDialog(hDlg,0);
            break;

        case ID_PAUSE:
            if (bPaused)
            {
                SendDlgItemMessage(hAudit,ID_PAUSE,WM_SETTEXT,0,(LPARAM)win32_getstring(WIN32_audit_pause));
                SetTimer(hDlg,0,1,NULL);
                bPaused = FALSE;
            }
            else
            {
                KillTimer(hDlg, 0);
                SendDlgItemMessage(hAudit,ID_PAUSE,WM_SETTEXT,0,(LPARAM)win32_getstring(WIN32_audit_continue));
                bPaused = TRUE;
            }
            break;
        }
        return 1;
    }
    return 0;
}

/* Callback for the Audit property sheet */
INT_PTR CALLBACK GameAuditDialogProc(HWND hDlg,UINT Msg,WPARAM wParam,LPARAM lParam)
{
    switch (Msg)
    {
    case WM_INITDIALOG :
        unzip_cache_clear();
        hAudit = hDlg;
        Static_SetText(GetDlgItem(hDlg, IDC_PROP_TITLE), GameInfoTitle(rom_index));
        SetTimer(hDlg,0,1,NULL);
        return 1;

    case WM_TIMER :
        KillTimer(hDlg, 0);
        {
            int iStatus;

            iStatus = VerifyRomSet(rom_index, DetailsPrintf);
            SetHasRoms(rom_index, (iStatus == CORRECT || iStatus == BEST_AVAILABLE) ? 1 : 0);
            SetWindowText(GetDlgItem(hDlg, IDC_PROP_ROMS), StatusString(iStatus));
        }
        ShowWindow(hDlg, SW_SHOW);
        break;
    }
    return 0;
}

static void ProcessNextRom()
{
    int retval;
    char buffer[200];

    retval = VerifyRomSet(rom_index, DetailsPrintf);
    switch (retval)
    {
    case BEST_AVAILABLE: /* correct, incorrect or separate count? */
    case CORRECT:
        roms_correct++;
        sprintf(buffer, "%i", roms_correct);
        SendDlgItemMessage(hAudit, IDC_ROMS_CORRECT, WM_SETTEXT, 0, (LPARAM)buffer);
        sprintf(buffer, "%i", roms_correct + roms_incorrect);
        SendDlgItemMessage(hAudit, IDC_ROMS_TOTAL, WM_SETTEXT, 0, (LPARAM)buffer);
        break;

    case NOTFOUND:
        break;

    case INCORRECT:
        roms_incorrect++;
        sprintf(buffer, "%i", roms_incorrect);
        SendDlgItemMessage(hAudit, IDC_ROMS_INCORRECT, WM_SETTEXT, 0, (LPARAM)buffer);
        sprintf(buffer, "%i", roms_correct + roms_incorrect);
        SendDlgItemMessage(hAudit, IDC_ROMS_TOTAL, WM_SETTEXT, 0, (LPARAM)buffer);
        break;
    }

    SetHasRoms(rom_index, (retval == CORRECT || retval == BEST_AVAILABLE) ? 1 : 0);
    rom_index++;
    SendDlgItemMessage(hAudit, IDC_ROMS_PROGRESS, PBM_SETPOS, rom_index, 0);

    if (rom_index == GetNumGames())
    {
        DetailsPrintf((char *)win32_getstring(WIN32_audit_complete));
        SendDlgItemMessage(hAudit,IDCANCEL,WM_SETTEXT,0,(LPARAM)win32_getstring(WIN32_audit_close));
        rom_index = -1;
    }
}

static void __cdecl DetailsPrintf(char *fmt,...)
{
    HWND hEdit;
    va_list marker;
    char buffer[2000];
    char buffer2[4000];
    int source,dest;

    hEdit = GetDlgItem(hAudit,IDC_AUDIT_DETAILS);

    va_start(marker,fmt);

    vsprintf(buffer,fmt,marker);

    va_end(marker);

    /* convert \n's to \r\n */
    source = 0;
    dest = 0;
    while (buffer[source] != 0)
    {
       if (buffer[source] == '\n')
       {
          buffer2[dest++] = '\r';
          buffer2[dest++] = '\n';
       }
       else
          buffer2[dest++] = buffer[source];
       source++;
    }
    buffer2[dest] = 0;

    Edit_SetSel(hEdit, Edit_GetTextLength(hEdit), Edit_GetTextLength(hEdit));
    Edit_ReplaceSel(hEdit,buffer2);
}

static char * StatusString(int iStatus)
{
    static const char *ptr;

    ptr = win32_getstring(WIN32_audit_nonerequired);

    switch (iStatus)
    {
    case CORRECT :
        ptr = win32_getstring(WIN32_audit_passed);
        break;

    case BEST_AVAILABLE :
        ptr = win32_getstring(WIN32_audit_bestavailable);
        break;

    case CLONE_NOTFOUND:
    case NOTFOUND:
        ptr = win32_getstring(WIN32_audit_notfound);
        break;

    case INCORRECT :
        ptr = win32_getstring(WIN32_audit_failed);
        break;
    }

    return (char *)ptr;
}

