/*********************************************************************

	cheat.c

	This is a massively rewritten version of the cheat engine.

	Busted (probably permanently) is the ability to use the UI_CONFIGURE
	key to pop the menu system on and off while retaining the current
	menu selection. The sheer number of submenus involved makes this
	a difficult task to pull off here.

	TODO:

	* key shortcuts to the search menus
		how should I implement this? add interface keys?

		if I use a modifier to enable the cheat keys, if people customize
		their key settings they will need to do a KEY and (not MODIFIER)
		config arrangment to bind to the unmodified key (that's not ideal)

		ideas:
			control + fkeys
			YUIOP
			use F5 to enable the cheat engine keys, then use YUIOP
	* check for compatibility with MESS
	* shift code over to flags-based cheat types
	* help text
	* show maximum number of characters available for cheat descriptions without overflow
	* better text editing
	* searching for values longer than one byte
	* remove duplicate cheats
	* variable length arrays

  	04092001	Ian Patterson	cheat searching by value
								cheat searching by comparison (energy search)
								saving
								watch cheats
								support for 'Get x NOW!' cheats
								moved add/remove cheat buttons to A and D
									(I use a portable which doesn't have INS and DEL keys)
								bugfixes
									mostly in textedit
	04102001	Ian Patterson	fixed specifying watches from the results list
								adding cheats from the result list
								made the message boxes go away faster
								bugfixes
	04112001	Ian Patterson	time cheat searching (delta search)
								bit cheat searching
								deleting a result from the result list
	04122001	Ian Patterson	commenting cheats
								textedit fixes
								user-specified value cheats
								bugfixes
								updated cheat special documentation (below)
								to include 22-24 and 42-44
								disabling watches (individually and all at once)
								fixed restoring previous search data
	04132001	Ian Patterson	user-configurable keys
								ui_text support
								watching cheats from edit menu
	04142001	Ian Patterson	fixed compile errors
								changed 'lives (or some other value)' to 'lives
								(or another value)'
								added plus sign before delta search delta
								'help not yet available' message
								converted all hex display to upper case
								one shot cheats now say 'set' (and act like a button)
								selection bar skips over comment cheats in activation menu
								cheat name templates
								conflict checking
								bugfixes
	04152001	Ian Patterson	stupid bugfix
	04172001	Ian Patterson	saving MESS cheats? (untested)
	04202001	Ian Patterson	auto-pause on entry of menu (compile-time option)
								converting watches into cheats
								saving watches
	04282001	Ian Patterson	fixed small text corruption bug
	05042001	Steph			added cheat types 100 and 101 which patch ROM areas
	06012001	Steph			fixed cheat saving bug (incompatibility with multiple cheat files)
				Ian Patterson	fixed small bug with conflict checking and new ROM cheat types
	06022001	Ian Patterson	added 'speed ramping' for auto-repeat
	06032001	Ian Patterson	fixed integer size bug which prevented menus from having >128 items
								increased MAX_LOADEDCHEATS to 500
	06062001	Ian Patterson	added slow but sure search
								added options menu
								added user memory area selection
	06072001	Ian Patterson	finalized user memory area selection
								updated ui_text
	06082001	Ian Patterson	added true backspace support
	06092001	Ian Patterson	kludging around missing osd_readkey_unicode in MAMEW
								skips multiple consecutive sound CPUs
								address alt step change to 0x100
								address display length truncation based on CPU address bits
	06102001	Ian Patterson	increased MAX_LOADEDCHEATS to 3000 (ddsomj has about 1200 right now)
	06112001	Ian Patterson	fixed typing error
								ANSI compatibility *sigh*
	06122001	Ian Patterson	documented cheat types 100 and 101
								results list shows first and last search values
								added page up/down support to watchpoints and results list
								more comments
								compile fixes
								added "very slow" speed (all non constant memory)
								added W+S (not A+D) support to enable/disable cheat menu
	06132001	Ian Patterson	made base horizontal change speed faster
								added activation key support
								general bugfixing
	06162001	Ian Patterson	changed activation key input function to code_pressed to avoid eating keydowns
	06182001	Ian Patterson	ANSI again
								bugfixes
								shift+return for value search searches for exact value
								watch saving bugfix
								added cheat types 102, 103, 110, 111, 112, and 113
								edit cheat menu labels fixed
								endian issues solved? (heh...)
	06192001	Ian Patterson	watching a cheat with subcheats watches _all_ addresses
								edit cheat menu labels fixed more
								added watch/save cheat support to the cheat edit window
								fixed redraw for vector games
								added extra info viewing for one-shot cheats (shift+return)
	06222001	Ian Patterson	removed "more info..." message
								fixed extra info viewing
								fixed max/min display
	06242001	Ian Patterson	added hex/dec/binary display for watches
								fixed user select with poke plus one
								added add/delete/save to edit watchpoint menu
								removed useless submenu_choice variables
	06252001	Ian Patterson	fixed activation keys
								fixed watch redraw
								removed 'magic numbers'
								added '+' prefix to title display for commented cheats
	06272001	Ian Patterson	fixed user select cheats
								ROM modification cheats now revert to their original value when disabled
								MESS compatibility?
								removed debug stuff
	06282001	Ian Patterson	removed useless decimal viewing for BCD user select
								fixed activation keys (and memory thrashing)
	07022001	Ian Patterson	fixed ROM cheat value reversion
								converting watches to cheats now makes normal cheats
								fixed for CPSMAME
	07092001	Ian Patterson	added 'all memory' search speed
	07122001	Ian Patterson	when cheat conflicts are found, the conflicting cheat is disabled
	07132001	Ian Patterson	removed CW-specific macro
								fixed conflict disabling
	07182001	Ian Patterson	added cheat types 15-17
	08062001	Ian Patterson	S and W keys now deactivated when typing names/descriptions
								added direct input to user select, edit (max/data field),
								start/continue search value
								added save/add all support to watch menu
	08182001	Ian Patterson	updated conflict checking algorithm to support rom and memory handler cheats
	09162001	Ian Patterson	added eeprom patching cheat types
	09272001	Ian Patterson	added cheat types 80-95
	09292001	Ian Patterson	fixes for rc.c

(Description from Pugsy's cheat.dat)

; all fields are separated by a colon (:)
;  -Name of the game (short name) [zip file or directory]
;  -No of the CPU usually 0, only different in multiple CPU games
;  -Address in Hexadecimal (where to poke)
;  -Data to put at this address in hexadecimal (what to poke)
;  -Special (see below) usually 000
;   -000 : the byte is poked every time and the cheat remains in active list.
;   -001 : the byte is poked once and the cheat is removed from active list.
;   -002 : the byte is poked every one second and the cheat remains in active
;          list.
;   -003 : the byte is poked every two seconds and the cheat remains in active
;          list.
;   -004 : the byte is poked every five seconds and the cheat remains in active
;          list.
;   -005 : the byte is poked one second after the original value has changed
;          and the cheat remains in active list.
;   -006 : the byte is poked two seconds after the original value has changed
;          and the cheat remains in active list.
;   -007 : the byte is poked five seconds after the original value has changed
;          and the cheat remains in active list.
;   -008 : the byte is poked unless the original value in being decremented by
;          1 each frame and the cheat remains in active list.
;   -009 : the byte is poked unless the original value in being decremented by
;          2 each frame and the cheat remains in active list.
;   -010 : the byte is poked unless the original value in being decremented by
;          3 each frame and the cheat remains in active list.
;   -011 : the byte is poked unless the original value in being decremented by
;          4 each frame and the cheat remains in active list.
;   -015 : the byte is poked one second after the original value has changed
;          and the cheat is removed from the active list.
;   -016 : the byte is poked two seconds after the original value has changed
;          and the cheat is removed from the active list.
;   -017 : the byte is poked five seconds after the original value has changed
;          and the cheat is removed from the active list.
;   -020 : the bits are set every time and the cheat remains in active list.
;   -021 : the bits are set once and the cheat is removed from active list.
;	-022 : the bits are set every second and the cheat remains in active list.
;	-023 : the bits are set every two seconds and the cheat remains in active
;          list.
;	-024 : the bits are set every five seconds and the cheat remains in active
;          list.
;   -040 : the bits are reset every time and the cheat remains in active list.
;   -041 : the bits are reset once and the cheat is removed from active list.
;	-042 : the bits are reset every second and the cheat remains in active
;          list.
;	-043 : the bits are reset every two seconds and the cheat remains in active
;          list.
;	-044 : the bits are reset every five seconds and the cheat remains in
;          active list.
;   -060 : the user selects a decimal value from 0 to byte
;          (display : 0 to byte) - the value is poked once when it changes and
;          the cheat is removed from the active list.
;   -061 : the user selects a decimal value from 0 to byte
;          (display : 1 to byte+1) - the value is poked once when it changes
;          and the cheat is removed from the active list.
;   -062 : the user selects a decimal value from 1 to byte
;          (display : 1 to byte) - the value is poked once when it changes and
;          the cheat is removed from the active list.
;   -063 : the user selects a BCD value from 0 to byte
;          (display : 0 to byte) - the value is poked once when it changes and
;          the cheat is removed from the active list.
;   -064 : the user selects a BCD value from 0 to byte
;          (display : 1 to byte+1) - the value is poked once when it changes
;          and the cheat is removed from the active list.
;   -065 : the user selects a decimal value from 1 to byte
;          (display : 1 to byte) - the value is poked once when it changes and
;          the cheat is removed from the active list.
;   -070 : the user selects a decimal value from 0 to byte
;          (display : 0 to byte) - the value is poked once and the cheat is
;          removed from the active list.
;   -071 : the user selects a decimal value from 0 to byte
;          (display : 1 to byte+1) - the value is poked once and the cheat is
;          removed from the active list.
;   -072 : the user selects a decimal value from 1 to byte
;          (display : 1 to byte) - the value is poked once and the cheat is
;          removed from the active list.
;   -073 : the user selects a BCD value from 0 to byte
;          (display : 0 to byte) - the value is poked once and the cheat is
;          removed from the active list.
;   -074 : the user selects a BCD value from 0 to byte
;          (display : 1 to byte+1) - the value is poked once and the cheat is
;          removed from the active list.
;   -075 : the user selects a decimal value from 1 to byte
;          (display : 1 to byte) - the value is poked once and the cheat is
;          removed from the active list.
;   -080 : the user selects a decimal value from 0 to byte
;          (display : 0 to byte) - the value is poked once when it changes and
;          the cheat stays in the active list.
;   -081 : the user selects a decimal value from 0 to byte
;          (display : 1 to byte+1) - the value is poked once when it changes
;          and cheat stays in the active list.
;   -082 : the user selects a decimal value from 1 to byte
;          (display : 1 to byte) - the value is poked once when it changes and
;          the cheat stays in the active list.
;   -083 : the user selects a BCD value from 0 to byte
;          (display : 0 to byte) - the value is poked once when it changes and
;          the cheat stays in the active list.
;   -084 : the user selects a BCD value from 0 to byte
;          (display : 1 to byte+1) - the value is poked once when it changes
;          and the cheat stays in the active list.
;   -085 : the user selects a decimal value from 1 to byte
;          (display : 1 to byte) - the value is poked once when it changes and
;          the cheat stays in the active list.
;   -090 : the user selects a decimal value from 0 to byte
;          (display : 0 to byte) - the value is poked once and the cheat
;          stays in the active list.
;   -091 : the user selects a decimal value from 0 to byte
;          (display : 1 to byte+1) - the value is poked once and the cheat
;          stays in the active list.
;   -092 : the user selects a decimal value from 1 to byte
;          (display : 1 to byte) - the value is poked once and the cheat
;          stays in the active list.
;   -093 : the user selects a BCD value from 0 to byte
;          (display : 0 to byte) - the value is poked once and the cheat
;          stays in the active list.
;   -094 : the user selects a BCD value from 0 to byte
;          (display : 1 to byte+1) - the value is poked once and the cheat
;          stays in the active list.
;   -095 : the user selects a decimal value from 1 to byte
;          (display : 1 to byte) - the value is poked once and the cheat
;          stays in the active list.
;	-100 : constantly pokes the value to the selected CPU's ROM region, adjusting for endian issues
;	-101 : pokes the value to the selected CPU's ROM region and the cheat is
;	       removed from the active list, adjusting for endian issues
;	-102 : constantly pokes the value to the selected CPU's ROM region
;	-103 : pokes the value to the selected CPU's ROM region and the cheat is
;	       removed from the active list
;	-110 : pokes the value to the memory mapped to the memory handler, adjusting for endian issues
;	-111 : pokes the value to the memory mapped to the memory handler, adjusting for endian issues, and
;	       the cheat is removed
;	-112 : pokes the value to the memory mapped to the memory handler
;	-113 : pokes the value to the memory mapped to the memory handler, and the cheat is removed
;	-120 : pokes the value to the eeprom buffer, and the cheat is removed
;	-121 : sets the bits in the eeprom buffer, and the cheat is removed
;	-122 : clears the bits in the eeprom buffer, and the cheat is removed
;   -500 to 575: These cheat types are identical to types 000 to 075 except
;                they are used in linked cheats (i.e. of 1/8 type). The first
;                cheat in the link list will be the normal type (eg type 000)
;                and the remaining cheats (eg 2/8...8/8) will be of this type
;                (eg type 500).
;   -998 : this is used as a watch cheat, ideal for showing answers in quiz
;          games .
;   -999 : this is used for comments only, cannot be enabled/selected by the
;          user.
;  -Name of the cheat
;  -Description for the cheat

*********************************************************************/

#include "driver.h"
#include "ui_text.h"

extern struct GameDriver driver_neogeo;

/*****************************************************************************/

#define MEMORY_READ(index,offset) \
    ((*cpuintf[Machine->drv->cpu[index].cpu_type & ~CPU_FLAGS_MASK]. \
    memory_read)(offset))
#define MEMORY_WRITE(index,offset,data) \
    ((*cpuintf[Machine->drv->cpu[index].cpu_type & ~CPU_FLAGS_MASK]. \
    memory_write)(offset,data))

static void computer_writemem_byte(int cpu, int addr, int value)
{
    int oldcpu = cpu_getactivecpu();
    memory_set_context(cpu);
    MEMORY_WRITE(cpu, addr, value);
    if (oldcpu != cpu)
        memory_set_context(oldcpu);
}

static int computer_readmem_byte(int cpu, int addr)
{
    int oldcpu = cpu_getactivecpu(), result;
    memory_set_context(cpu);
    result = MEMORY_READ(cpu, addr);
    if (oldcpu != cpu)
        memory_set_context(oldcpu);
    return result;
}

/******************************************
 *
 * Cheats
 *
 */

#define MAX_LOADEDCHEATS		3000
#define CHEAT_FILENAME_MAXLEN	255
#define MAX_SUBCHEATS			80

enum
{
	kCheatFlagActive =					1 << 0,
	kCheatFlagWatch =					1 << 1,
	kCheatFlagComment =					1 << 2,
	kCheatFlagDecPrompt =				1 << 3,
	kCheatFlagBCDPrompt =				1 << 4,
	kCheatFlagOneShot =					1 << 5,
	kCheatFlagActivationKeyPressed =	1 << 6,
	kCheatFlagDispPlusOne =				1 << 7,

	kCheatFlagPromptMask =	kCheatFlagDecPrompt |
							kCheatFlagBCDPrompt,
	kCheatFlagStatusMask =	kCheatFlagActive |
							kCheatFlagActivationKeyPressed,
	kCheatFlagModeMask =	kCheatFlagWatch |
							kCheatFlagComment |
							kCheatFlagPromptMask |
							kCheatFlagOneShot |
							kCheatFlagDispPlusOne
};

enum
{
	kSubcheatFlagDone =				1 << 0,
	kSubcheatFlagTimed =			1 << 1,
	kSubcheatFlagBitModify =		1 << 2,
	kSubcheatFlagByteModify =		1 << 3,
	kSubcheatFlagROMRegion =		1 << 4,
	kSubcheatFlagCustomRegion =		1 << 4,
	kSubcheatFlagPrompt =			1 << 6,
	kSubcheatFlagWatch =			1 << 7,
	kSubcheatFlagNoConflictCheck =	1 << 8,

	kSubcheatFlagModifyMask =	kSubcheatFlagBitModify |
								kSubcheatFlagByteModify,
	kSubcheatFlagStatusMask =	kSubcheatFlagDone |
								kSubcheatFlagTimed,
	kSubcheatFlagRegionMask =	kSubcheatFlagROMRegion |
								kSubcheatFlagCustomRegion,
	kSubcheatFlagModeMask =		kSubcheatFlagBitModify |
								kSubcheatFlagByteModify |
								kSubcheatFlagROMRegion |
								kSubcheatFlagCustomRegion |
								kSubcheatFlagPrompt |
								kSubcheatFlagWatch |
								kSubcheatFlagPrompt |
								kSubcheatFlagNoConflictCheck
};

struct subcheat_struct
{
	int		cpu;
	offs_t	address;
	data8_t	data;
	data8_t	olddata;			/* old data for code patch when cheat is turned OFF */
	data8_t	backup;				/* The original value of the memory location, checked against the current */
	UINT32	code;
	UINT16	flags;
	data8_t	min;
	data8_t	max;
	UINT32	frames_til_trigger;	/* the number of frames until this cheat fires (does not change) */
	UINT32	frame_count;		/* decrementing frame counter to determine if cheat should fire */
};

struct cheat_struct
{
	char					* name;
	char					* comment;
	UINT8					flags;
	int						num_sub;	/* number of cheat cpu/address/data/code combos for this one cheat */
	struct subcheat_struct	* subcheat;	/* a variable-number of subcheats are attached to each "master" cheat */
	int						activate_key;
};

enum
{
	kCheatSpecial_Poke =						0,
	kCheatSpecial_PokeRemove =					1,
	kCheatSpecial_Poke1 =						2,
	kCheatSpecial_Poke2 =						3,
	kCheatSpecial_Poke5 =						4,
	kCheatSpecial_Delay1 =						5,
	kCheatSpecial_Delay2 =						6,
	kCheatSpecial_Delay5 =						7,
	kCheatSpecial_Backup1 =						8,
	kCheatSpecial_Backup2 =						9,
	kCheatSpecial_Backup3 =						10,
	kCheatSpecial_Backup4 =						11,
	kCheatSpecial_Delay1Remove =				15,
	kCheatSpecial_Delay2Remove =				16,
	kCheatSpecial_Delay5Remove =				17,
	kCheatSpecial_SetBit =						20,
	kCheatSpecial_SetBitRemove =				21,
	kCheatSpecial_SetBit1 =						22,
	kCheatSpecial_SetBit2 =						23,
	kCheatSpecial_SetBit5 =						24,
	kCheatSpecial_ResetBit =					40,
	kCheatSpecial_ResetBitRemove =				41,
	kCheatSpecial_ResetBit1 =					42,
	kCheatSpecial_ResetBit2 =					43,
	kCheatSpecial_ResetBit5 =					44,
	kCheatSpecial_UserFirst =					60,
	kCheatSpecial_m0d0cRemove =					60,		/* minimum value 0, display range 0 to byte, poke when changed */
	kCheatSpecial_m0d1cRemove =					61,		/* minimum value 0, display range 1 to byte+1, poke when changed */
	kCheatSpecial_m1d1cRemove =					62,		/* minimum value 1, display range 1 to byte, poke when changed */
	kCheatSpecial_m0d0bcdcRemove =				63,		/* BCD, minimum value 0, display range 0 to byte, poke when changed */
	kCheatSpecial_m0d1bcdcRemove =				64,		/* BCD, minimum value 0, display range 1 to byte+1, poke when changed */
	kCheatSpecial_m1d1bcdcRemove =				65,		/* BCD, minimum value 1, display range 1 to byte, poke when changed */
	kCheatSpecial_m0d0Remove =					70,		/* minimum value 0, display range 0 to byte */
	kCheatSpecial_m0d1Remove =					71,		/* minimum value 0, display range 1 to byte+1 */
	kCheatSpecial_m1d1Remove =					72,		/* minimum value 1, display range 1 to byte */
	kCheatSpecial_m0d0bcdRemove =				73,		/* BCD, minimum value 0, display range 0 to byte */
	kCheatSpecial_m0d1bcdRemove =				74,		/* BCD, minimum value 0, display range 1 to byte+1 */
	kCheatSpecial_m1d1bcdRemove =				75,		/* BCD, minimum value 1, display range 1 to byte */
	kCheatSpecial_m0d0c =						80,		/* minimum value 0, display range 0 to byte, poke when changed */
	kCheatSpecial_m0d1c =						81,		/* minimum value 0, display range 1 to byte+1, poke when changed */
	kCheatSpecial_m1d1c =						82,		/* minimum value 1, display range 1 to byte, poke when changed */
	kCheatSpecial_m0d0bcdc =					83,		/* BCD, minimum value 0, display range 0 to byte, poke when changed */
	kCheatSpecial_m0d1bcdc =					84,		/* BCD, minimum value 0, display range 1 to byte+1, poke when changed */
	kCheatSpecial_m1d1bcdc =					85,		/* BCD, minimum value 1, display range 1 to byte, poke when changed */
	kCheatSpecial_m0d0 =						90,		/* minimum value 0, display range 0 to byte */
	kCheatSpecial_m0d1 =						91,		/* minimum value 0, display range 1 to byte+1 */
	kCheatSpecial_m1d1 =						92,		/* minimum value 1, display range 1 to byte */
	kCheatSpecial_m0d0bcd =						93,		/* BCD, minimum value 0, display range 0 to byte */
	kCheatSpecial_m0d1bcd =						94,		/* BCD, minimum value 0, display range 1 to byte+1 */
	kCheatSpecial_m1d1bcd =						95,		/* BCD, minimum value 1, display range 1 to byte */
	kCheatSpecial_UserLast =					95,
	kCheatSpecial_PokeROM =						100,
	kCheatSpecial_PokeROMRemove =				101,
	kCheatSpecial_PokeROMRaw =					102,
	kCheatSpecial_PokeROMRemoveRaw =			103,
	kCheatSpecial_ForcePokeRegion =				110,
	kCheatSpecial_ForcePokeRegionRemove =		111,
	kCheatSpecial_ForcePokeRegionRaw =			112,
	kCheatSpecial_ForcePokeRegionRemoveRaw =	113,
	kCheatSpecial_Last =						199,
	kCheatSpecial_LinkStart =					500,	/* only used when loading the database */
	kCheatSpecial_LinkEnd =						699,	/* only used when loading the database */
	kCheatSpecial_Watch =						998,
	kCheatSpecial_Comment =						999,
	kCheatSpecial_Timed =						1000
};

/* Steph 2001.05.04 - added types 100 and 101 */
const int kSupportedCheatTypes[] =
{
	kCheatSpecial_Poke,
	kCheatSpecial_PokeRemove,
	kCheatSpecial_Poke1,
	kCheatSpecial_Poke2,
	kCheatSpecial_Poke5,
	kCheatSpecial_Delay1,
	kCheatSpecial_Delay2,
	kCheatSpecial_Delay5,
	kCheatSpecial_Backup1,
	kCheatSpecial_Backup2,
	kCheatSpecial_Backup3,
	kCheatSpecial_Backup4,
	kCheatSpecial_Delay1Remove,
	kCheatSpecial_Delay2Remove,
	kCheatSpecial_Delay5Remove,
	kCheatSpecial_SetBit,
	kCheatSpecial_SetBitRemove,
	kCheatSpecial_SetBit1,
	kCheatSpecial_SetBit2,
	kCheatSpecial_SetBit5,
	kCheatSpecial_ResetBit,
	kCheatSpecial_ResetBitRemove,
	kCheatSpecial_ResetBit1,
	kCheatSpecial_ResetBit2,
	kCheatSpecial_ResetBit5,
	kCheatSpecial_m0d0cRemove,
	kCheatSpecial_m0d1cRemove,
	kCheatSpecial_m1d1cRemove,
	kCheatSpecial_m0d0bcdcRemove,
	kCheatSpecial_m0d1bcdcRemove,
	kCheatSpecial_m1d1bcdcRemove,
	kCheatSpecial_m0d0Remove,
	kCheatSpecial_m0d1Remove,
	kCheatSpecial_m1d1Remove,
	kCheatSpecial_m0d0bcdRemove,
	kCheatSpecial_m0d1bcdRemove,
	kCheatSpecial_m1d1bcdRemove,
	kCheatSpecial_m0d0cRemove,
	kCheatSpecial_m0d1cRemove,
	kCheatSpecial_m1d1cRemove,
	kCheatSpecial_m0d0bcdc,
	kCheatSpecial_m0d1bcdc,
	kCheatSpecial_m1d1bcdc,
	kCheatSpecial_m0d0,
	kCheatSpecial_m0d1,
	kCheatSpecial_m1d1,
	kCheatSpecial_m0d0bcd,
	kCheatSpecial_m0d1bcd,
	kCheatSpecial_m1d1bcd,
	kCheatSpecial_PokeROM,
	kCheatSpecial_PokeROMRemove,
	kCheatSpecial_PokeROMRaw,
	kCheatSpecial_PokeROMRemoveRaw,
	kCheatSpecial_ForcePokeRegion,
	kCheatSpecial_ForcePokeRegionRemove,
	kCheatSpecial_ForcePokeRegionRaw,
	kCheatSpecial_ForcePokeRegionRemoveRaw,
	kCheatSpecial_Watch,
	kCheatSpecial_Comment,
	-1
};

char *cheatfile = "cheat.dat";

char database[CHEAT_FILENAME_MAXLEN + 1];

int he_did_cheat;

/******************************************
 *
 * Searches
 *
 */

/* Defines */
#define MAX_SEARCHES 500

enum
{
	kRestore_NoInit = 0,
	kRestore_NoSave,
	kRestore_Done,
	kRestore_OK
};

enum
{
	kEnergy_Equals = 0,
	kEnergy_Less,
	kEnergy_Greater,
	kEnergy_LessOrEquals,
	kEnergy_GreaterOrEquals,
	kEnergy_NotEquals,
	kEnergy_FuzzyEquals
};

/******************************************
 *
 * Watchpoints
 *
 */

#define MAX_WATCHES 20

const int	kMaxWatchBytes = 16;

enum
{
	kWatchLabel_None = 0,
	kWatchLabel_Address,
	kWatchLabel_String,

	kWatchLabel_MaxPlusOne
};

enum
{
	kWatchDisplayType_Hex = 0,
	kWatchDisplayType_Decimal,
	kWatchDisplayType_Binary,

	kWatchDisplayType_MaxPlusOne
};

struct watch_struct
{
	int		cheat_num;		/* if this watchpoint is tied to a cheat, this is the index into the cheat array. -1 if none */
	UINT32	address;
	INT16	cpu;
	UINT8	num_bytes;		/* number of consecutive bytes to display */
	UINT8	display_type;	/* hex, dec, bin */
	UINT8	label_type;		/* none, address, text */
	char	label[255];		/* optional text label */
	UINT16	x;				/* position of watchpoint on screen */
	UINT16	y;
};

static struct watch_struct	watches[MAX_WATCHES];
static int					is_watch_active;		/* true if at least one watchpoint is active */
static int					is_watch_visible;		/* we can toggle the visibility for all on or off */

/* Some macros to simplify the code */
#define READ_CHEAT				computer_readmem_byte(subcheat->cpu, subcheat->address)
#define WRITE_CHEAT				computer_writemem_byte(subcheat->cpu, subcheat->address, subcheat->data)
#define COMPARE_CHEAT			(computer_readmem_byte(subcheat->cpu, subcheat->address) != subcheat->data)

/* Local prototypes */
static INT32	CommentMenu(struct osd_bitmap * bitmap, INT32 selected, int cheat_index);
static INT32	UserSelectValueMenu(struct osd_bitmap * bitmap, int selection, int cheat_num);
static int		FindFreeWatch(void);
static void		reset_table(struct ExtMemory * table);
static void		UnloadCheatDatabase(void);

/* Local variables */
static int					ActiveCheatTotal;				/* number of cheats currently active */
static int					LoadedCheatTotal;				/* total number of cheats */
static struct cheat_struct	CheatTable[MAX_LOADEDCHEATS];

static int					CheatEnabled = 1;
static int					cheatEngineWasActive;

static UINT8				memoryRegionEnabled[MAX_MEMORY_REGIONS];

/*****************************************************************************/

static int CPUAddressWidth(int cpu)
{
	int bits = cpunum_address_bits(cpu);

	if (bits & 0x3)
		return (bits >> 2) + 1;
	else
		return bits >> 2;
}

typedef union {
	struct	{ UINT8		h3, h2, h, l; } b;
	struct	{ UINT16	h, l; } w;
	UINT32	l;
}	PAIR_32;

typedef union {
	struct	{ UINT8		h, l; } b;
	UINT16	w;
}	PAIR_16;

/*	returns if a byte is BCD */
static int IsBCD(int value)
{
	return	((value & 0x0F) <= 0x09) &&
			((value & 0xF0) <= 0x90);
}

void poke_endian_fix(struct subcheat_struct * subcheat, UINT8 * basePtr, UINT8 inData)
{
	UINT32	offset;
	UINT32	dataWidth;
	UINT32	data;

	offset =	subcheat->address;
	dataWidth =	cpunum_databus_width(subcheat->cpu);
	data =		inData & 0xFF;

	switch (dataWidth)
	{
		case 8:
			offset = subcheat->address;
			break;

		case 16:
			offset = subcheat->address & ~1;
			break;

		case 32:
			offset = subcheat->address & ~3;
			break;

		default:
			logerror("poke_endian_fix: cpu %d databus width (%d) unsupported\n", subcheat->cpu, dataWidth);
			return;
	}

	basePtr = &basePtr[offset];

	switch (dataWidth)
	{
		case 8:
		{
			*basePtr = data;

			/* logerror("poke_endian_fix: 8 bit write (%.8X)\n", *basePtr); */
		}
		break;

		case 16:
		{
			PAIR_16	* target = (PAIR_16 *)basePtr;

			switch (subcheat->address & 1)
			{
				case 0:	target->b.h = data;	break;
				case 1:	target->b.l = data; break;
			}

			/* logerror("poke_endian_fix: 16 bit write (%.4X)\n", target->w); */
		}
		break;

		case 32:
		{
			PAIR_32	* target = (PAIR_32 *)basePtr;

			switch (subcheat->address & 3)
			{
				case 0:	target->b.h3 = data;	break;
				case 1:	target->b.h2 = data;	break;
				case 2:	target->b.h = data;		break;
				case 3:	target->b.l = data;		break;
			}

			/* logerror("poke_endian_fix: 32 bit write (%.8X)\n", target->l); */
		}
		break;
	}
}

UINT8 read_endian_fix(struct subcheat_struct * subcheat, UINT8 * basePtr)
{
	UINT32	offset;
	UINT32	dataWidth;

	offset =	subcheat->address;
	dataWidth =	cpunum_databus_width(subcheat->cpu);

	switch (dataWidth)
	{
		case 8:
			offset = subcheat->address;
			break;

		case 16:
			offset = subcheat->address & ~1;
			break;

		case 32:
			offset = subcheat->address & ~3;
			break;

		default:
			logerror("poke_endian_fix: cpu %d databus width (%d) unsupported\n", subcheat->cpu, dataWidth);
			return 0;
	}

	basePtr = &basePtr[offset];

	switch (dataWidth)
	{
		case 8:
		{
			return *basePtr;
		}
		break;

		case 16:
		{
			PAIR_16	* target = (PAIR_16 *)basePtr;

			switch (subcheat->address & 1)
			{
				case 0:	return target->b.h;
				case 1:	return target->b.l;
			}
		}
		break;

		case 32:
		{
			PAIR_32	* target = (PAIR_32 *)basePtr;

			switch (subcheat->address & 3)
			{
				case 0:	return target->b.h3;
				case 1:	return target->b.h2;
				case 2:	return target->b.h;
				case 3:	return target->b.l;
			}
		}
		break;
	}

	return 0;
}

/***************************************************************************

  patch_rom		Steph			2001.05.04
				Tourniquet		2001.05.10
				Ian Patterson	2001.06.06
								2001.06.12

  This patches ROM area by using a direct write to memory_region
  instead of using the write handlers

***************************************************************************/

void patch_rom(struct subcheat_struct * subcheat, UINT8 data)
{
	poke_endian_fix(subcheat, memory_region(REGION_CPU1 + subcheat->cpu), data);
}

UINT8 read_patch_rom(struct subcheat_struct * subcheat)
{
	return read_endian_fix(subcheat, memory_region(REGION_CPU1 + subcheat->cpu));
}

void patch_rom_raw(struct subcheat_struct * subcheat, UINT8 data)
{
	UINT8	* rom = memory_region(REGION_CPU1 + subcheat->cpu);

	rom[subcheat->address] = data;
}

UINT8 read_rom_raw(struct subcheat_struct * subcheat)
{
	UINT8	* rom = memory_region(REGION_CPU1 + subcheat->cpu);

	return rom[subcheat->address];
}

void force_poke_region(struct subcheat_struct * subcheat, UINT8 data)
{
	const struct Memory_WriteAddress	* mwa = Machine->drv->cpu[subcheat->cpu].memory_write;

	while (!IS_MEMPORT_END(mwa))
	{
		if (!IS_MEMPORT_MARKER(mwa))
		{
			if ((subcheat->address >= mwa->start) &&
				(subcheat->address <= mwa->end))
			{
				subcheat->address -= mwa->start;

				poke_endian_fix(subcheat, *mwa->base, data);

				subcheat->address += mwa->start;

				return;
			}
		}

		mwa++;
	}
}

UINT8 force_read_region(struct subcheat_struct * subcheat)
{
	const struct Memory_WriteAddress	* mwa = Machine->drv->cpu[subcheat->cpu].memory_write;
	UINT8								data;

	while (!IS_MEMPORT_END(mwa))
	{
		if (!IS_MEMPORT_MARKER(mwa))
		{
			if ((subcheat->address >= mwa->start) &&
				(subcheat->address <= mwa->end))
			{
				subcheat->address -= mwa->start;

				data = read_endian_fix(subcheat, *mwa->base);

				subcheat->address += mwa->start;

				return data;
			}
		}

		mwa++;
	}

	return 0;
}

void force_poke_region_raw(struct subcheat_struct * subcheat, UINT8 data)
{
	const struct Memory_WriteAddress	* mwa = Machine->drv->cpu[subcheat->cpu].memory_write;

	while (!IS_MEMPORT_END(mwa))
	{
		if (!IS_MEMPORT_MARKER(mwa))
		{
			if ((subcheat->address >= mwa->start) &&
				(subcheat->address <= mwa->end))
			{
				(*mwa->base)[subcheat->address - mwa->start] = data;

				return;
			}
		}

		mwa++;
	}
}

UINT8 force_read_region_raw(struct subcheat_struct * subcheat)
{
	const struct Memory_WriteAddress	* mwa = Machine->drv->cpu[subcheat->cpu].memory_write;

	while (!IS_MEMPORT_END(mwa))
	{
		if (!IS_MEMPORT_MARKER(mwa))
		{
			if ((subcheat->address >= mwa->start) &&
				(subcheat->address <= mwa->end))
			{
				return (*mwa->base)[subcheat->address - mwa->start];
			}
		}

		mwa++;
	}

	return 0;
}

/***************************************************************************

  cheat_set_code

  Given a cheat code, sets the various attribues of the cheat structure.
  This is to aid in making the cheat engine more flexible in the event that
  someday the codes are restructured or the case statement in DoCheat is
  simplified from its current form.

***************************************************************************/

void cheat_set_code(struct subcheat_struct * subcheat, int code, int cheat_num, int initMode)
{
	/* clear all flags (except mode flags) */
	CheatTable[cheat_num].flags &= ~kCheatFlagModeMask;
	subcheat->flags &= ~kSubcheatFlagModeMask;

	switch (code)
	{
		case kCheatSpecial_Poke1:
		case kCheatSpecial_Delay1:
		case kCheatSpecial_Delay1Remove:
		case kCheatSpecial_SetBit1:
		case kCheatSpecial_ResetBit1:
			subcheat->frames_til_trigger = 1 * Machine->drv->frames_per_second;
			break;

		case kCheatSpecial_Poke2:
		case kCheatSpecial_Delay2:
		case kCheatSpecial_Delay2Remove:
		case kCheatSpecial_SetBit2:
		case kCheatSpecial_ResetBit2:
			subcheat->frames_til_trigger = 2 * Machine->drv->frames_per_second;
			break;

		case kCheatSpecial_Poke5:
		case kCheatSpecial_Delay5:
		case kCheatSpecial_Delay5Remove:
		case kCheatSpecial_SetBit5:
		case kCheatSpecial_ResetBit5:
			subcheat->frames_til_trigger = 5 * Machine->drv->frames_per_second;
			break;

		case kCheatSpecial_Comment:
			subcheat->frames_til_trigger =	0;

			if (initMode)
			{
				subcheat->address =				0;
				subcheat->data =				0;
			}

			CheatTable[cheat_num].flags |=	kCheatFlagComment;
			break;

		case kCheatSpecial_Watch:
			subcheat->frames_til_trigger =	0;

			if (initMode)
			{
				subcheat->data =				0;
			}

			CheatTable[cheat_num].flags |=	kCheatFlagWatch;
			subcheat->flags |=				kSubcheatFlagWatch;
			break;

		default:
			subcheat->frames_til_trigger =	0;
			break;
	}

	if ((code == kCheatSpecial_PokeRemove) ||
	   ((code >= kCheatSpecial_Delay1Remove) &&
		(code <= kCheatSpecial_Delay5Remove)) ||
		(code == kCheatSpecial_SetBitRemove) ||
		(code == kCheatSpecial_ResetBitRemove) ||
		(code == kCheatSpecial_PokeROMRemove) ||		/* Steph 2001.05.04 */
		(code == kCheatSpecial_ForcePokeRegionRemove) ||
		(code == kCheatSpecial_ForcePokeRegionRemoveRaw) ||
		(	(code >= kCheatSpecial_m0d0cRemove) &&
			(code <= kCheatSpecial_m1d1bcdRemove)))
	{
		CheatTable[cheat_num].flags |=	kCheatFlagOneShot;
	}

	/* Set the minimum value */
	if ((code == kCheatSpecial_m1d1cRemove) ||
		(code == kCheatSpecial_m1d1bcdcRemove) ||
		(code == kCheatSpecial_m1d1Remove) ||
		(code == kCheatSpecial_m1d1bcdRemove) ||
		(code == kCheatSpecial_m1d1c) ||
		(code == kCheatSpecial_m1d1bcdc) ||
		(code == kCheatSpecial_m1d1) ||
		(code == kCheatSpecial_m1d1bcd))
	{
		subcheat->min = 1;
	}
	else
	{
		subcheat->min = 0;
	}

	/* Set the maximum value */
	if ((code >= kCheatSpecial_UserFirst) &&
		(code <= kCheatSpecial_UserLast))
	{
		if (initMode)
		{
			subcheat->max = subcheat->data;
		}
	}
	else
	{
		subcheat->max = 0xff;
	}

	/* Set prompting flags */
	if ((	(code >= kCheatSpecial_m0d0cRemove) &&
			(code <= kCheatSpecial_m1d1cRemove)) ||
		(	(code >= kCheatSpecial_m0d0Remove) &&
			(code <= kCheatSpecial_m1d1Remove)) ||
		(	(code >= kCheatSpecial_m0d0c) &&
			(code <= kCheatSpecial_m1d1c)) ||
		(	(code >= kCheatSpecial_m0d0) &&
			(code <= kCheatSpecial_m1d1)))
	{
		CheatTable[cheat_num].flags |=	kCheatFlagDecPrompt;
		subcheat->flags |=				kSubcheatFlagPrompt;
	}
	else if (	(	(code >= kCheatSpecial_m0d0bcdcRemove) &&
					(code <= kCheatSpecial_m1d1bcdcRemove)) ||
				(	(code >= kCheatSpecial_m0d0bcdRemove) &&
					(code <= kCheatSpecial_m1d1bcdRemove)) ||
				(	(code >= kCheatSpecial_m0d0bcdc) &&
					(code <= kCheatSpecial_m1d1bcdc)) ||
				(	(code >= kCheatSpecial_m0d0bcd) &&
					(code <= kCheatSpecial_m1d1bcd)))
	{
		CheatTable[cheat_num].flags |=	kCheatFlagBCDPrompt;
		subcheat->flags |=				kSubcheatFlagPrompt;
	}

	if ((code == kCheatSpecial_m0d1c) ||
		(code == kCheatSpecial_m0d1bcdc) ||
		(code == kCheatSpecial_m0d1) ||
		(code == kCheatSpecial_m0d1bcd))
	{
		CheatTable[cheat_num].flags |= kCheatFlagDispPlusOne;
	}

	if ((	(code >= kCheatSpecial_SetBit) &&
			(code <= kCheatSpecial_SetBit5)) ||
		(	(code >= kCheatSpecial_ResetBit) &&
			(code <= kCheatSpecial_ResetBit5)))
	{
		subcheat->flags |= kSubcheatFlagBitModify;
	}

	if ((	(code >= kCheatSpecial_Poke) &&
			(code <= kCheatSpecial_Backup4)) ||
		(code == kCheatSpecial_PokeROM) ||		/* Steph 2001.05.04 */
		(code == kCheatSpecial_PokeROMRemove) ||
		(code == kCheatSpecial_PokeROMRaw) ||
		(code == kCheatSpecial_PokeROMRemoveRaw) ||
		(code == kCheatSpecial_ForcePokeRegion) ||
		(code == kCheatSpecial_ForcePokeRegionRemove) ||
		(code == kCheatSpecial_ForcePokeRegionRaw) ||
		(code == kCheatSpecial_ForcePokeRegionRemoveRaw) ||
		(	(code >= kCheatSpecial_m0d0cRemove) &&
			(code <= kCheatSpecial_m1d1bcd)))
	{
		subcheat->flags |= kSubcheatFlagByteModify;
	}

	if ((code == kCheatSpecial_PokeROM) ||
		(code == kCheatSpecial_PokeROMRemove) ||
		(code == kCheatSpecial_PokeROMRaw) ||
		(code == kCheatSpecial_PokeROMRemoveRaw))
	{
		subcheat->flags |= kSubcheatFlagROMRegion;
	}

	if ((code == kCheatSpecial_ForcePokeRegion) ||
		(code == kCheatSpecial_ForcePokeRegionRemove) ||
		(code == kCheatSpecial_ForcePokeRegionRaw) ||
		(code == kCheatSpecial_ForcePokeRegionRemoveRaw))
	{
		subcheat->flags |= kSubcheatFlagCustomRegion;
	}

	subcheat->code = code;
}

/***************************************************************************

  cheat_set_status

  Given an index into the cheat table array, make the selected cheat
  either active or inactive.

***************************************************************************/

void cheat_set_status(int cheat_num, int active)
{
	int i, j, k;

	if (cheat_num >= LoadedCheatTotal)
		return;

	if (active) /* enable the cheat */
	{
		if (CheatTable[cheat_num].flags & kCheatFlagActive)
			return;

		/* check for conflict */
		for (i = 0; i <= CheatTable[cheat_num].num_sub; i++)
		{
			UINT32	address =	CheatTable[cheat_num].subcheat[i].address;
			UINT8	data =		CheatTable[cheat_num].subcheat[i].data;
			UINT32	flags =		CheatTable[cheat_num].subcheat[i].flags;
			UINT8	cpu =		CheatTable[cheat_num].subcheat[i].cpu;
			UINT8	conflict =	0;

			if (!flags & kSubcheatFlagNoConflictCheck)
			{
				for (j = 0; j < LoadedCheatTotal; j++)
				{
					if ((j != cheat_num) &&
						(CheatTable[j].flags & kCheatFlagActive))
					{
						for (k = 0; k <= CheatTable[j].num_sub; k++)
						{
							if ((CheatTable[j].subcheat[k].address == address) &&
								(CheatTable[j].subcheat[k].cpu == cpu) &&
								((CheatTable[j].subcheat[k].flags & kSubcheatFlagRegionMask) == (flags & kSubcheatFlagRegionMask)) &&
								!(CheatTable[j].subcheat[k].flags & kSubcheatFlagNoConflictCheck))
							{
								if (flags & kSubcheatFlagBitModify)
								{
									if ((CheatTable[j].subcheat[k].flags & kSubcheatFlagByteModify) &&
										data)
									{
										conflict = 1;
									}

									if ((CheatTable[j].subcheat[k].flags & kSubcheatFlagBitModify) &&
										(CheatTable[j].subcheat[k].data & data))
									{
										conflict = 1;
									}
								}

								if (flags & kSubcheatFlagByteModify)
								{
									if (CheatTable[j].subcheat[k].flags & kSubcheatFlagByteModify)
									{
										conflict = 1;
									}

									if ((CheatTable[j].subcheat[k].flags & kSubcheatFlagBitModify) &&
										CheatTable[j].subcheat[k].data)
									{
										conflict = 1;
									}
								}

								if (conflict)
								{
									/* no possible recursion for disabling */
									cheat_set_status(j, 0);

									if (CheatTable[j].name && CheatTable[j].name[0])
									{
										usrintf_showmessage_secs(1, "%s %s", ui_getstring(UI_conflict_found), CheatTable[j].name);
									}
									else
									{
										usrintf_showmessage_secs(1, "%s %s", ui_getstring(UI_conflict_found), ui_getstring(UI_none));
									}
								}

								conflict = 0;
							}
						}
					}
				}
			}
		}

		for (i = 0; i <= CheatTable[cheat_num].num_sub; i++)
		{
			struct subcheat_struct	* subcheat = &CheatTable[cheat_num].subcheat[i];

			/* Reset the active variables */
			subcheat->frame_count =	0;
			subcheat->backup =		0;
			subcheat->flags &=		~kSubcheatFlagStatusMask;

			/* add to the watch list (if needed) */
			if (subcheat->flags & kSubcheatFlagWatch)
			{
				int	freeWatch;

				freeWatch = FindFreeWatch();

				if (freeWatch != -1)
				{
					watches[freeWatch].cheat_num =		cheat_num;
					watches[freeWatch].address =		subcheat->address;
					watches[freeWatch].cpu =			subcheat->cpu;
					watches[freeWatch].num_bytes =		1;
					watches[freeWatch].display_type =	kWatchDisplayType_Hex;
					watches[freeWatch].label_type =		kWatchLabel_None;
					watches[freeWatch].label[0] =		0;

					is_watch_active = 1;
				}
			}

			switch (subcheat->code)
			{
				case kCheatSpecial_PokeROM:
				case kCheatSpecial_PokeROMRemove:
					subcheat->olddata = read_patch_rom(subcheat);
					break;

				case kCheatSpecial_PokeROMRaw:
				case kCheatSpecial_PokeROMRemoveRaw:
					subcheat->olddata = read_rom_raw(subcheat);
					break;

				case kCheatSpecial_ForcePokeRegion:
				case kCheatSpecial_ForcePokeRegionRemove:
					subcheat->olddata = force_read_region(subcheat);
					break;

				case kCheatSpecial_ForcePokeRegionRaw:
				case kCheatSpecial_ForcePokeRegionRemoveRaw:
					subcheat->olddata = force_read_region_raw(subcheat);
					break;
			}
		}

		CheatTable[cheat_num].flags |= kCheatFlagActive;
		ActiveCheatTotal++;

		/* tell the MAME core that we're cheaters! */
		he_did_cheat = 1;
	}
	else /* disable the cheat (case 0, 2) */
	{
		for (i = 0; i <= CheatTable[cheat_num].num_sub; i++)
		{
			struct subcheat_struct	* subcheat = &CheatTable[cheat_num].subcheat[i];

			/* Reset the active variables */
			subcheat->frame_count = 0;
			subcheat->backup = 0;

			switch (subcheat->code)
			{
				case kCheatSpecial_PokeROM:
					patch_rom(subcheat, subcheat->olddata);
					break;

				case kCheatSpecial_PokeROMRaw:
					patch_rom_raw(subcheat, subcheat->olddata);
					break;

				case kCheatSpecial_ForcePokeRegion:
					force_poke_region(subcheat, subcheat->olddata);
					break;

				case kCheatSpecial_ForcePokeRegionRaw:
					force_poke_region_raw(subcheat, subcheat->olddata);
					break;
			}
		}

		/* only add if there's a cheat active already */
		if (CheatTable[cheat_num].flags & kCheatFlagActive)
		{
			ActiveCheatTotal--;
		}

		CheatTable[cheat_num].flags &= ~kCheatFlagActive;

		/* disable watches associated with cheats */
		for (i = 0; i < MAX_WATCHES; i++)
		{
			if (watches[i].cheat_num == cheat_num)
			{
				watches[i].cheat_num =		-1;
				watches[i].address =		0;
				watches[i].cpu =			0;
				watches[i].num_bytes =		0;
				watches[i].display_type =	kWatchDisplayType_Hex;
				watches[i].label_type =		kWatchLabel_None;
				watches[i].label[0] =		0;

				schedule_full_refresh();
			}
		}
	}
}

/*	loads cheats from a specified file, optionally replacing all current cheats */
void LoadCheatFile(int merge, char * filename)
{
	void					* f;
	char					curline[2048];
	int						name_length;
	struct subcheat_struct	* subcheat;
	int						sub = 0;
	int						temp;

	f = osd_fopen (NULL, filename, OSD_FILETYPE_CHEAT, 0);

	if (!f)
		return;

	if (!merge)
	{
		ActiveCheatTotal = 0;
		LoadedCheatTotal = 0;
	}

	name_length = strlen(Machine->gamedrv->name);

	/* Load the cheats for that game */
	/* Ex.: pacman:0:4E14:06:000:1UP Unlimited lives:Coded on 1 byte */
	while ((osd_fgets (curline, 2048, f) != NULL) && (LoadedCheatTotal < MAX_LOADEDCHEATS))
	{
		char			* ptr;
		int				temp_cpu;
		offs_t			temp_address;
		data8_t			temp_data;
		INT32			temp_code;

		/* Take a few easy-out cases to speed things up */
		if (curline[name_length] != ':')
			continue;

		if (strncmp(curline, Machine->gamedrv->name, name_length))
			continue;

		if (curline[0] == ';')
			continue;

#if 0
		if (sologame)
			if ((strstr(str, "2UP") != NULL) || (strstr(str, "PL2") != NULL) ||
				(strstr(str, "3UP") != NULL) || (strstr(str, "PL3") != NULL) ||
				(strstr(str, "4UP") != NULL) || (strstr(str, "PL4") != NULL) ||
				(strstr(str, "2up") != NULL) || (strstr(str, "pl2") != NULL) ||
				(strstr(str, "3up") != NULL) || (strstr(str, "pl3") != NULL) ||
				(strstr(str, "4up") != NULL) || (strstr(str, "pl4") != NULL))
			continue;
#endif

		/* Extract the fields from the line */
		/* Skip the driver name */
		ptr = strtok(curline, ":");
		if (!ptr)
			continue;

		/* CPU number */
		ptr = strtok(NULL, ":");
		if (!ptr)
			continue;

		sscanf(ptr, "%d", &temp_cpu);

		/* skip if it's a sound cpu and the audio is off */
		/*
		if (CPU_AUDIO_OFF(temp_cpu))
			continue;
		*/

		/* Address */
		ptr = strtok(NULL, ":");
		if (!ptr)
			continue;

		sscanf(ptr, "%X", &temp_address);
		temp_address &= cpunum_address_mask(temp_cpu);

		/* data byte */
		ptr = strtok(NULL, ":");
		if (!ptr)
			continue;

		sscanf(ptr,"%x", &temp);
		temp_data = temp;

		temp_data &= 0xff;

		/* special code */
		ptr = strtok(NULL, ":");
		if (!ptr)
			continue;

		sscanf(ptr, "%d", &temp_code);

		/* skip if this is a bogus CPU */
		if ((temp_code != kCheatSpecial_PokeROM) &&
			(temp_code != kCheatSpecial_PokeROMRemove) &&
			(temp_cpu >= cpu_gettotalcpu()))
		{
			continue;
		}

		/* Is this a subcheat? */
		if ((temp_code >= kCheatSpecial_LinkStart) &&
			(temp_code <= kCheatSpecial_LinkEnd))
		{
			sub++;

			/* don't overflow */
			if ((sub + 1) > MAX_SUBCHEATS)
				continue;

			/* Adjust the special flag */
			temp_code -= kCheatSpecial_LinkStart;

			/* point to the last valid main cheat entry */
			LoadedCheatTotal--;
		}
		else
		{
			/* no, make this the first cheat in the series */
			sub = 0;
		}

		/* Allocate (or resize) storage for the subcheat array */
		CheatTable[LoadedCheatTotal].subcheat = MyRealloc(CheatTable[LoadedCheatTotal].subcheat, sizeof(struct subcheat_struct) * (sub + 1));
		if (CheatTable[LoadedCheatTotal].subcheat == NULL)
			continue;

		/* Store the current number of subcheats embodied by this code */
		CheatTable[LoadedCheatTotal].num_sub = sub;

		subcheat = &CheatTable[LoadedCheatTotal].subcheat[sub];

		/* Reset the cheat */
		subcheat->frames_til_trigger =	0;
		subcheat->frame_count =			0;
		subcheat->backup =				0;
		subcheat->flags =				0;

		/* Copy the cheat data */
		subcheat->cpu =					temp_cpu;
		subcheat->address =				temp_address;
		subcheat->data =				temp_data;

		cheat_set_code(subcheat, temp_code, LoadedCheatTotal, 1);

		/* don't bother with the names & comments for subcheats */
		if (sub)
			goto next;

		/* Disable the cheat */
		CheatTable[LoadedCheatTotal].flags &= ~kCheatFlagStatusMask;

		CheatTable[LoadedCheatTotal].activate_key = -1;

		/* cheat name */
		CheatTable[LoadedCheatTotal].name = NULL;
		ptr = strtok(NULL, ":");
		if (!ptr)
			continue;

		/* Allocate storage and copy the name */
		CheatTable[LoadedCheatTotal].name = MyMalloc(strlen(ptr) + 1);
		strcpy(CheatTable[LoadedCheatTotal].name, ptr);

		/* Strip line-ending if needed */
		if (strstr(CheatTable[LoadedCheatTotal].name, "\n") != NULL)
			CheatTable[LoadedCheatTotal].name[strlen(CheatTable[LoadedCheatTotal].name) - 1] = 0;

		/* read the "comment" field if there */
		ptr = strtok(NULL, ":");
		if (ptr)
		{
			/* Allocate storage and copy the comment */
			CheatTable[LoadedCheatTotal].comment = MyMalloc(strlen(ptr) + 1);
			strcpy(CheatTable[LoadedCheatTotal].comment, ptr);

			/* Strip line-ending if needed */
			if (strstr(CheatTable[LoadedCheatTotal].comment, "\n"))
				CheatTable[LoadedCheatTotal].comment[strlen(CheatTable[LoadedCheatTotal].comment) - 1] = 0;
		}
		else
		{
			CheatTable[LoadedCheatTotal].comment = MyCalloc(1, sizeof(char));
		}

		next:

		LoadedCheatTotal++;
	}

	osd_fclose(f);
}

/*	loads cheats from all database files */
void LoadCheatFiles(void)
{
	char	* ptr;
	char	str[CHEAT_FILENAME_MAXLEN + 1];
	char	filename[CHEAT_FILENAME_MAXLEN + 1];
	int		pos1;
	int		pos2;

	ActiveCheatTotal = 0;
	LoadedCheatTotal = 0;

	/* start off with the default cheat file, cheat.dat */
	strcpy(str, cheatfile);
	ptr = strtok(str, ";");

	/* append any additional cheat files */
	strcpy(database, ptr);
	strcpy(str, cheatfile);
	str[strlen(str) + 1] = 0;

	pos1 = 0;
	while (str[pos1])
	{
		pos2 = pos1;

		while ((str[pos2]) && (str[pos2] != ';'))
			pos2++;

		if (pos1 != pos2)
		{
			memset(filename, '\0', sizeof(filename));
			strncpy(filename, &str[pos1], pos2 - pos1);

			LoadCheatFile(1, filename);

			pos1 = pos2 + 1;
		}
	}
}

static void UnloadCheatDatabase(void)
{
	int i;

	for (i = 0; i < LoadedCheatTotal; i++)
	{
		/* free storage for the strings */
		if (CheatTable[i].name)
		{
			MyFree(CheatTable[i].name);
			CheatTable[i].name = NULL;
		}
		if (CheatTable[i].comment)
		{
			MyFree(CheatTable[i].comment);
			CheatTable[i].comment = NULL;
		}
	}

	memset(CheatTable, 0, sizeof(struct cheat_struct) * MAX_LOADEDCHEATS);

	LoadedCheatTotal = 0;
}

/*	inits the cheat engine */
void InitCheat(void)
{
	int i;

	he_did_cheat = 0;
	CheatEnabled = 1;

	memset(CheatTable, 0, sizeof(struct cheat_struct) * MAX_LOADEDCHEATS);

	/* Reset the watchpoints to their defaults */
	is_watch_active = 0;
	is_watch_visible = 1;

	for (i = 0; i < MAX_WATCHES; i++)
	{
		/* disable this watchpoint */
		watches[i].num_bytes =		0;

		watches[i].cheat_num =		-1;

		watches[i].cpu =			0;
		watches[i].label[0] =		0;
		watches[i].display_type =	kWatchDisplayType_Hex;
		watches[i].label_type =		kWatchLabel_None;
		watches[i].address =		0;

		/* set the screen position */
		watches[i].x =				0;
		watches[i].y =				i * Machine->uifontheight;
	}

	LoadCheatFiles();
}

/*	shows the "Enable/Disable a Cheat" menu */
INT32 cheat_menu(struct osd_bitmap *bitmap, int selected, int shortcut)
{
	INT32			sel;
	static INT32	submenu_choice;
	const char		* menu_item[MAX_LOADEDCHEATS + 4];
	const char		* menu_subitem[MAX_LOADEDCHEATS + 4];
	static char		buf[MAX_LOADEDCHEATS][80];	/* static to take load off stack */
	INT32			i;
	INT32			total = 0;

	sel = selected - 1;

	/* If a submenu has been selected, go there */
	if (submenu_choice)
	{
		switch (submenu_choice)
		{
			case 1:
				submenu_choice = CommentMenu(bitmap, submenu_choice, sel);
				break;

			case 2:
				submenu_choice = UserSelectValueMenu(bitmap, submenu_choice, sel);
				break;
		}

		if (submenu_choice == -1)
		{
			submenu_choice = 0;
			sel = -2;
		}

		return sel + 1;
	}

	/* No submenu active, do the watchpoint menu */
	for (i = 0; i < LoadedCheatTotal; i++)
	{
		if (CheatTable[i].comment && CheatTable[i].comment[0])
		{
			sprintf(buf[total], "+%s", CheatTable[i].name);
		}
		else
		{
			sprintf(buf[total], "%s", CheatTable[i].name);
		}

		menu_item[total] = buf[total];

		menu_subitem[total] = NULL;

		/* add submenu options for all cheats that are not comments */
		if (!(CheatTable[i].flags & kCheatFlagComment))
		{
			if (CheatTable[i].flags & kCheatFlagOneShot)
			{
				menu_subitem[total] = ui_getstring(UI_set);
			}
			else
			{
				if (CheatTable[i].flags & kCheatFlagActive)
				{
					menu_subitem[total] = ui_getstring(UI_on);
				}
				else
				{
					menu_subitem[total] = ui_getstring(UI_off);
				}
			}
		}

		total++;
	}

	if (shortcut)
		menu_item[total] = ui_getstring(UI_returntogame);
	else
		menu_item[total] = ui_getstring(UI_returntomain);

	menu_subitem[total] = NULL;
	total++;

	menu_item[total] = 0;	/* terminate array */
	menu_subitem[total] = 0;

	if (sel < 0)
		sel = 0;
	if (sel > (total - 1))
		sel = total - 1;

	while (	(sel < total - 1) &&
			CheatTable[sel].flags & kCheatFlagComment)
			sel++;

	ui_displaymenu(bitmap, menu_item, menu_subitem, 0, sel, 0);

	if (input_ui_pressed_repeat(IPT_UI_DOWN,8))
	{
		sel++;

		if (sel >= total)
			sel = 0;

		while (	(sel < total - 1) &&
				CheatTable[sel].flags & kCheatFlagComment)
			sel++;
	}

	if (input_ui_pressed_repeat(IPT_UI_UP,8))
	{
		sel--;

		if (sel < 0)
		{
			sel = total - 1;
		}
		else
		{
			while (	(sel != total - 1) &&
					CheatTable[sel].flags & kCheatFlagComment)
			{
				sel--;

				if (sel < 0)
					sel = total - 1;
			}
		}
	}

	if (code_pressed_memory_repeat(KEYCODE_PGUP, 8))
	{
		sel -= Machine->uiheight / (3 * Machine->uifontheight / 2) - 1;

		if (sel < 0)
		{
			sel = 0;

			while (	(sel < total - 1) &&
					CheatTable[sel].flags & kCheatFlagComment)
				sel++;
		}
		else
		{
			while (	(sel != total - 1) &&
					CheatTable[sel].flags & kCheatFlagComment)
			{
				sel--;

				if (sel < 0)
					sel = total - 1;
			}
		}
	}

	if (code_pressed_memory_repeat(KEYCODE_PGDN, 8))
	{
		sel += Machine->uiheight / (3 * Machine->uifontheight / 2) - 1;

		if (sel >= total)
		{
			sel = total - 1;
		}
		else
		{
			while (	(sel < total - 1) &&
					CheatTable[sel].flags & kCheatFlagComment)
				sel++;
		}
	}

	if ((input_ui_pressed_repeat(IPT_UI_LEFT,8)) || (input_ui_pressed_repeat(IPT_UI_RIGHT,8)))
	{
		if ((sel < (total - 1)) &&
			!(CheatTable[sel].flags & kCheatFlagComment) &&
			!(CheatTable[sel].flags & kCheatFlagOneShot))
		{
			int active = CheatTable[sel].flags & kCheatFlagActive;

			active ^= 0x01;

			/* get the user's selected value if needed */
			if ((CheatTable[sel].flags & kCheatFlagPromptMask) && active)
			{
				submenu_choice = 2;
				schedule_full_refresh();
			}
			else
			{
				cheat_set_status(sel, active);

				CheatEnabled = 1;
			}
		}
	}

	if (input_ui_pressed(IPT_UI_SELECT))
	{
		if (sel == (total - 1))
		{
			/* return to prior menu */
			submenu_choice = 0;
			sel = -1;
		}
		else if (sel < (total - 1))
		{
#if 0	// test code
			if (ShiftKeyPressed())
			{
				if (CheatTable[sel].comment && (CheatTable[sel].comment[0] != 0x00))
				{
					submenu_choice = 1;
					schedule_full_refresh();
				}
			}
			else
#endif	// test code
			{
				if (CheatTable[sel].flags & kCheatFlagPromptMask)
				{
					submenu_choice = 2;
					schedule_full_refresh();
				}
				else
				{
					cheat_set_status(sel, 1);

					CheatEnabled = 1;
				}
			}
		}
	}

	/* Cancel pops us up a menu level */
	if (input_ui_pressed(IPT_UI_CANCEL))
		sel = -1;

	/* The UI key takes us all the way back out */
	if (input_ui_pressed(IPT_UI_CONFIGURE))
		sel = -2;

	if (sel == -1 || sel == -2)
	{
		schedule_full_refresh();
	}

	return sel + 1;
}

/*	lets the user select a value for selection cheats */
static INT32 UserSelectValueMenu(struct osd_bitmap * bitmap, int selection, int cheat_num)
{
	char					buf[2048];
	int						sel;
	struct cheat_struct		* cheat;
	struct subcheat_struct	* subcheat;
	static int				value = -1;
	int						delta = 0;
	int						displayValue;
	int						keyValue;

	if (cheat_num >= LoadedCheatTotal)
	{
		schedule_full_refresh();

		return 0;
	}

	sel =		selection - 1;

	cheat =		&CheatTable[cheat_num];
	subcheat =	&cheat->subcheat[0];

	if (value == -1)
		value = subcheat->data;

	displayValue = value;

	if (cheat->flags & kCheatFlagDispPlusOne)
	{
		displayValue++;

		if (cheat->flags & kCheatFlagBCDPrompt)
		{
			while (!IsBCD(displayValue))
			{
				displayValue++;
			}
		}
	}

	if (cheat->flags & kCheatFlagBCDPrompt)
	{
		sprintf(buf, "\t%s\n\t%.2X\n", ui_getstring(UI_search_select_value), displayValue);
	}
	else
	{
		sprintf(buf, "\t%s\n\t%.2X (%d)\n", ui_getstring(UI_search_select_value), displayValue, displayValue);
	}

	/* menu system, use the normal menu keys */
	strcat(buf, "\t");
	strcat(buf, ui_getstring(UI_lefthilight));
	strcat(buf, " ");
	strcat(buf, ui_getstring(UI_OK));
	strcat(buf, " ");
	strcat(buf, ui_getstring(UI_righthilight));

	ui_displaymessagewindow(bitmap, buf);

	if (input_ui_pressed_repeat(IPT_UI_LEFT,8))
	{
		delta = -1;
	}
	if (input_ui_pressed_repeat(IPT_UI_RIGHT,8))
	{
		delta = 1;
	}

	if (input_ui_pressed(IPT_UI_SELECT))
	{
		if (value != -1)
		{
			subcheat->data = value;

			cheat_set_status(cheat_num, 1);

			CheatEnabled = 1;
		}

		value = -1;
		sel = -1;
	}

	if (input_ui_pressed(IPT_UI_CANCEL))
	{
		sel = -1;
	}

	if (input_ui_pressed(IPT_UI_CONFIGURE))
	{
		sel = -2;
	}

	keyValue = code_read_hex_async();

	if (keyValue != -1)
	{
		if (cheat->flags & kCheatFlagBCDPrompt)
		{
			value *= 10;
			value &= 0xFF;
			value += keyValue;
		}
		else
		{
			value <<= 4;
			value &= 0xF0;
			value |= keyValue & 0x0F;
		}

		delta = 0;
	}

	if (sel == -1 || sel == -2)
	{
		schedule_full_refresh();
	}

	{
		int	count = 0;

		do
		{
			value += delta;

			if (value < subcheat->min)
			{
				value = subcheat->max;
			}
			if (value > subcheat->max)
			{
				value = subcheat->min;
			}

			if (delta > 0)
				delta = 1;
			else if (delta < 0)
				delta = -1;
			else if (delta == 0)
				delta = 1;

			count++;
		}
		while (	(cheat->flags & kCheatFlagBCDPrompt) &&
				!IsBCD(value) &&
				(count < 1000));
	}

	return sel + 1;
}

/*	shows a cheat's comment */
static INT32 CommentMenu(struct osd_bitmap * bitmap, INT32 selected, int cheat_index)
{
	char	buf[2048];
	int		sel;

	if (cheat_index >= LoadedCheatTotal)
	{
		schedule_full_refresh();

		return 0;
	}

	sel = selected - 1;

	buf[0] = 0;

	if (CheatTable[cheat_index].comment[0])
	{
		sprintf(buf, "\t%s\n\t%s\n\n", ui_getstring (UI_moreinfoheader), CheatTable[cheat_index].name);

		strcat(buf, CheatTable[cheat_index].comment);
	}
	else
	{
		sel = -1;
		buf[0] = 0;
	}

	/* menu system, use the normal menu keys */
	strcat(buf, "\n\n\t");
	strcat(buf, ui_getstring (UI_lefthilight));
	strcat(buf, " ");
	strcat(buf, ui_getstring (UI_returntoprior));
	strcat(buf, " ");
	strcat(buf, ui_getstring (UI_righthilight));

	ui_displaymessagewindow(bitmap, buf);

	if (input_ui_pressed(IPT_UI_SELECT))
		sel = -1;

	if (input_ui_pressed(IPT_UI_CANCEL))
		sel = -1;

	if (input_ui_pressed(IPT_UI_CONFIGURE))
		sel = -2;

	if (sel == -1 || sel == -2)
	{
		schedule_full_refresh();
	}

	return sel + 1;
}

/* free all the memory and init the table */
static void reset_table(struct ExtMemory * table)
{
	struct ExtMemory	* ext;

	for (ext = table; ext->data; ext++)
		MyFree(ext->data);

	memset(table, 0, sizeof(struct ExtMemory) * MAX_EXT_MEMORY);
}

/*	prints a value in binary notation */
static void PrintBinary(char * buf, int value)
{
	/*	*cough* */
	buf[0] = (value & 0x80) ? '1' : '0';
	buf[1] = (value & 0x40) ? '1' : '0';
	buf[2] = (value & 0x20) ? '1' : '0';
	buf[3] = (value & 0x10) ? '1' : '0';
	buf[4] = (value & 0x08) ? '1' : '0';
	buf[5] = (value & 0x04) ? '1' : '0';
	buf[6] = (value & 0x02) ? '1' : '0';
	buf[7] = (value & 0x01) ? '1' : '0';
	buf[8] = 0;
}

/*	displays all watches */
void DisplayWatches(struct osd_bitmap * bitmap)
{
	int		i, j;
	char	buf[256];

	if (!is_watch_active || !is_watch_visible)
		return;

	for (i = 0; i < MAX_WATCHES; i++)
	{
		/* Is this watchpoint active? */
		if (watches[i].num_bytes != 0)
		{
			char buf2[80];

			switch (watches[i].display_type)
			{
				default:
				case kWatchDisplayType_Hex:
					/* Display the first byte */
					sprintf(buf, "%.2X", computer_readmem_byte(watches[i].cpu, watches[i].address));

					/* If this is for more than one byte, display the rest */
					for (j = 1; j < watches[i].num_bytes; j++)
					{
						sprintf(buf2, " %.2X", computer_readmem_byte(watches[i].cpu, watches[i].address + j));
						strcat(buf, buf2);
					}
					break;

				case kWatchDisplayType_Decimal:
					/* Display the first byte */
					sprintf(buf, "%.3d", computer_readmem_byte(watches[i].cpu, watches[i].address));

					/* If this is for more than one byte, display the rest */
					for (j = 1; j < watches[i].num_bytes; j++)
					{
						sprintf(buf2, " %.3d", computer_readmem_byte(watches[i].cpu, watches[i].address + j));
						strcat(buf, buf2);
					}
					break;

				case kWatchDisplayType_Binary:
					/* Display the first byte */
					PrintBinary(buf, computer_readmem_byte(watches[i].cpu, watches[i].address));

					/* If this is for more than one byte, display the rest */
					for (j = 1; j < watches[i].num_bytes; j++)
					{
						PrintBinary(buf2, computer_readmem_byte(watches[i].cpu, watches[i].address + j));

						strcat(buf, " ");
						strcat(buf, buf2);
					}
					break;
			}

			/* Handle any labels */
			switch (watches[i].label_type)
			{
				case kWatchLabel_None:
				default:
					break;

				case kWatchLabel_Address:
					sprintf(buf2, " (%.*X)", CPUAddressWidth(watches[i].cpu), watches[i].address);
					strcat(buf, buf2);
					break;

				case kWatchLabel_String:
					sprintf(buf2, " (%s)", watches[i].label);
					strcat(buf, buf2);
					break;
			}

			ui_text(bitmap, buf, watches[i].x, watches[i].y);
		}
	}
}

/*	returns the index of a free watch (-1 if none is found) */
static int FindFreeWatch(void)
{
	int i;

	for (i = 0; i < MAX_WATCHES; i++)
		if (watches[i].num_bytes == 0)
			return i;

	/* indicate no free watch found */
	return -1;
}

/*	shuts down the cheat engine */
void StopCheat(void)
{
	UnloadCheatDatabase();
}

/*	performs all cheat actions, called once per frame */
void DoCheat(struct osd_bitmap *bitmap)
{
	DisplayWatches(bitmap);

	if (CheatEnabled)
	{
		int i, j;
		int done;

		for (i = 0; i < LoadedCheatTotal; i++)
		{
			if ((CheatTable[i].activate_key != -1) &&
				(!(CheatTable[i].flags & kCheatFlagComment)) &&
				(!(CheatTable[i].flags & kCheatFlagPromptMask)) &&
				!cheatEngineWasActive)
			{
				/* use code_pressed to avoid issues */
				if (code_pressed(CheatTable[i].activate_key))
				{
					if (!(CheatTable[i].flags & kCheatFlagActivationKeyPressed))
					{
						if (CheatTable[i].flags & kCheatFlagOneShot)
						{
							cheat_set_status(i, 1);
						}
						else
						{
							if (CheatTable[i].flags & kCheatFlagActive)
								cheat_set_status(i, 0);
							else
								cheat_set_status(i, 1);
						}

						CheatTable[i].flags |= kCheatFlagActivationKeyPressed;
					}
				}
				else
				{
					CheatTable[i].flags &= ~kCheatFlagActivationKeyPressed;
				}
			}

			/* skip if this isn't an active cheat */
			if (!(CheatTable[i].flags & kCheatFlagActive))
				continue;

			/* loop through all subcheats */
			for (j = 0; j <= CheatTable[i].num_sub; j++)
			{
				struct subcheat_struct	* subcheat = &CheatTable[i].subcheat[j];

				if (subcheat->flags & kSubcheatFlagDone)
					continue;

				/* most common case: 0 */
				if (subcheat->code == kCheatSpecial_Poke)
				{
					WRITE_CHEAT;
				}
				/* Check special function if cheat counter is ready */
				else if (subcheat->frame_count == 0)
				{
					switch (subcheat->code)
					{
						case kCheatSpecial_PokeRemove:
							WRITE_CHEAT;
							subcheat->flags |= kSubcheatFlagDone;
							break;

						case kCheatSpecial_Poke1:
						case kCheatSpecial_Poke2:
						case kCheatSpecial_Poke5:
							WRITE_CHEAT;
							subcheat->frame_count = subcheat->frames_til_trigger;
							break;

						/* 5,6,7 check if the value has changed, if yes, start a timer. */
						/* When the timer ends, change the location */
						case kCheatSpecial_Delay1:
						case kCheatSpecial_Delay2:
						case kCheatSpecial_Delay5:
							if (subcheat->flags & kSubcheatFlagTimed)
							{
								WRITE_CHEAT;
								subcheat->flags &= ~kSubcheatFlagTimed;
							}
							else if (COMPARE_CHEAT)
							{
								subcheat->frame_count = subcheat->frames_til_trigger;
								subcheat->flags |= kSubcheatFlagTimed;
							}
							break;

						/* 8,9,10,11 do not change the location if the value change by X every frames
						  This is to try to not change the value of an energy bar
				 		  when a bonus is awarded to it at the end of a level
				 		  See Kung Fu Master */
						case kCheatSpecial_Backup1:
						case kCheatSpecial_Backup2:
						case kCheatSpecial_Backup3:
						case kCheatSpecial_Backup4:
							if (subcheat->flags & kSubcheatFlagTimed)
							{
								/* Check the value to see if it has increased over the original value by 1 or more */
								if (READ_CHEAT != subcheat->backup - (kCheatSpecial_Backup1 - subcheat->code + 1))
									WRITE_CHEAT;

								subcheat->flags &= ~kSubcheatFlagTimed;
							}
							else
							{
								subcheat->backup = READ_CHEAT;
								subcheat->frame_count = 1;
								subcheat->flags |= kSubcheatFlagTimed;
							}
							break;

						case kCheatSpecial_Delay1Remove:
						case kCheatSpecial_Delay2Remove:
						case kCheatSpecial_Delay5Remove:
							if (subcheat->flags & kSubcheatFlagTimed)
							{
								WRITE_CHEAT;
								subcheat->flags &= ~kSubcheatFlagTimed;

								subcheat->flags |= kSubcheatFlagDone;
							}
							else if (COMPARE_CHEAT)
							{
								subcheat->frame_count = subcheat->frames_til_trigger;
								subcheat->flags |= kSubcheatFlagTimed;
							}
							break;

						/* 20-24: set bits */
						case kCheatSpecial_SetBit:
							computer_writemem_byte(subcheat->cpu, subcheat->address, READ_CHEAT | subcheat->data);
							break;

						case kCheatSpecial_SetBitRemove:
							computer_writemem_byte(subcheat->cpu, subcheat->address, READ_CHEAT | subcheat->data);
							subcheat->flags |= kSubcheatFlagDone;
							break;

						case kCheatSpecial_SetBit1:
						case kCheatSpecial_SetBit2:
						case kCheatSpecial_SetBit5:
							computer_writemem_byte(subcheat->cpu, subcheat->address, READ_CHEAT | subcheat->data);
							subcheat->frame_count = subcheat->frames_til_trigger;
							break;

						/* 40-44: reset bits */
						case kCheatSpecial_ResetBit:
							computer_writemem_byte(subcheat->cpu, subcheat->address, READ_CHEAT & ~subcheat->data);
							break;

						case kCheatSpecial_ResetBitRemove:
							computer_writemem_byte(subcheat->cpu, subcheat->address, READ_CHEAT & ~subcheat->data);
							subcheat->flags |= kSubcheatFlagDone;
							break;

						case kCheatSpecial_ResetBit1:
						case kCheatSpecial_ResetBit2:
						case kCheatSpecial_ResetBit5:
							computer_writemem_byte(subcheat->cpu, subcheat->address, READ_CHEAT & ~subcheat->data);
							subcheat->frame_count = subcheat->frames_til_trigger;
							break;

						/* 60-65: user select, poke when changes */
						case kCheatSpecial_m0d0cRemove:
						case kCheatSpecial_m0d1cRemove:
						case kCheatSpecial_m1d1cRemove:
						case kCheatSpecial_m0d0bcdcRemove:
						case kCheatSpecial_m0d1bcdcRemove:
						case kCheatSpecial_m1d1bcdcRemove:
							if (subcheat->flags & kSubcheatFlagTimed)
							{
								if (READ_CHEAT != subcheat->backup)
								{
									WRITE_CHEAT;
									subcheat->flags |= kSubcheatFlagDone;
								}
							}
							else
							{
								subcheat->backup = READ_CHEAT;
								subcheat->frame_count = 1;
								subcheat->flags |= kSubcheatFlagTimed;
							}
							break;

						/* 70-75: user select, poke once */
						case kCheatSpecial_m0d0Remove:
						case kCheatSpecial_m0d1Remove:
						case kCheatSpecial_m1d1Remove:
						case kCheatSpecial_m0d0bcdRemove:
						case kCheatSpecial_m0d1bcdRemove:
						case kCheatSpecial_m1d1bcdRemove:
							WRITE_CHEAT;
							subcheat->flags |= kSubcheatFlagDone;
							break;

						/* 80-85: user select, poke when changes, don't remove */
						case kCheatSpecial_m0d0c:
						case kCheatSpecial_m0d1c:
						case kCheatSpecial_m1d1c:
						case kCheatSpecial_m0d0bcdc:
						case kCheatSpecial_m0d1bcdc:
						case kCheatSpecial_m1d1bcdc:
							if (subcheat->flags & kSubcheatFlagTimed)
							{
								if (READ_CHEAT != subcheat->backup)
								{
									WRITE_CHEAT;
								}
							}
							else
							{
								subcheat->backup = READ_CHEAT;
								subcheat->frame_count = 1;
								subcheat->flags |= kSubcheatFlagTimed;
							}
							break;

						/* 70-75: user select, poke once, don't remove */
						case kCheatSpecial_m0d0:
						case kCheatSpecial_m0d1:
						case kCheatSpecial_m1d1:
						case kCheatSpecial_m0d0bcd:
						case kCheatSpecial_m0d1bcd:
						case kCheatSpecial_m1d1bcd:
							WRITE_CHEAT;
							break;

						case kCheatSpecial_PokeROM:
							patch_rom(subcheat, subcheat->data);
							break;

						case kCheatSpecial_PokeROMRemove:
							patch_rom(subcheat, subcheat->data);
							subcheat->flags |= kSubcheatFlagDone;
							break;

						case kCheatSpecial_PokeROMRaw:
							patch_rom_raw(subcheat, subcheat->data);
							break;

						case kCheatSpecial_PokeROMRemoveRaw:
							patch_rom_raw(subcheat, subcheat->data);
							subcheat->flags |= kSubcheatFlagDone;
							break;

						case kCheatSpecial_ForcePokeRegion:
							force_poke_region(subcheat, subcheat->data);
							break;

						case kCheatSpecial_ForcePokeRegionRemove:
							force_poke_region(subcheat, subcheat->data);
							subcheat->flags |= kSubcheatFlagDone;
							break;

						case kCheatSpecial_ForcePokeRegionRaw:
							force_poke_region_raw(subcheat, subcheat->data);
							break;

						case kCheatSpecial_ForcePokeRegionRemoveRaw:
							force_poke_region_raw(subcheat, subcheat->data);
							subcheat->flags |= kSubcheatFlagDone;
							break;
					}
				}
				else
				{
					subcheat->frame_count--;
				}
			}

			done = 1;

			for (j = 0; j <= CheatTable[i].num_sub; j++)
			{
				if (!(CheatTable[i].subcheat[j].flags & kSubcheatFlagDone))
				{
					done = 0;

					break;
				}
			}

			/* disable cheat if 100% done */
			if (done)
			{
				cheat_set_status(i, 0);
			}
		} /* end for */
	}

	/* IPT_UI_TOGGLE_CHEAT Enable/Disable the active cheats on the fly. Required for some cheats. */
	if (input_ui_pressed(IPT_UI_TOGGLE_CHEAT))
	{
		/* Hold down shift to toggle the watchpoints */
#if 0	// test code
		if (ShiftKeyPressed())
		{
			is_watch_visible ^= 1;
			usrintf_showmessage_secs(1, "%s %s", ui_getstring(UI_watchpoints), (is_watch_visible ? ui_getstring (UI_on) : ui_getstring (UI_off)));
		}
		else
#endif	// test code
			if (ActiveCheatTotal)
		{
			CheatEnabled ^= 1;
			usrintf_showmessage_secs(1, "%s %s", ui_getstring(UI_cheats), (CheatEnabled ? ui_getstring (UI_on) : ui_getstring (UI_off)));
		}
	}

	cheatEngineWasActive = 0;
}

