/******************************************************************************

  palette.c

  Palette handling functions.

******************************************************************************/

#include "driver.h"


/* helper macro for 16-bit mode */
#define rgbpenindex(r,g,b) ((((r)>>3)<<10)+(((g)>>3)<<5)+((b)>>3))

unsigned short *game_pen;

int palette_start(void)
{
	game_pen      = MyMalloc((Machine->drv->total_colors * 2 + 16) * sizeof(UINT16));
	Machine->pens = MyMalloc((Machine->drv->total_colors * 2 + 16) * sizeof(UINT16));
#ifdef MAME_DEBUG
	Machine->debug_remapped_colortable = MyMalloc(2*DEBUGGER_TOTAL_COLORS*DEBUGGER_TOTAL_COLORS * sizeof(UINT16));
	Machine->debug_pens = MyMalloc(DEBUGGER_TOTAL_COLORS * sizeof(short));
#endif

#ifdef MAME_DEBUG
	if (!game_pen || !Machine->pens || !Machine->debug_remapped_colortable || !Machine->debug_pens)
#else
	if (!game_pen || !Machine->pens)
#endif
	{
		palette_stop();
		return 1;
	}

	return 0;
}


void palette_stop(void)
{
	MyFree(game_pen);
	MyFree(Machine->pens);
	Machine->pens = game_pen = 0;
#ifdef MAME_DEBUG
	MyFree(Machine->debug_remapped_colortable);
	MyFree(Machine->debug_pens);
	Machine->debug_pens = Machine->debug_remapped_colortable = 0;
#endif
}


int palette_init(void)
{
	int i;
#ifdef MAME_DEBUG
	UINT8 *debug_palette;
	UINT16 *debug_pens;

	if (mame_debug)
	{
		debug_palette = debugger_palette;
		debug_pens = Machine->debug_pens;
	}
	else
	{
		debug_palette = NULL;
		debug_pens = NULL;
	}

	if (osd_allocate_colors(debug_palette, debug_pens))
		return 1;
#else
	if (osd_allocate_colors())
		return 1;
#endif

	for (i = 0;i < Machine->drv->total_colors * 2 + 16; i++)
	{
		game_pen[i] = 0;
		Machine->pens[i] = i;
	}

	for (i = 0 ; i < COLOR_MAX_NUMBER ; i++)
	{
		Machine->system_color[i] = i + 0x2000;
	}

#ifdef MAME_DEBUG
	for (i = 0;i < DEBUGGER_TOTAL_COLORS*DEBUGGER_TOTAL_COLORS;i++)
	{
		Machine->debug_remapped_colortable[2*i+0] = Machine->debug_pens[i / DEBUGGER_TOTAL_COLORS];
		Machine->debug_remapped_colortable[2*i+1] = Machine->debug_pens[i % DEBUGGER_TOTAL_COLORS];
	}
#endif

	return 0;
}


void palette_change_color(int color, unsigned char red, unsigned char green, unsigned char blue)
{
	unsigned short pen = rgbpenindex(red, green, blue);

	if (game_pen[color] == pen) return;

	game_pen[color] = pen;

	osd_modify_pen(color, red, green, blue);
}
