/*********************************************************************

  usrintrf.c

  Functions used to handle MAME's crude user interface.

*********************************************************************/

#include "driver.h"
#include "info.h"
#include "datafile.h"
#include <stdarg.h>
#include <stdlib.h>
#include "ui_text.h"
#include "wav.h"
#include "drivers/neogeo.h"
#include <windows.h>
#include <time.h>
#include "win32/mame32.h"
#include "win32/avi.h"

#ifdef KAILLERA
#include "KailleraChat.h"
extern int kPlay;
#endif

/***************************************************************************
    macros
***************************************************************************/

#define BOXADJUST (Machine->uifontheight / 2)
#define ROWHEIGHT (Machine->uifontheight + 2)
#define ROWADJUST (Machine->uifontheight / 8)
#define MAX_PLAYLIST 512
#define MIN(a, b) ((a) < (b) ? (a) : (b))


/***************************************************************************
    external values
***************************************************************************/

#ifdef MAME_DEBUG
extern int mame_debug;
#endif

/* Variables for stat menu */
extern char build_version[];
extern unsigned int dispensed_tickets;
extern unsigned int coins[COIN_COUNTERS];
extern unsigned int coinlockedout[COIN_COUNTERS];

/* MARTINEZ.F 990207 Memory Card */
extern int mcd_action;
extern int mcd_number;
extern int memcard_status;
extern int memcard_number;
extern int memcard_manager;

extern int show_input;
extern char command_buf[];
extern int command_counter;

extern int saveslot;
extern int combosaveslot;
extern int record_combo;

/***************************************************************************
    external functions
***************************************************************************/

extern int neogeo_memcard_load(int);
extern void neogeo_memcard_save(void);
extern void neogeo_memcard_eject(void);
extern int neogeo_memcard_create(int);
/* MARTINEZ.F 990207 Memory Card End */

void NeoMVSDrawGfx16(struct osd_bitmap *bitmap,const struct GfxElement *gfx,
		unsigned int code,unsigned int color,int flipx,int flipy,int sx,int sy,
		int zx,int zy,const struct rectangle *clip);

void neojuke_send_message(UINT16 sound_code);

int count_object_number(void);
void alt_backcolor_exp(void);

/***************************************************************************
    internal values
***************************************************************************/

static int menu_lastselected = 0;
static int osd_lastselected = 0;
static int setup_selected = 0;
static int osd_selected = 0;
static int single_step = 0;
static int command_lastselected = 0;
static int playlist_selected = 0;
static int soundtest_selected = 1;

static int wave_counter = 0;
static int jukebox = 0;
static int line_volume[2] = { 0, 0 };
static int line_counter[2] = { 0, 0 };
static int volume_delay = 0;
static int random_play = 0;
static int playing_time = 0;
static int playing_wait = 0;
static int nosound_time = 0;
static int randomplay_time = 0;
static int fadeout_timing = 0;
#ifdef MAME_DEBUG
static int show_profiler = 0;
static int show_total_colors = 0;
#endif
static char messagetext[80];
static int messagecounter = 0;
static int scroll_reset = 0;
static int lhw, rhw, uaw, daw, law, raw;

static char *history_buf = 0;
static int history_scroll = 0;
static char *commandlist_buf = 0;
static int command_scroll = 0;

static int object = 0;
static int object_no = 0;
static int object_selected = 1;
static int object_cpu = 0;

static int command_sc = 0;
static int command_sel = 1;
static int command_load = 1;

static int cheat_sc = 0;
static int cheat_sel = 1;

static int show_time = 0;

enum {
	CODETYPE_BGM,
	CODETYPE_SE
};

enum {
	JUKE_STOP = 0,
	JUKE_STOPWAIT,
	JUKE_START,
	JUKE_STARTWAIT,
	JUKE_PLAYING,
	JUKE_PLAYWAIT,
	JUKE_RECORDING,
	JUKE_RECORDWAIT,
	JUKE_IDLE
};

static struct jukebox_t
{
	int num_bank;
	int curr_bank;
	int status;
	int command;
	int counter;
	int playlist;
	int playlist_bak;
	int menu_init;
	UINT16 start;
	UINT16 stop1;
	UINT16 stop2;
	UINT16 bank[8];
	UINT16 code[2];
	UINT16 que[8];
	UINT16 level[2];
} neojuke;

static struct playlist_t
{
	UINT16 code;
	char desc[48];
	int flag;
} playlist[MAX_PLAYLIST];

char *sound_filename = "sound.dat";

/***************************************************************************
    global value
***************************************************************************/

int paused = 0;
int in_jukebox = 0;
int wave_recording;
int wave_minutes;
int wave_seconds;
int skip_object[0x300 >> 1];
int front_layer = 0;
int alt_backcolor = 0;
int fadeout_frame = 600;
int fadeout_counter = 0;

/***************************************************************************
    prototype
***************************************************************************/

static int wordwrap_text_buffer(char *buffer, int maxcols, int fonttype);
static int count_lines_in_buffer (char *buffer);
static int display_scroll_message(struct osd_bitmap *bitmap, int *scroll,
                                  int width, int height, char *buf, int fonttype);

/**************************************************************************/

static int IsCloseButtonPressed(void)
{
	if (MAME32App.Done())
	{
		extern usres;
		usres |= 1;
		return usres;
	}
	return 0;
}

/**************************************************************************/

void set_ui_visarea (int xmin, int ymin, int xmax, int ymax)
{
	Machine->uiwidth  = xmax - xmin + 1;
	Machine->uiheight = ymax - ymin + 1;
	Machine->uixmin   = xmin;
	Machine->uiymin   = ymin;
}


static void erase_screen(struct osd_bitmap *bitmap)
{
	fillbitmap(bitmap, 0);
	schedule_full_refresh();
}


/***************************************************************************

  Display text on the screen. If erase is 0, it superimposes the text on
  the last frame displayed.

***************************************************************************/

void displaytext(struct osd_bitmap *bitmap, const struct DisplayText *dt, int fonttype)
{
	int next_sy = 0;
	int fontcolor;

	while (dt->text)
	{
		if (dt->color)	fontcolor = BUTTON_COLOR_YELLOW;
		else	fontcolor = FONT_COLOR_NORMAL;

		next_sy = uifont_drawfont(bitmap, dt->text, dt->x, dt->y + next_sy, UI_COLOR_NORMAL, fontcolor, TRANSPARENCY_FONT2, fonttype);
		dt++;

		if (dt->y + next_sy >= Machine->uiheight)
			break;
	}
}


/* Writes messages on the screen. */
static void ui_text_ex(struct osd_bitmap *bitmap,const char* buf_begin, const char* buf_end, int x, int y, int color)
{
	int len;
	char buf[256];

	memset(buf, 0, 256);

	len = (buf_end - buf_begin) + 1;
	if (*buf_end == ' ')
		len--;

	memcpy(buf, buf_begin, len);

	uifont_drawfont(bitmap, buf, x, y, color, FONT_COLOR_NORMAL, TRANSPARENCY_FONT, FONT_PROPORTIONAL);
}


/* Writes messages on the screen. */
void ui_text(struct osd_bitmap *bitmap,const char *buf,int x,int y)
{
	uifont_drawfont(bitmap, buf, x, y, UI_COLOR_NORMAL, FONT_COLOR_NORMAL, TRANSPARENCY_FONT2, FONT_PROPORTIONAL);
}


void ui_drawbox(struct osd_bitmap *bitmap, int leftx, int topy, int width, int height, int trans)
{
	unsigned short black,white;


	if (leftx < 0) leftx = 0;
	if (topy < 0) topy = 0;
	if (width > Machine->uiwidth) width = Machine->uiwidth;
	if (height > Machine->uiheight) height = Machine->uiheight;

	leftx += Machine->uixmin;
	topy += Machine->uiymin;

	black = Machine->uifont->colortable[0];
	white = Machine->uifont->colortable[1];

	plot_box(bitmap,leftx,        topy,         width,  1,       white, 0    );
	plot_box(bitmap,leftx,        topy+height-1,width,  1,       white, 0    );
	plot_box(bitmap,leftx,        topy,         1,      height,  white, 0    );
	plot_box(bitmap,leftx+width-1,topy,         1,      height,  white, 0    );
	plot_box(bitmap,leftx+1,      topy+1,       width-2,height-2,black, trans);
}


void ui_drawbox2(struct osd_bitmap *bitmap, int leftx, int topy, int width, int height, int color)
{
	if (leftx < 0) leftx = 0;
	if (topy < 0) topy = 0;
	if (width > Machine->uiwidth) width = Machine->uiwidth;
	if (height > Machine->uiheight) height = Machine->uiheight;

	leftx += Machine->uixmin;
	topy += Machine->uiymin;

	plot_box(bitmap,leftx, topy, width, height, color, 0);
}

#define drawbox0(x, y, w, h) \
	plot_box(bitmap, x,     y,     w,     1, SYSTEM_COLOR_FRAMEMIDIUM, 0); \
	plot_box(bitmap, x,     y + 1, 1, h - 1, SYSTEM_COLOR_FRAMEMIDIUM, 0); \
	plot_box(bitmap, x+w-1, y + 1, 1, h - 2, SYSTEM_COLOR_FRAMEMIDIUM, 0); \
	plot_box(bitmap, x,     y+h-1, w,     1, SYSTEM_COLOR_FRAMEMIDIUM, 0);

#define drawbox1(x, y, w, h) \
	plot_box(bitmap, x,     y,     w,     1, SYSTEM_COLOR_FRAMELIGHT, 0);  \
	plot_box(bitmap, x,     y + 1, 1, h - 1, SYSTEM_COLOR_FRAMELIGHT, 0);  \
	plot_box(bitmap, x+w-1, y + 1, 1, h - 2, SYSTEM_COLOR_FRAMEDARK,  0);   \
	plot_box(bitmap, x,     y+h-1, w,     1, SYSTEM_COLOR_FRAMEDARK,  0);

#define drawbox2(x, y, w, h) \
	plot_box(bitmap, x,     y,     w,     1,     SYSTEM_COLOR_FRAMEDARK,  0);   \
	plot_box(bitmap, x,     y + 1, 1,     h,     SYSTEM_COLOR_FRAMEDARK,  0);   \
	plot_box(bitmap, x+w-1, y,     1,     h - 1, SYSTEM_COLOR_FRAMELIGHT, 0);  \
	plot_box(bitmap, x+1,   y+h-1, w - 1, 1,     SYSTEM_COLOR_FRAMELIGHT, 0);


static void draw3dbox(struct osd_bitmap *bitmap, int x, int y, int w, int h, int t)
{
	ui_drawbox(bitmap,x-1, y-1, w+2, h+2, t);

	x += Machine->uixmin;
	y += Machine->uiymin;

	drawbox2(x-1, y-1, w+2, h+2);
	drawbox0(x-2, y-2, w+4, h+4);
	drawbox1(x-3, y-3, w+6, h+6);
}


static void drawbar(struct osd_bitmap *bitmap, int leftx, int topy,
                    int width, int height, int percentage, int default_percentage)
{
	unsigned short black,white;


	if (leftx < 0) leftx = 0;
	if (topy < 0) topy = 0;
	if (width > Machine->uiwidth) width = Machine->uiwidth;
	if (height > Machine->uiheight) height = Machine->uiheight;

	leftx += Machine->uixmin;
	topy += Machine->uiymin;

	black = Machine->uifont->colortable[0];
	white = Machine->uifont->colortable[1];

	plot_box(bitmap,leftx+(width-1)*default_percentage/100,topy,1,height/8,white,0);

	plot_box(bitmap,leftx,topy+height/8,width,1,white,0);

	plot_box(bitmap,leftx,topy+height/8,1+(width-1)*percentage/100,height-2*(height/8),white,0);

	plot_box(bitmap,leftx,topy+height-height/8-1,width,1,white,0);

	plot_box(bitmap,leftx+(width-1)*default_percentage/100,topy+height-height/8,1,height/8,white,0);
}

static void multiline_size(int *dx, int *dy, char *text, unsigned widthlimit)
{
	unsigned rows  = 0;
	unsigned width = 0;
	int n, n2, ct, code;
	int maxwidth = 0;

	n = 0;

	while (text[n])
	{
		ct = get_chartype(&text[n]);

		if (ct == ASCII || ct == HKANA)
		{
			if (text[n] == '\n')
			{
				rows++;
				width = 0;
				n++;
			}
			else if (text[n] == ' ')
			{
				int maxwidth2, width2;

				n++;
				n2 = n;
				maxwidth2 = maxwidth - width;
				width2 = 0;

				while (text[n2])
				{
					ct = get_chartype(&text[n2]);

					if (ct == ASCII || ct == HKANA)
					{
						if (text[n2] == '\n' || text[n2] == ' ')
							break;

						n2++;
						width2 += Machine->uifontwidth;
					}
					else
					{
						code = get_uifont_code(&text[n2]);
						width2 += get_fontwidth(code, FONT_PROPORTIONAL);
						n2 += 2;
					}
				}

				if (width + width2 > widthlimit)
				{
					text[n] = '\n';
				}
				else
				{
					width += width2 + Machine->uifontwidth;
					n = n2;
				}
			}
			else
			{
				n++;
				width += Machine->uifontwidth;
			}
		}
		else
		{
			code = get_uifont_code(&text[n]);

			n += 2;
			width += get_fontwidth(code, FONT_PROPORTIONAL);
		}

		if (width > maxwidth)
			maxwidth = width;
	}

	*dx = maxwidth;
	*dy = rows * (Machine->uifontheight+2) + Machine->uifontheight;
}


/* Compute the output size of a multiline string with box */
static void multilinebox_size(int *dx, int *dy, char *text, unsigned maxwidth)
{
	multiline_size(dx,dy,text,maxwidth);
	*dx += Machine->uifontwidth;
	*dy += Machine->uifontheight;
}


/* Display a multiline string with box */
static void ui_multitextbox_ex(struct osd_bitmap *bitmap,const char* begin, const char* end, unsigned max, int x, int y, int dx, int dy, int color)
{
	x -= Machine->uifontwidth / 2;
	draw3dbox(bitmap, x, y, dx, dy, 0);
	x += Machine->uifontwidth / 2;
	y += Machine->uifontheight / 2;
	ui_text_ex(bitmap, begin, end, x, y, color);
}


void ui_displaymenu(struct osd_bitmap *bitmap,const char **items,const char **subitems,char *flag,int selected,int arrowize_subitem)
{
	struct DisplayText dt[256];
	int curr_dt,curr_y;
	const char *lefthilight = ui_getstring (UI_lefthilight);
	const char *righthilight = ui_getstring (UI_righthilight);
	const char *uparrow = ui_getstring (UI_uparrow);
	const char *downarrow = ui_getstring (UI_downarrow);
	const char *leftarrow = ui_getstring (UI_leftarrow);
	const char *rightarrow = ui_getstring (UI_rightarrow);
	int i,count,len,maxlen,highlen,maxlen2,len2;
	int leftoffs,topoffs,visible,topitem;
	int selected_long;


	i = 0;
	maxlen = maxlen2 = 0;
	highlen = Machine->uiwidth / Machine->uifontwidth;
	while (items[i])
	{
		len = strwidth(items[i], FONT_PROPORTIONAL);
		if (subitems && subitems[i])
		{
			len2 = 2 * Machine->uifontwidth + strwidth(subitems[i], FONT_PROPORTIONAL);
			if (len2 > maxlen2) maxlen2 = len2;
		}
		if (len > maxlen)
			maxlen = len;
		i++;
	}
	maxlen += maxlen2;
	count = i;

	maxlen += 4 * Machine->uifontwidth;
	if (maxlen > Machine->uiwidth - 2*Machine->uifontwidth)
		maxlen = Machine->uiwidth - 2*Machine->uifontwidth;

	visible = (Machine->uiheight / ROWHEIGHT) - 1;

	if (neojuke.playlist && neojuke.status != JUKE_STOP)	visible -= 4;
	
	topitem = 0;
	if (visible > count) visible = count;
	else
	{
		topitem = selected - visible / 2;
		if (topitem < 0) topitem = 0;
		if (topitem > count - visible) topitem = count - visible;
	}

	leftoffs = (Machine->uiwidth - maxlen) / 2;
	if (neojuke.playlist && neojuke.status != JUKE_STOP)	topoffs  = (Machine->uiheight - 15 * ROWHEIGHT) / 2 + 5;
	else	topoffs  = (Machine->uiheight - visible * ROWHEIGHT) / 2 + 1;

	/* black background */
	if (neojuke.playlist && neojuke.status != JUKE_STOP)	draw3dbox(bitmap,leftoffs,topoffs - BOXADJUST / 2 - 1,maxlen,visible * ROWHEIGHT + BOXADJUST,1);
	else	draw3dbox(bitmap,leftoffs,topoffs - BOXADJUST / 2 - 1,maxlen,visible * ROWHEIGHT + BOXADJUST,1);

	selected_long = 0;
	curr_dt = 0;
	for (i = 0;i < visible;i++)
	{
		int item = i + topitem;

		curr_y = topoffs + i * ROWHEIGHT + ROWADJUST;

		if (i == 0 && item > 0)
		{
			dt[curr_dt].text = uparrow;
			dt[curr_dt].color = UI_COLOR_NORMAL;
			dt[curr_dt].x = (Machine->uiwidth - uaw) / 2;
			dt[curr_dt].y = curr_y;
			curr_dt++;
		}
		else if (i == visible - 1 && item < count - 1)
		{
			dt[curr_dt].text = downarrow;
			dt[curr_dt].color = UI_COLOR_NORMAL;
			dt[curr_dt].x = (Machine->uiwidth - daw) / 2;
			dt[curr_dt].y = curr_y;
			curr_dt++;
		}
		else
		{
			if (subitems && subitems[item])
			{
				int sublen;

				len = strwidth(items[item], FONT_PROPORTIONAL);
				dt[curr_dt].text = items[item];
				dt[curr_dt].color = UI_COLOR_NORMAL;
				dt[curr_dt].x = leftoffs + 3 * Machine->uifontwidth / 2;
				dt[curr_dt].y = curr_y;
				curr_dt++;

				sublen = strwidth(subitems[item], FONT_PROPORTIONAL);
				if (sublen > maxlen-5*Machine->uifontwidth-len)
				{
					dt[curr_dt].text = "...";
					sublen = strwidth(dt[curr_dt].text, FONT_PROPORTIONAL);
					if (item == selected)
						selected_long = 1;
				}
				else
					dt[curr_dt].text = subitems[item];

				/* If this item is flagged, draw it in inverse print */
				dt[curr_dt].color = (flag && flag[item]) ? UI_COLOR_INVERSE : UI_COLOR_NORMAL;
				dt[curr_dt].x = leftoffs + maxlen - Machine->uifontwidth - strwidth(dt[curr_dt].text, FONT_PROPORTIONAL) - Machine->uifontwidth/2;
				dt[curr_dt].y = curr_y;
				curr_dt++;
			}
			else
			{
				dt[curr_dt].text = items[item];
				dt[curr_dt].color = UI_COLOR_NORMAL;
				dt[curr_dt].x = (Machine->uiwidth - strwidth(items[item], FONT_PROPORTIONAL)) / 2;
				dt[curr_dt].y = curr_y;
				curr_dt++;
			}
		}
	}

	i = selected - topitem;

	curr_y = topoffs + i * ROWHEIGHT + ROWADJUST;

	if (subitems && subitems[selected] && arrowize_subitem)
	{
		if (arrowize_subitem & 1)
		{
			int sublen;

			len = strwidth(items[selected], FONT_PROPORTIONAL);

			dt[curr_dt].text = leftarrow;
			dt[curr_dt].color = UI_COLOR_NORMAL;

			sublen = strwidth(subitems[selected], FONT_PROPORTIONAL);
			if (sublen > maxlen-5*Machine->uifontwidth-len)
				sublen = strwidth("...", FONT_PROPORTIONAL);

			dt[curr_dt].x = leftoffs + maxlen-2*Machine->uifontwidth - strwidth(subitems[selected], FONT_PROPORTIONAL) - Machine->uifontwidth/2 - 1;
			dt[curr_dt].y = curr_y;
			curr_dt++;
		}
		if (arrowize_subitem & 2)
		{
			dt[curr_dt].text = rightarrow;
			dt[curr_dt].color = UI_COLOR_NORMAL;
			dt[curr_dt].x = leftoffs + maxlen-Machine->uifontwidth - Machine->uifontwidth/2;
			dt[curr_dt].y = curr_y;
			curr_dt++;
		}
	}
	else
	{
		dt[curr_dt].text = righthilight;
		dt[curr_dt].color = UI_COLOR_NORMAL;
		dt[curr_dt].x = leftoffs + maxlen-Machine->uifontwidth - Machine->uifontwidth/2;
		dt[curr_dt].y = curr_y;
		curr_dt++;
	}
	dt[curr_dt].text = lefthilight;
	dt[curr_dt].color = UI_COLOR_NORMAL;
	dt[curr_dt].x = leftoffs + Machine->uifontwidth/2;
	dt[curr_dt].y = curr_y;
	curr_dt++;

	dt[curr_dt].text = 0;	/* terminate array */

	plot_box(bitmap, leftoffs + 8, curr_y + ROWHEIGHT + 1, maxlen, ROWHEIGHT, CURSOR_COLOR, 0);
	
	displaytext(bitmap,dt,FONT_PROPORTIONAL);

	if (selected_long)
	{
		int long_dx;
		int long_dy;
		int long_x;
		int long_y;
		unsigned long_max;
		char textbuf[512];

		memset(textbuf, 0, 512);
		strcpy(textbuf, subitems[selected]);

		long_max = (Machine->uiwidth / Machine->uifontwidth) - 4;//2;

		multilinebox_size(&long_dx,&long_dy,textbuf,long_max*Machine->uifontwidth);

		long_x = Machine->uiwidth - long_dx;
		long_y = topoffs + (i+1) * ROWHEIGHT;

		/* if too low display up */
		if (long_y + long_dy > Machine->uiheight)
			long_y = topoffs + i * ROWHEIGHT - long_dy;

		ui_multitextbox_ex(bitmap, textbuf, textbuf + strwidth(subitems[selected], FONT_PROPORTIONAL), long_max,long_x, long_y, long_dx, long_dy, UI_COLOR_NORMAL);
	}
}


int ui_displaymessagewindow(struct osd_bitmap *bitmap, const char *text)
{
	struct DisplayText dt[256];
	char *c, *c2;
	char textcopy[2048];
	int curr_dt;
	int i, widthlimit, lines;
	int leftoffs, topoffs;
	int maxrows, maxcols, maxwidth;
#ifndef BUILD_ENGLISH
	int ct, code;
#endif
	int width;
	static int scroll = 0;

	maxcols = (Machine->uiwidth / Machine->uifontwidth) - 2;
	maxrows = (Machine->uiheight / ROWHEIGHT) - 1;
	widthlimit = Machine->uiwidth - 4 * Machine->uifontwidth;

	/* copy text, calculate max len, count lines, wrap long lines and crop height to fit */
	maxwidth = 0;
	lines = 0;
	c = (char *)text;
	c2 = textcopy;

#ifndef BUILD_ENGLISH
	{
		while (*c)
		{
			width = 0;

			while (*c && *c != '\n')
			{
				ct = get_chartype(c);

				if (ct == ASCII || ct == HKANA)
					i = Machine->uifontwidth;
				else
				{
					code = get_uifont_code(c);
					i = get_fontwidth(code, FONT_PROPORTIONAL);
				}

				if (width + i >= widthlimit)
				{
					*c2++ = '\n'; /* insert wrap */
					break;
				}

				width += i;

				*c2++ = *c++;
				if (ct == ZKANJI)
					*c2++ = *c++;
			}

			if (*c == '\n')
				*c2++ = *c++;

			if (width > maxwidth) maxwidth = width;

			lines++;

			if (lines == maxrows)
				break;
		}
	}
#else
	{
		while (*c)
		{
			width = 0;

			while (*c && *c != '\n')
			{
				*c2++ = *c++;
				width += Machine->uifontwidth;
				if (width >= widthlimit && *c != '\n')
				{
					/* attempt word wrap */
					char *csave = c, *c2save = c2;
					int widthsave = width;

					/* back up to last space or beginning of line */
					while (*c != ' ' && *c != '\n' && c > text)
						--c, --c2, width -= Machine->uifontwidth;

					/* if no space was found, hard wrap instead */
					if (*c != ' ')
						c = csave, c2 = c2save, width = widthsave;
					else
						c++;

					*c2++ = '\n'; /* insert wrap */
					break;
				}
			}

			if (*c == '\n')
				*c2++ = *c++;

			if (width > maxwidth) maxwidth = width;

			lines++;
			if (lines == maxrows)
				break;
		}
	}
#endif
	*c2 = '\0';

	maxwidth += Machine->uifontwidth;

	leftoffs = (Machine->uiwidth - maxwidth) / 2;
	if (leftoffs < 0) leftoffs = 0;

	topoffs = (Machine->uiheight - lines * (Machine->uifontheight + 2)) / 2 + 1;

	draw3dbox(bitmap,leftoffs, topoffs - 4, maxwidth, lines * (Machine->uifontheight + 2) + 7, 1);

	curr_dt = 0;
	c = textcopy;
	i = 0;
	while (*c)
	{
		c2 = c;
		while (*c && *c != '\n')
			c++;

		if (*c == '\n')
		{
			*c = '\0';
			c++;
		}

		if (*c2 == '\t')	/* center text */
		{
			c2++;
			dt[curr_dt].x = (Machine->uiwidth - strwidth(c2, FONT_PROPORTIONAL)) / 2;
		}
		else
			dt[curr_dt].x = leftoffs + Machine->uifontwidth/2;

		dt[curr_dt].text = c2;
		dt[curr_dt].color = UI_COLOR_NORMAL;
		dt[curr_dt].y = topoffs + i * (Machine->uifontheight + 2) + 2;
		curr_dt++;

		i++;
	}

	dt[curr_dt].text = 0;	/* terminate array */

	displaytext(bitmap, dt, FONT_PROPORTIONAL);

	return 0;
}


static void showcharset(struct osd_bitmap *bitmap)
{
	int i;
	char buf[80];
	int bank,color,firstdrawn;
	int palpage;
	int changed;
	int quit;
	extern int no_of_tiles;

	bank = -1;
	color = 0;
	firstdrawn = 0;
	palpage = 0;

	changed = 1;
	quit = 0;

	do
	{
		int cpx,cpy,skip_chars;

		if (bank >= 0)
		{
			cpx = Machine->uiwidth / Machine->gfx[bank]->width;
			cpy = (Machine->uiheight - Machine->uifontheight) / Machine->gfx[bank]->height;
			skip_chars = cpx * cpy;
		}
		else cpx = cpy = skip_chars = 0;

		if (changed)
		{
			int lastdrawn=0;

			erase_screen(bitmap);

			/* validity chack after char bank change */
			if (bank >= 0)
			{
				if (firstdrawn >= Machine->gfx[bank]->total_elements)
				{
					firstdrawn = Machine->gfx[bank]->total_elements - skip_chars;
					if (firstdrawn < 0) firstdrawn = 0;
				}
			}

			if (bank != 2)
			{
				if (bank >= 0)
				{
					for (i = 0; i+firstdrawn < Machine->gfx[bank]->total_elements && i<cpx*cpy; i++)
					{
						drawgfx(bitmap,Machine->gfx[bank],
								i+firstdrawn,color,  /*sprite num, color*/
								0, 0,
								(i % cpx) * Machine->gfx[bank]->width + Machine->uixmin,
								Machine->uifontheight + (i / cpx) * Machine->gfx[bank]->height + Machine->uiymin,
								0,TRANSPARENCY_NONE,0);

						lastdrawn = i+firstdrawn;
					}
				}
				else
				{
					int sx, sy, colors;

					colors = Machine->drv->total_colors - 256 * palpage;
					if (colors > 256) colors = 256;

					for (i = 0; i < 16; i++)
					{
						char bf[40];

						sx = 3 * Machine->uifontwidth + 2 * Machine->uifontwidth * (i % 16) - Machine->uifontwidth/2;
						sprintf(bf, "%2X", i);
						ui_text(bitmap,bf,sx,Machine->uifontheight);

						if (16 * i < colors)
						{
							sy = 2 * Machine->uifontheight + Machine->uifontheight * (i % 16);
							sprintf(bf," %02X",i+16*palpage);
							ui_text(bitmap,bf,0,sy);
						}
					}

					for (i = 0; i < colors; i++)
					{
						sx = Machine->uixmin + 3 * Machine->uifontwidth + (Machine->uifontwidth * 2)*(i % 16);
						sy = Machine->uiymin + 2 * Machine->uifontheight + Machine->uifontheight * (i / 16);
						plot_box(bitmap,sx,sy,Machine->uifontwidth*2-1,Machine->uifontheight-1,Machine->pens[i + 256*palpage],0);
					}
				}
			}
			else	/* neogeo sprite tiles */
			{
				struct rectangle clip;

				clip.min_x = Machine->uixmin;
				clip.max_x = Machine->uixmin + Machine->uiwidth - 1;
				clip.min_y = Machine->uiymin;
				clip.max_y = Machine->uiymin + Machine->uiheight - 1;

				for (i = 0; i+firstdrawn < no_of_tiles && i<cpx*cpy; i++)
				{
					NeoMVSDrawGfx16(bitmap, Machine->gfx[bank],
						i + firstdrawn,color,  /*sprite num, color*/
						0,0,
						(i % cpx) * Machine->gfx[bank]->width + Machine->uixmin,
						Machine->uifontheight+1 + (i / cpx) * Machine->gfx[bank]->height + Machine->uiymin,
						16,16,&clip);

					lastdrawn = i+firstdrawn;
				}
			}

			if (bank >= 0)
				sprintf(buf,"GFXSET %d COLOR %2X CODE %X-%X",bank,color,firstdrawn,lastdrawn);
			else
				strcpy(buf,"PALETTE");
			ui_text(bitmap,buf,0,0);

			changed = 0;
		}

		update_video_and_audio();

		quit = input_ui_pressed(IPT_UI_SHOW_GFX) | input_ui_pressed(IPT_UI_CANCEL) | IsCloseButtonPressed();

		if (code_pressed(KEYCODE_LCONTROL) || code_pressed(KEYCODE_RCONTROL))
		{
			skip_chars = cpx;
		}
		if (code_pressed(KEYCODE_LSHIFT) || code_pressed(KEYCODE_RSHIFT))
		{
			skip_chars = 1;
		}


		if (input_ui_pressed_repeat(IPT_UI_RIGHT,8))
		{
			if (bank+1 < MAX_GFX_ELEMENTS && Machine->gfx[bank + 1])
			{
				bank++;
//				firstdrawn = 0;
				changed = 1;
			}
		}

		if (input_ui_pressed_repeat(IPT_UI_LEFT,8))
		{
			if (bank > -1)
			{
				bank--;
//				firstdrawn = 0;
				changed = 1;
			}
		}

		if (code_pressed_memory_repeat(KEYCODE_PGDN,4))
		{
			if (bank >= 0)
			{
				if (firstdrawn + skip_chars < Machine->gfx[bank]->total_elements)
				{
					firstdrawn += skip_chars;
					changed = 1;
				}
			}
			else
			{
				if (256 * (palpage + 1) < Machine->drv->total_colors)
				{
					palpage++;
					changed = 1;
				}
			}
		}

		if (code_pressed_memory_repeat(KEYCODE_PGUP,4))
		{
			if (bank >= 0)
			{
				firstdrawn -= skip_chars;
				if (firstdrawn < 0) firstdrawn = 0;
				changed = 1;
			}
			else
			{
				if (palpage > 0)
				{
					palpage--;
					changed = 1;
				}
			}
		}

		if (input_ui_pressed_repeat(IPT_UI_UP,4))
		{
			if (bank >= 0)
			{
				if (color < Machine->gfx[bank]->total_colors - 1)
				{
					color++;
					changed = 1;
				}
			}
		}

		if (input_ui_pressed_repeat(IPT_UI_DOWN,4))
		{
			if (color > 0)
			{
				color--;
				changed = 1;
			}
		}

		if (input_ui_pressed(IPT_UI_SNAPSHOT))
			osd_save_snapshot(bitmap);

	} while (!quit);

	schedule_full_refresh();
}


static int setdipswitches(struct osd_bitmap *bitmap,int selected)
{
	const char *menu_item[128];
	const char *menu_subitem[128];
	struct InputPort *entry[128];
	char flag[40];
	int i,sel;
	struct InputPort *in;
	int total;
	int arrowize;


	sel = selected - 1;


	in = Machine->input_ports;

	total = 0;
	while (in->type != IPT_END)
	{
		if ((in->type & ~IPF_MASK) == IPT_DIPSWITCH_NAME && input_port_name(in) != 0 &&
				(in->type & IPF_UNUSED) == 0 &&
				!(!options.cheat && (in->type & IPF_CHEAT)))
		{
			entry[total] = in;
			menu_item[total] = input_port_name(in);

			total++;
		}

		in++;
	}

	if (total == 0) return 0;

	menu_item[total] = ui_getstring (UI_returntomain);
	menu_item[total + 1] = 0;	/* terminate array */
	total++;


	for (i = 0;i < total;i++)
	{
		flag[i] = 0; /* TODO: flag the dip if it's not the real default */
		if (i < total - 1)
		{
			in = entry[i] + 1;
			while ((in->type & ~IPF_MASK) == IPT_DIPSWITCH_SETTING &&
					in->default_value != entry[i]->default_value)
				in++;

			if ((in->type & ~IPF_MASK) != IPT_DIPSWITCH_SETTING)
				menu_subitem[i] = ui_getstring (UI_INVALID);
			else menu_subitem[i] = input_port_name(in);
		}
		else menu_subitem[i] = 0;	/* no subitem */
	}

	arrowize = 0;
	if (sel < total - 1)
	{
		in = entry[sel] + 1;
		while ((in->type & ~IPF_MASK) == IPT_DIPSWITCH_SETTING &&
				in->default_value != entry[sel]->default_value)
			in++;

		if ((in->type & ~IPF_MASK) != IPT_DIPSWITCH_SETTING)
			/* invalid setting: revert to a valid one */
			arrowize |= 1;
		else
		{
			if (((in-1)->type & ~IPF_MASK) == IPT_DIPSWITCH_SETTING &&
					!(!options.cheat && ((in-1)->type & IPF_CHEAT)))
				arrowize |= 1;
		}
	}
	if (sel < total - 1)
	{
		in = entry[sel] + 1;
		while ((in->type & ~IPF_MASK) == IPT_DIPSWITCH_SETTING &&
				in->default_value != entry[sel]->default_value)
			in++;

		if ((in->type & ~IPF_MASK) != IPT_DIPSWITCH_SETTING)
			/* invalid setting: revert to a valid one */
			arrowize |= 2;
		else
		{
			if (((in+1)->type & ~IPF_MASK) == IPT_DIPSWITCH_SETTING &&
					!(!options.cheat && ((in+1)->type & IPF_CHEAT)))
				arrowize |= 2;
		}
	}

	ui_displaymenu(bitmap,menu_item,menu_subitem,flag,sel,arrowize);

	if (input_ui_pressed_repeat(IPT_UI_DOWN,4))
		sel = (sel + 1) % total;

	if (input_ui_pressed_repeat(IPT_UI_UP,4))
		sel = (sel + total - 1) % total;

	if (input_ui_pressed_repeat(IPT_UI_RIGHT,8))
	{
		if (sel < total - 1)
		{
			in = entry[sel] + 1;
			while ((in->type & ~IPF_MASK) == IPT_DIPSWITCH_SETTING &&
					in->default_value != entry[sel]->default_value)
				in++;

			if ((in->type & ~IPF_MASK) != IPT_DIPSWITCH_SETTING)
				/* invalid setting: revert to a valid one */
				entry[sel]->default_value = (entry[sel]+1)->default_value & entry[sel]->mask;
			else
			{
				if (((in+1)->type & ~IPF_MASK) == IPT_DIPSWITCH_SETTING &&
						!(!options.cheat && ((in+1)->type & IPF_CHEAT)))
					entry[sel]->default_value = (in+1)->default_value & entry[sel]->mask;
			}

			schedule_full_refresh();
		}
	}

	if (input_ui_pressed_repeat(IPT_UI_LEFT,8))
	{
		if (sel < total - 1)
		{
			in = entry[sel] + 1;
			while ((in->type & ~IPF_MASK) == IPT_DIPSWITCH_SETTING &&
					in->default_value != entry[sel]->default_value)
				in++;

			if ((in->type & ~IPF_MASK) != IPT_DIPSWITCH_SETTING)
				/* invalid setting: revert to a valid one */
				entry[sel]->default_value = (entry[sel]+1)->default_value & entry[sel]->mask;
			else
			{
				if (((in-1)->type & ~IPF_MASK) == IPT_DIPSWITCH_SETTING &&
						!(!options.cheat && ((in-1)->type & IPF_CHEAT)))
					entry[sel]->default_value = (in-1)->default_value & entry[sel]->mask;
			}

			schedule_full_refresh();
		}
	}

	if (input_ui_pressed(IPT_UI_SELECT))
	{
		if (sel == total - 1) sel = -1;
	}

	if (input_ui_pressed(IPT_UI_CANCEL))
		sel = -1;

	if (input_ui_pressed(IPT_UI_CONFIGURE))
		sel = -2;

	if (sel == -1 || sel == -2)
	{
		schedule_full_refresh();
	}

	return sel + 1;
}

/* This flag is used for record OR sequence of key/joy */
/* when is !=0 the first sequence is record, otherwise the first free */
/* it's used byt setdefkeysettings, setdefjoysettings, setkeysettings, setjoysettings */
static int record_first_insert = 1;

static char menu_subitem_buffer[400][96];

static int setdefcodesettings(struct osd_bitmap *bitmap,int selected)
{
	const char *menu_item[400];
	const char *menu_subitem[400];
	struct ipd *entry[400];
	char flag[400];
	int i, sel, visible;
	struct ipd *in;
	int total;
	extern struct ipd inputport_defaults[];

	sel = selected - 1;
	visible = (Machine->uiheight / ROWHEIGHT) - 3;

	if (Machine->input_ports == 0)
		return 0;

	in = inputport_defaults;

	total = 0;
	while (in->type != IPT_END)
	{
		if (in->name != 0  && (in->type & ~IPF_MASK) != IPT_UNKNOWN && (in->type & IPF_UNUSED) == 0
			&& !(!options.cheat && (in->type & IPF_CHEAT)))
		{
			entry[total] = in;
			menu_item[total] = in->name;

			total++;
		}

		in++;
	}

	if (total == 0) return 0;

	menu_item[total] = ui_getstring (UI_returntomain);
	menu_item[total + 1] = 0;	/* terminate array */
	total++;

	for (i = 0;i < total;i++)
	{
		if (i < total - 1)
		{
			seq_name(&entry[i]->seq,menu_subitem_buffer[i],sizeof(menu_subitem_buffer[0]));
			menu_subitem[i] = menu_subitem_buffer[i];
		} else
			menu_subitem[i] = 0;	/* no subitem */
		flag[i] = 0;
	}

	if (sel > SEL_MASK)   /* are we waiting for a new key? */
	{
		int ret;

		menu_subitem[sel & SEL_MASK] = "    ";
		ui_displaymenu(bitmap,menu_item,menu_subitem,flag,sel & SEL_MASK,3);

		ret = seq_read_async(&entry[sel & SEL_MASK]->seq,record_first_insert);

		if (ret >= 0)
		{
			sel &= 0xff;

			if (ret > 0 || seq_get_1(&entry[sel]->seq) == CODE_NONE)
			{
				seq_set_1(&entry[sel]->seq,CODE_NONE);
				ret = 1;
			}

			schedule_full_refresh();

			record_first_insert = ret != 0;
		}


		return sel + 1;
	}


	ui_displaymenu(bitmap,menu_item,menu_subitem,flag,sel,0);

	if (input_ui_pressed_repeat(IPT_UI_DOWN,4))
	{
		sel = (sel + 1) % total;
		record_first_insert = 1;
	}

	if (input_ui_pressed_repeat(IPT_UI_UP,4))
	{
		sel = (sel + total - 1) % total;
		record_first_insert = 1;
	}

	if (input_ui_pressed_repeat(IPT_UI_LEFT, 4))
	{
		sel -= visible;
		if (sel < 0)	sel = 0;
		record_first_insert = 1;
	}

	if (input_ui_pressed_repeat(IPT_UI_RIGHT, 4))
	{
		sel += visible;
		if (sel >= total)	sel = total - 1;
		record_first_insert = 1;
	}

	if (input_ui_pressed(IPT_UI_SELECT))
	{
		if (sel == total - 1) sel = -1;
		else
		{
			seq_read_async_start();

			sel |= 1 << SEL_BITS;	/* we'll ask for a key */

			schedule_full_refresh();
		}
	}

	if (input_ui_pressed(IPT_UI_CANCEL))
		sel = -1;

	if (input_ui_pressed(IPT_UI_CONFIGURE))
		sel = -2;

	if (sel == -1 || sel == -2)
	{
		schedule_full_refresh();

		record_first_insert = 1;
	}

	return sel + 1;
}



static int setcodesettings(struct osd_bitmap *bitmap,int selected)
{
	const char *menu_item[400];
	const char *menu_subitem[400];
	struct InputPort *entry[400];
	char flag[400];
	int i, sel, visible;
	struct InputPort *in;
	int total;


	sel = selected - 1;
	visible = (Machine->uiheight / ROWHEIGHT) - 3;

	if (Machine->input_ports == 0)
		return 0;

	in = Machine->input_ports;

	total = 0;
	while (in->type != IPT_END)
	{
		if (input_port_name(in) != 0 && seq_get_1(&in->seq) != CODE_NONE && (in->type & ~IPF_MASK) != IPT_UNKNOWN)
		{
			entry[total] = in;
			menu_item[total] = input_port_name(in);

			total++;
		}

		in++;
	}

	if (total == 0) return 0;

	menu_item[total] = ui_getstring (UI_returntomain);
	menu_item[total + 1] = 0;	/* terminate array */
	total++;

	for (i = 0;i < total;i++)
	{
		if (i < total - 1)
		{
			seq_name(input_port_seq(entry[i]),menu_subitem_buffer[i],sizeof(menu_subitem_buffer[0]));
			menu_subitem[i] = menu_subitem_buffer[i];

			/* If the key isn't the default, flag it */
			if (seq_get_1(&entry[i]->seq) != CODE_DEFAULT)
				flag[i] = 1;
			else
				flag[i] = 0;

		} else
			menu_subitem[i] = 0;	/* no subitem */
	}

	if (sel > SEL_MASK)   /* are we waiting for a new key? */
	{
		int ret;

		menu_subitem[sel & SEL_MASK] = "    ";
		ui_displaymenu(bitmap,menu_item,menu_subitem,flag,sel & SEL_MASK,3);

		ret = seq_read_async(&entry[sel & SEL_MASK]->seq,record_first_insert);

		if (ret >= 0)
		{
			sel &= 0xff;

			if (ret > 0 || seq_get_1(&entry[sel]->seq) == CODE_NONE)
			{
				seq_set_1(&entry[sel]->seq, CODE_DEFAULT);
				ret = 1;
			}

			schedule_full_refresh();

			record_first_insert = ret != 0;
		}

		return sel + 1;
	}


	ui_displaymenu(bitmap,menu_item,menu_subitem,flag,sel,0);

	if (input_ui_pressed_repeat(IPT_UI_DOWN,4))
	{
		sel = (sel + 1) % total;
		record_first_insert = 1;
	}

	if (input_ui_pressed_repeat(IPT_UI_UP,4))
	{
		sel = (sel + total - 1) % total;
		record_first_insert = 1;
	}

	if (input_ui_pressed_repeat(IPT_UI_LEFT, 4))
	{
		sel -= visible;
		if (sel < 0)	sel = 0;
		record_first_insert = 1;
	}

	if (input_ui_pressed_repeat(IPT_UI_RIGHT, 4))
	{
		sel += visible;
		if (sel >= total)	sel = total - 1;
		record_first_insert = 1;
	}

	if (input_ui_pressed(IPT_UI_SELECT))
	{
		if (sel == total - 1) sel = -1;
		else
		{
			seq_read_async_start();

			sel |= 1 << SEL_BITS;	/* we'll ask for a key */

			schedule_full_refresh();
		}
	}

	if (input_ui_pressed(IPT_UI_CANCEL))
		sel = -1;

	if (input_ui_pressed(IPT_UI_CONFIGURE))
		sel = -2;

	if (sel == -1 || sel == -2)
	{
		schedule_full_refresh();

		record_first_insert = 1;
	}

	return sel + 1;
}


static int calibratejoysticks(struct osd_bitmap *bitmap,int selected)
{
	const char *msg;
	char buf[2048];
	int sel;
	static int calibration_started = 0;

	sel = selected - 1;

	if (calibration_started == 0)
	{
		osd_joystick_start_calibration();
		calibration_started = 1;
		strcpy (buf, "");
	}

	if (sel > SEL_MASK) /* Waiting for the user to acknowledge joystick movement */
	{
		if (input_ui_pressed(IPT_UI_CANCEL))
		{
			calibration_started = 0;
			sel = -1;
		}
		else if (input_ui_pressed(IPT_UI_SELECT))
		{
			osd_joystick_calibrate();
			sel &= 0xff;
		}

		ui_displaymessagewindow(bitmap,buf);
	}
	else
	{
		msg = osd_joystick_calibrate_next();
		schedule_full_refresh();
		if (msg == 0)
		{
			calibration_started = 0;
			osd_joystick_end_calibration();
			sel = -1;
		}
		else
		{
			strcpy (buf, msg);
			ui_displaymessagewindow(bitmap,buf);
			sel |= 1 << SEL_BITS;
		}
	}

	if (input_ui_pressed(IPT_UI_CONFIGURE))
		sel = -2;

	if (sel == -1 || sel == -2)
	{
		schedule_full_refresh();
	}

	return sel + 1;
}


static int settraksettings(struct osd_bitmap *bitmap,int selected)
{
	const char *menu_item[40];
	const char *menu_subitem[40];
	struct InputPort *entry[40];
	int i,sel;
	struct InputPort *in;
	int total,total2;
	int arrowize;


	sel = selected - 1;


	if (Machine->input_ports == 0)
		return 0;

	in = Machine->input_ports;

	/* Count the total number of analog controls */
	total = 0;
	while (in->type != IPT_END)
	{
		if (((in->type & 0xff) > IPT_ANALOG_START) && ((in->type & 0xff) < IPT_ANALOG_END)
				&& !(!options.cheat && (in->type & IPF_CHEAT)))
		{
			entry[total] = in;
			total++;
		}
		in++;
	}

	if (total == 0) return 0;

	/* Each analog control has 3 entries - key & joy delta, reverse, sensitivity */

#define ENTRIES 3

	total2 = total * ENTRIES;

	menu_item[total2] = ui_getstring (UI_returntomain);
	menu_item[total2 + 1] = 0;	/* terminate array */
	total2++;

	arrowize = 0;
	for (i = 0;i < total2;i++)
	{
		if (i < total2 - 1)
		{
			char label[30][40];
			char setting[30][40];
			int sensitivity,delta;
			int reverse;

			strcpy (label[i], input_port_name(entry[i/ENTRIES]));
			sensitivity = IP_GET_SENSITIVITY(entry[i/ENTRIES]);
			delta = IP_GET_DELTA(entry[i/ENTRIES]);
			reverse = (entry[i/ENTRIES]->type & IPF_REVERSE);

			strcat (label[i], " ");
			switch (i%ENTRIES)
			{
				case 0:
					strcat (label[i], ui_getstring (UI_keyjoyspeed));
					sprintf(setting[i],"%d",delta);
					if (i == sel) arrowize = 3;
					break;
				case 1:
					strcat (label[i], ui_getstring (UI_reverse));
					if (reverse)
						sprintf(setting[i],ui_getstring (UI_on));
					else
						sprintf(setting[i],ui_getstring (UI_off));
					if (i == sel) arrowize = 3;
					break;
				case 2:
					strcat (label[i], ui_getstring (UI_sensitivity));
					sprintf(setting[i],"%3d%%",sensitivity);
					if (i == sel) arrowize = 3;
					break;
			}

			menu_item[i] = label[i];
			menu_subitem[i] = setting[i];

			in++;
		}
		else menu_subitem[i] = 0;	/* no subitem */
	}

	ui_displaymenu(bitmap,menu_item,menu_subitem,0,sel,arrowize);

	if (input_ui_pressed_repeat(IPT_UI_DOWN,4))
		sel = (sel + 1) % total2;

	if (input_ui_pressed_repeat(IPT_UI_UP,4))
		sel = (sel + total2 - 1) % total2;

	if (input_ui_pressed_repeat(IPT_UI_LEFT,8))
	{
		if ((sel % ENTRIES) == 0)
		/* keyboard/joystick delta */
		{
			int val = IP_GET_DELTA(entry[sel/ENTRIES]);

			val --;
			if (val < 1) val = 1;
			IP_SET_DELTA(entry[sel/ENTRIES],val);
		}
		else if ((sel % ENTRIES) == 1)
		/* reverse */
		{
			int reverse = entry[sel/ENTRIES]->type & IPF_REVERSE;
			if (reverse)
				reverse=0;
			else
				reverse=IPF_REVERSE;
			entry[sel/ENTRIES]->type &= ~IPF_REVERSE;
			entry[sel/ENTRIES]->type |= reverse;
		}
		else if ((sel % ENTRIES) == 2)
		/* sensitivity */
		{
			int val = IP_GET_SENSITIVITY(entry[sel/ENTRIES]);

			val --;
			if (val < 1) val = 1;
			IP_SET_SENSITIVITY(entry[sel/ENTRIES],val);
		}
	}

	if (input_ui_pressed_repeat(IPT_UI_RIGHT,8))
	{
		if ((sel % ENTRIES) == 0)
		/* keyboard/joystick delta */
		{
			int val = IP_GET_DELTA(entry[sel/ENTRIES]);

			val ++;
			if (val > 255) val = 255;
			IP_SET_DELTA(entry[sel/ENTRIES],val);
		}
		else if ((sel % ENTRIES) == 1)
		/* reverse */
		{
			int reverse = entry[sel/ENTRIES]->type & IPF_REVERSE;
			if (reverse)
				reverse=0;
			else
				reverse=IPF_REVERSE;
			entry[sel/ENTRIES]->type &= ~IPF_REVERSE;
			entry[sel/ENTRIES]->type |= reverse;
		}
		else if ((sel % ENTRIES) == 2)
		/* sensitivity */
		{
			int val = IP_GET_SENSITIVITY(entry[sel/ENTRIES]);

			val ++;
			if (val > 255) val = 255;
			IP_SET_SENSITIVITY(entry[sel/ENTRIES],val);
		}
	}

	if (input_ui_pressed(IPT_UI_SELECT))
	{
		if (sel == total2 - 1) sel = -1;
	}

	if (input_ui_pressed(IPT_UI_CANCEL))
		sel = -1;

	if (input_ui_pressed(IPT_UI_CONFIGURE))
		sel = -2;

	if (sel == -1 || sel == -2)
	{
		schedule_full_refresh();
	}

	return sel + 1;
}

static int mame_stats(struct osd_bitmap *bitmap,int selected)
{
	char temp[10];
	char buf[2048];
	int sel, i;


	sel = selected - 1;

	buf[0] = 0;

	for (i=0; i<COIN_COUNTERS; i++)
	{
		strcat(buf, ui_getstring (UI_coin));
		sprintf(temp, " %c: ", i+'A');
		strcat(buf, temp);
		if (!coins[i])
			strcat (buf, ui_getstring (UI_NA));
		else
		{
			sprintf (temp, "%d", coins[i]);
			strcat (buf, temp);
		}
		if (coinlockedout[i])
		{
			strcat(buf, " ");
			strcat(buf, ui_getstring (UI_locked));
			strcat(buf, "\n");
		}
		else
		{
			strcat(buf, "\n");
		}
	}

	{
		int res;

		/* menu system, use the normal menu keys */
		strcat(buf,"\n\t");
		strcat(buf,ui_getstring (UI_lefthilight));
		strcat(buf," ");
		strcat(buf,ui_getstring (UI_returntomain));
		strcat(buf," ");
		strcat(buf,ui_getstring (UI_righthilight));

		res = ui_displaymessagewindow(bitmap,buf);
		if (res != -1)
		{
			if (res == 1 || input_ui_pressed(IPT_UI_SELECT))
				sel = -1;

			if (res == 2 || input_ui_pressed(IPT_UI_CANCEL))
				sel = -1;

			if (res == 3 || input_ui_pressed(IPT_UI_CONFIGURE))
				sel = -2;
		}
	}

	if (sel == -1 || sel == -2)
	{
		scroll_reset = 1;
		schedule_full_refresh();
	}

	return sel + 1;
}

static int setcustombutton(struct osd_bitmap *bitmap, int selected)
{
	const char *menu_item[10];
	const char *menu_subitem[10];
	const char *cbname[10];
	const char *afname[4];
	int i, player, button, sel, arrowize, val;
	int total;
	char buf[8][32], buf2[8][32];
	extern UINT16 custom_button[2][4];

	for (i = 0 ; i < 10 ; i++)
	{
		cbname[i] = ui_getstring(i + UI_custombutton_AB);
	}

	for (i = 0 ; i < 4 ; i++)
	{
		afname[i] = ui_getstring(i + UI_autofire_A);
	}

	sel = selected - 1;

	total = 0;
	arrowize = 0;
	for (player = 0; player < 2; player++)
	{
		for (button = 0; button < 4; button++)
		{
			i = player * 4 + button;
			val = custom_button[player][button];

			sprintf(buf[i], "P%d %s%d", player + 1, ui_getstring(UI_custombutton), button + 1);
			if (val == 0)
				sprintf(buf2[i], "%s", ui_getstring(UI_custombutton_notuse));
			else if (val <= 10)
				strcpy(buf2[i], cbname[val - 1]);
			else if (val <= 18)
				sprintf(buf2[i], "%s %d", ui_getstring(UI_custombutton_combo), val - 10);
			else
				strcpy(buf2[i], afname[val - 19]);

			menu_item[i] = buf[i];
			menu_subitem[i] = buf2[i];

			if (i == sel)
			{
				if (custom_button[player][button] == BTN_NONE)
					arrowize = 2;
				else if (custom_button[player][button] == BTN_AF_D)
					arrowize = 1;
				else
					arrowize = 3;
			}

			total++;
		}
	}

	menu_item[total] = ui_getstring(UI_returntomain);
	menu_subitem[total]  = 0;
	menu_item[total + 1] = 0;	/* terminate array */
	total++;

	ui_displaymenu(bitmap, menu_item,menu_subitem,0,sel,arrowize);

	if (input_ui_pressed_repeat(IPT_UI_DOWN,4))
		sel = (sel + 1) % total;

	if (input_ui_pressed_repeat(IPT_UI_UP,4))
		sel = (sel + total - 1) % total;

	player = sel / 4;
	button = sel % 4;

	if (input_ui_pressed_repeat(IPT_UI_RIGHT,8))
	{
		if (sel < total - 1)
		{
			if (custom_button[player][button] < BTN_AF_D)
			{
				custom_button[player][button]++;

				schedule_full_refresh();
			}
		}
	}

	if (input_ui_pressed_repeat(IPT_UI_LEFT,8))
	{
		if (sel < total - 1)
		{
			if (custom_button[player][button] > BTN_NONE)
			{
				custom_button[player][button]--;

				schedule_full_refresh();
			}
		}
	}

	if (input_ui_pressed(IPT_UI_SELECT))
		if (sel == total - 1)
			sel = -1;

	if (input_ui_pressed(IPT_UI_CANCEL))
		sel = -1;

	if (input_ui_pressed(IPT_UI_CONFIGURE))
		sel = -2;

	if (sel == -1 || sel == -2)
	{
		schedule_full_refresh();
	}

	return sel + 1;
}


static int setautofire(struct osd_bitmap *bitmap, int selected)
{
	const char *menu_item[12];
	const char *menu_subitem[12];
	char flag[12];
	int i, player, button, sel, arrowize;
	int total;
	char buf[8][64], buf2[8][4];
	extern UINT16 autofire_delay[2][4];

	sel = selected - 1;

	total = 0;
	arrowize = 0;
	for (player = 0; player < 2; player++)
	{
		for (button = 0; button < 4; button++)
		{
			i = player * 4 + button;
			sprintf(buf[i], "P%d \x81%c%s %s", player + 1, button + 0xc0,
				ui_getstring(UI_autofire_button), ui_getstring(UI_autofire_delay));
			sprintf(buf2[i], "%d", autofire_delay[player][button] + 1);

			menu_item[i] = buf[i];
			menu_subitem[i] = buf2[i];
			flag[i]      = 0;
			total++;

			if (i == sel)
			{
				if (autofire_delay[player][button] == 0)
					arrowize = 2;
				if (autofire_delay[player][button] == 9)
					arrowize |= 1;
			}
		}
	}

	menu_item[total] = ui_getstring(UI_returntomain);
	menu_subitem[total]  = 0;
	menu_item[total + 1] = 0;	/* terminate array */
	total++;

	ui_displaymenu(bitmap, menu_item,menu_subitem,flag,sel,arrowize);

	if (input_ui_pressed_repeat(IPT_UI_DOWN,4))
		sel = (sel + 1) % total;

	if (input_ui_pressed_repeat(IPT_UI_UP,4))
		sel = (sel + total - 1) % total;

	player = sel / 4;
	button = sel % 4;

	if (input_ui_pressed_repeat(IPT_UI_RIGHT,8))
	{
		if (sel < total - 1)
		{
			if (autofire_delay[player][button] < 9)
			{
				autofire_delay[player][button]++;

				schedule_full_refresh();
			}
		}
	}

	if (input_ui_pressed_repeat(IPT_UI_LEFT,8))
	{
		if (sel < total - 1)
		{
			if (autofire_delay[player][button] > 0)
			{
				autofire_delay[player][button]--;

				schedule_full_refresh();
			}
		}
	}

	if (input_ui_pressed(IPT_UI_SELECT))
	{
		if (sel == total - 1) sel = -1;
	}

	if (input_ui_pressed(IPT_UI_CANCEL))
		sel = -1;

	if (input_ui_pressed(IPT_UI_CONFIGURE))
		sel = -2;

	if (sel == -1 || sel == -2)
	{
		schedule_full_refresh();
	}

	return sel + 1;
}


int showcopyright(struct osd_bitmap *bitmap)
{
	int done, res;
	char buf[1000];
	char buf2[256];

	strcpy (buf, ui_getstring(UI_copyright1));
	strcat (buf, "\n\n");
	sprintf(buf2, ui_getstring(UI_copyright2), Machine->gamedrv->description);
	strcat (buf, buf2);
	strcat (buf, "\n\n");
	strcat (buf, ui_getstring(UI_copyright3));

	erase_screen(bitmap);
	ui_displaymessagewindow(bitmap,buf);

	setup_selected = -1;
	done = 0;
	do
	{
		res = ui_displaymessagewindow(bitmap,buf);
		update_video_and_audio();
		if (res != -1)
		{
			if (res == 2 || input_ui_pressed(IPT_UI_CANCEL) || IsCloseButtonPressed())
			{
				setup_selected = 0;
				return 1;
			}
			if (keyboard_pressed_memory(KEYCODE_O) ||
					input_ui_pressed(IPT_UI_LEFT))
				done = 1;
			if (done == 1 && (keyboard_pressed_memory(KEYCODE_K) ||
					input_ui_pressed(IPT_UI_RIGHT)))
				done = 2;
		}
	} while (done < 2);

	scroll_reset = 1;
	setup_selected = 0;
	erase_screen(bitmap);
	update_video_and_audio();

	return 0;
}

static int displaygameinfo(struct osd_bitmap *bitmap,int selected)
{
	int i, res;
	char buf[4096];
	char buf2[256];
	int sel;

	sel = selected - 1;

	sprintf(buf,"%s\n%s %s\n\n%s:\n",Machine->gamedrv->description,Machine->gamedrv->year,Machine->gamedrv->manufacturer,
		ui_getstring (UI_cpu));
	i = 0;
	while (i < MAX_CPU && Machine->drv->cpu[i].cpu_type)
	{
		if (Machine->drv->cpu[i].cpu_clock >= 1000000)
			sprintf(&buf[strlen(buf)],"%s %d.%06d MHz",
					cputype_name(Machine->drv->cpu[i].cpu_type),
					Machine->drv->cpu[i].cpu_clock / 1000000,
					Machine->drv->cpu[i].cpu_clock % 1000000);
		else
			sprintf(&buf[strlen(buf)],"%s %d.%03d kHz",
					cputype_name(Machine->drv->cpu[i].cpu_type),
					Machine->drv->cpu[i].cpu_clock / 1000,
					Machine->drv->cpu[i].cpu_clock % 1000);

		if (Machine->drv->cpu[i].cpu_type & CPU_AUDIO_CPU)
		{
			sprintf (buf2, " (%s)", ui_getstring (UI_sound_lc));
			strcat(buf, buf2);
		}

		strcat(buf,"\n");

		i++;
	}

	sprintf(buf2, "\n%s", ui_getstring(UI_sound));
	strcat(buf, buf2);
	if (Machine->drv->sound_attributes & SOUND_SUPPORTS_STEREO)
		sprintf(&buf[strlen(buf)]," (%s)", ui_getstring (UI_stereo));
	strcat(buf,":\n");

	i = 0;
	while (i < MAX_SOUND && Machine->drv->sound[i].sound_type)
	{
		if (sound_num(&Machine->drv->sound[i]))
			sprintf(&buf[strlen(buf)],"%dx",sound_num(&Machine->drv->sound[i]));

		sprintf(&buf[strlen(buf)],"%s",sound_name(&Machine->drv->sound[i]));

		if (sound_clock(&Machine->drv->sound[i]))
		{
			if (sound_clock(&Machine->drv->sound[i]) >= 1000000)
				sprintf(&buf[strlen(buf)]," %d.%06d MHz",
						sound_clock(&Machine->drv->sound[i]) / 1000000,
						sound_clock(&Machine->drv->sound[i]) % 1000000);
			else
				sprintf(&buf[strlen(buf)]," %d.%03d kHz",
						sound_clock(&Machine->drv->sound[i]) / 1000,
						sound_clock(&Machine->drv->sound[i]) % 1000);
		}

		strcat(buf,"\n");

		i++;
	}

	{
		int pixelx,pixely,tmax,tmin,rem;

		pixelx = 4 * (Machine->visible_area.max_y - Machine->visible_area.min_y + 1);
		pixely = 3 * (Machine->visible_area.max_x - Machine->visible_area.min_x + 1);

		/* calculate MCD */
		if (pixelx >= pixely)
		{
			tmax = pixelx;
			tmin = pixely;
		}
		else
		{
			tmax = pixely;
			tmin = pixelx;
		}
		while ( (rem = tmax % tmin) )
		{
			tmax = tmin;
			tmin = rem;
		}
		/* tmin is now the MCD */

		pixelx /= tmin;
		pixely /= tmin;

		sprintf(&buf[strlen(buf)],"\n%s:\n", ui_getstring (UI_screenres));
		sprintf(&buf[strlen(buf)],"%d x %d (H) %f Hz\n",
				Machine->visible_area.max_x - Machine->visible_area.min_x + 1,
				Machine->visible_area.max_y - Machine->visible_area.min_y + 1,
				Machine->drv->frames_per_second);

		if (sel != -1)
		{
#ifdef BUILD_ENGLISH
			sprintf(&buf[strlen(buf)],"pixel aspect ratio %d:%d\n", pixelx, pixely);
			sprintf(&buf[strlen(buf)],"%d colors ",Machine->drv->total_colors);
#else
			sprintf(&buf[strlen(buf)],"AXyNg %d:%d\n", pixelx, pixely);
			sprintf(&buf[strlen(buf)],"%dF (v16-bit)\n", Machine->drv->total_colors);
#endif
		}
	}

	if (sel == -1)
	{
		/* startup info, print MAME version and ask for any key */

		sprintf(buf2, "\n\t%s ", ui_getstring(UI_mame));	/* \t means that the line will be centered */
		strcat(buf, buf2);

		strcat(buf,build_version);
		sel = 0;

		change_port_mask();
	}

	res = ui_displaymessagewindow(bitmap,buf);
	if (res >= 0)
	{
		if (res == 1 || input_ui_pressed(IPT_UI_SELECT))
			sel = -1;

		if (res == 2 || input_ui_pressed(IPT_UI_CANCEL) || IsCloseButtonPressed())
			sel = -1;

		if (res == 3 || input_ui_pressed(IPT_UI_CONFIGURE))
			sel = -2;
	}

	if (sel == -1 || sel == -2)
	{
		scroll_reset = 1;
		schedule_full_refresh();
	}

	return sel + 1;
}


int showgamewarnings(struct osd_bitmap *bitmap)
{
	erase_screen(bitmap);

	update_video_and_audio();

	/* clear the input memory */
	while (code_read_async() != CODE_NONE);

#ifdef KAILLERA
	if (!kPlay)		/* disable info with network game*/
#endif
	while (displaygameinfo(bitmap,0) == 1)
		update_video_and_audio();

	/* clear the input memory */
	while (code_read_async() != CODE_NONE);

	erase_screen(bitmap);
	/* make sure that the screen is really cleared, in case autoframeskip kicked in */
	update_video_and_audio();
	update_video_and_audio();
	update_video_and_audio();
	update_video_and_audio();

	return 0;
}


/* Word-wraps the text in the specified buffer to fit in maxwidth characters per line.
   The contents of the buffer are modified.
   Known limitations: Words longer than maxwidth cause the function to fail. */
int wordwrap_text_buffer(char *buffer, int maxwidth, int fonttype)
{
	char *c;
	int width, lines;

	width = 0;
	lines = 0;
	c = buffer;

#ifndef BUILD_ENGLISH
	{
		char *c2, *textcopy;
		int w, ct, code;

		if ((textcopy = MyMalloc(MAX_SCROLLBUF)) == 0)
			return 0;

		memset(textcopy, 0, MAX_SCROLLBUF);

		if (fonttype & FONT_PROPORTIONAL)
			maxwidth -= Machine->uifontwidth2p;

		c2 = textcopy;
		code = 0;

		while (*c)
		{
			width = 0;

			while (*c && *c != '\n')
			{
				ct = get_chartype(c);

				if (ct == ASCII || ct == HKANA)
				{
					code = 0;
					w = Machine->uifontwidth;
					if (*c == '\t') w = 0;
				}
				else
				{
					code = get_uifont_code(c);
					w = get_fontwidth(code, fonttype);
				}

				if (width + w >= maxwidth)
				{
					// v|[VitHg̏ꍇAǓ_ƁE`I͉sȂ(֑)
					if ((code >= 0x01 && code <= 0x09) &&
						fonttype == FONT_PROPORTIONAL)
					{
						*c2++ = *c++;
						*c2++ = *c++;
						*c2++ = '\n';
						break;
					}
					else
					{
						*c2++ = '\n';
						break;
					}
				}

				width += w;

				*c2++ = *c++;
				if (ct == ZKANJI)
					*c2++ = *c++;
			}

			if (*c == '\n')
				*c2++ = *c++;

			lines++;
		}

		*c2 = '\0';

		strcpy(buffer, textcopy);
		MyFree(textcopy);
	}
#else
	{
		while (*buffer)
		{
			if (*buffer == '\n')
			{
				buffer++;
				width = 0;
				continue;
			}

			width += Machine->uifontwidth;

			if (width > maxwidth - Machine->uifontwidth)
			{
				/* backtrack until a space is found */
				while (*buffer != ' ')
				{
					buffer--;
					width -= Machine->uifontwidth;
				}
				if (width < Machine->uifontwidth) break;	/* word too long */

				/* replace space with a newline */
				*buffer = '\n';
			}
			else
				buffer++;
		}

		while(*c)
			if (*c++ == '\n') lines++;
	}
#endif
	return lines;
}

static int count_lines_in_buffer(char *buffer)
{
	int lines = 0;
	char c;

	while ( (c = *buffer++) )
		if (c == '\n') lines++;

	return lines;
}

/* Display lines from buffer, starting with line 'scroll', in a width x height text window */
static int display_scroll_message (struct osd_bitmap *bitmap, int *scroll, int width, int height, char *buf, int fonttype)
{
	struct DisplayText dt[256];
	int curr_dt = 0;
	const char *uparrow = ui_getstring (UI_uparrow);
	const char *downarrow = ui_getstring (UI_downarrow);
	char textcopy[MAX_SCROLLBUF];
	char *copy;
	int leftoffs,topoffs;
	int first = *scroll;
	int buflines,showlines;
	int i, ret;


	/* draw box */
	leftoffs = (Machine->uiwidth - width) / 2;// - 1;
	if (leftoffs < 0) leftoffs = 0;
	topoffs = (Machine->uiheight - height * ROWHEIGHT) / 2;
	draw3dbox(bitmap,leftoffs,topoffs-BOXADJUST,width + 1,height*ROWHEIGHT+BOXADJUST*2,1);

	buflines = count_lines_in_buffer (buf);
	if (first > 0)
	{
		if (buflines <= height)
			first = 0;
		else
		{
			height--;
			if (first > (buflines - height))
				first = buflines - height;
		}
		*scroll = first;
	}

	if (first != 0)
	{
		/* indicate that scrolling upward is possible */
		dt[curr_dt].text = uparrow;
		dt[curr_dt].color = UI_COLOR_NORMAL;
		dt[curr_dt].x = (Machine->uiwidth - uaw) / 2;
		dt[curr_dt].y = topoffs + curr_dt * ROWHEIGHT;
		curr_dt++;
	}

	if ((buflines - first) > height)
		showlines = height - 1;
	else
		showlines = height;

	ret = buflines;

	if (first)
		ret |= SCR_PREV_PAGE; // Oy[W

	if (showlines == (height - 1))
		ret |= SCR_NEXT_PAGE; // y[W

	/* skip to first line */
	while (first > 0)
	{
		char c;

		while ( (c = *buf++) )
		{
			if (c == '\n')
			{
				first--;
				break;
			}
		}
	}

	/* copy 'showlines' lines from buffer, starting with line 'first' */
	copy = textcopy;
	for (i = 0; i < showlines; i++)
	{
		char *copystart = copy;

		while (*buf && *buf != '\n')
		{
			*copy = *buf;
			copy++;
			buf++;
		}
		*copy = '\0';
		copy++;
		if (*buf == '\n')
			buf++;

		if (*copystart == '\t') /* center text */
		{
			copystart++;
			dt[curr_dt].x = (Machine->uiwidth - strwidth(copystart, fonttype)) / 2;
		}
		else
			dt[curr_dt].x = leftoffs + Machine->uifontwidth;

		dt[curr_dt].text = copystart;
		dt[curr_dt].color = UI_COLOR_NORMAL;
		dt[curr_dt].y = topoffs + curr_dt*ROWHEIGHT;
		curr_dt++;
	}

	if (showlines == (height - 1))
	{
		/* indicate that scrolling downward is possible */
		dt[curr_dt].text = downarrow;
		dt[curr_dt].color = UI_COLOR_NORMAL;
		dt[curr_dt].x = (Machine->uiwidth - daw) / 2;
		dt[curr_dt].y = topoffs + curr_dt*ROWHEIGHT;
		curr_dt++;
	}

	dt[curr_dt].text = 0;	/* terminate array */

	displaytext(bitmap,dt,fonttype);

	return ret;
}


/* Display text entry for current driver from history.dat and mameinfo.dat. */
static int displayhistory (struct osd_bitmap *bitmap, int selected)
{
	int sel, ret = 0, buflines = 0, visible = 0;
	static int maxcols = 0, maxrows = 0;

	sel = selected - 1;

	if (!history_buf)
	{
		maxcols = Machine->uiwidth - Machine->uifontwidth * 4;
		maxrows = (Machine->uiheight / ROWHEIGHT) - 2;

		/* allocate a buffer for the text */
		if ((history_buf = MyMalloc(MAX_SCROLLBUF)))
		{
			int len;

			memset(history_buf, 0, MAX_SCROLLBUF);

			len = strlen(history_buf);
			
			/* try to load entry */
			if (!load_driver_history(Machine->gamedrv, history_buf, MAX_SCROLLBUF))
			{
				int lines;

				strcat(history_buf,"\n\t");
				strcat(history_buf,ui_getstring (UI_lefthilight));
				strcat(history_buf," ");
				strcat(history_buf,ui_getstring (UI_returntomain));
				strcat(history_buf," ");
				strcat(history_buf,ui_getstring (UI_righthilight));
				strcat(history_buf,"\n");

				lines = wordwrap_text_buffer(history_buf, maxcols, FONT_PROPORTIONAL);

				if (lines < maxrows)
					maxrows = lines;
			}
			else
			{
				MyFree(history_buf);
				history_buf = 0;
			}
		}
	}

	if (history_buf)
	{
		ret = display_scroll_message(bitmap, &history_scroll, maxcols, maxrows, history_buf, FONT_PROPORTIONAL);

		buflines = ret & SCR_PAGE_MASK;
		visible  = maxrows - ((ret & SCR_PREV_PAGE) != 0) - ((ret & SCR_NEXT_PAGE) != 0);
	}
	else
	{
		char msg[80];

		strcpy(msg,"\t");
		strcat(msg,ui_getstring(UI_historymissing));
		strcat(msg,"\n\n\t");
		strcat(msg,ui_getstring (UI_lefthilight));
		strcat(msg," ");
		strcat(msg,ui_getstring (UI_returntomain));
		strcat(msg," ");
		strcat(msg,ui_getstring (UI_righthilight));
		ui_displaymessagewindow(bitmap,msg);
	}

	if ((history_scroll > 0) && input_ui_pressed_repeat(IPT_UI_UP,4))
	{
		if (history_scroll == 2) history_scroll = 0;	/* 1 would be the same as 0, but with arrow on top */
		else history_scroll--;
	}

	if (input_ui_pressed_repeat(IPT_UI_DOWN,4))
	{
		if (history_scroll == 0) history_scroll = 2;	/* 1 would be the same as 0, but with arrow on top */
		else history_scroll++;
	}

	if ((ret & SCR_PREV_PAGE) && input_ui_pressed_repeat(IPT_UI_LEFT,6))
	{
		if (history_scroll == maxrows - 1)
			history_scroll = 0;
		else if (history_scroll - visible < 0)
			history_scroll = 0;
		else
			history_scroll -= visible;
	}

	if ((ret & SCR_NEXT_PAGE) && input_ui_pressed_repeat(IPT_UI_RIGHT,6))
	{
		if (history_scroll >= buflines - visible)
			history_scroll = buflines - visible;
		else
			history_scroll += visible;
	}

	if (input_ui_pressed(IPT_UI_SELECT))
		sel = -1;

	if (input_ui_pressed(IPT_UI_CANCEL))
		sel = -1;

	if (input_ui_pressed(IPT_UI_CONFIGURE))
		sel = -2;

	if (sel == -1 || sel == -2)
	{
		schedule_full_refresh();

		/* force buffer to be recreated */
		if (history_buf)
		{
			MyFree(history_buf);
			history_buf = 0;
		}
	}

	return sel + 1;
}

static void convert_command_move(char *buf)
{
	unsigned char *c = (unsigned char *)buf;

	while (*c)
	{
		if (iszen1(*c) && iszen2(*(c+1)))
		{
			c+=2;
		}
		else
		{
			if (*c == '_')
			{
				switch(*(c+1))
				{
					case '1':
					case '2':
					case '3':
					case '4':
					case '6':
					case '7':
					case '8':
					case '9':
						*c     = 0x81;
						*(c+1) = 0xd0 + (*(c+1) - '0');
						c+=2;
						break;

					case 'A':
					case 'B':
					case 'C':
					case 'D':
						*c     = 0x81;
						*(c+1) = 0xc0 + (*(c+1) - 'A');
						c+=2;
						break;

					case '+':
						*c     = 0x81;
						*(c+1) = 0xc4;
						c+=2;
						break;

					default:
						c++;
						break;
				}
			}
			else if (*c == '\\' && *(c+1) == 'n')
			{
				*c     = ' ';
				*(c+1) = '\n';
				c+=2;
			}
			else
				c++;
		}
	}
}


/* Display text entry for current driver from command.dat. */
static int displaycommand(struct osd_bitmap *bitmap, int selected, int shortcut)
{
	int sel, ret = 0, buflines = 0, visible = 0;
	static int maxcols = 0, maxrows = 0;

	sel = selected - 1;

	if (!commandlist_buf)
	{
		maxcols = Machine->uiwidth - 3*Machine->uifontwidth;
		maxrows = (Machine->uiheight / ROWHEIGHT) - 2;

		/* allocate a buffer for the text */
		if ((commandlist_buf = MyMalloc(MAX_SCROLLBUF)))
		{
			memset(commandlist_buf, 0, MAX_SCROLLBUF);

			/* try to load entry */
			if (!load_driver_command_ex (Machine->gamedrv, commandlist_buf, MAX_SCROLLBUF, sel))
			{
				int height = 1;
				static int presel = -1;
				char *c;

				if( presel != sel )
				{
					command_scroll = 0;
					presel = sel;
				}

				convert_command_move(commandlist_buf);

				strcat(commandlist_buf,"\n\t");
				strcat(commandlist_buf,ui_getstring(UI_lefthilight));
				strcat(commandlist_buf," ");
				strcat(commandlist_buf,ui_getstring(UI_returntoprior));
				strcat(commandlist_buf," ");
				strcat(commandlist_buf,ui_getstring(UI_righthilight));
				strcat(commandlist_buf,"\n");

				c = commandlist_buf;

				while(*c)
					if (*c++ == '\n') height++;

				if (height <= maxrows)
					maxrows = height - 1;
			}
			else
			{
				MyFree(commandlist_buf);
				commandlist_buf = 0;
			}
		}
	}

	if (commandlist_buf)
	{
		ret = display_scroll_message(bitmap, &command_scroll, maxcols, maxrows, commandlist_buf, FONT_NORMAL);

		buflines = ret & SCR_PAGE_MASK;
		visible  = maxrows - ((ret & SCR_PREV_PAGE) != 0) - ((ret & SCR_NEXT_PAGE) != 0);
	}
	else
	{
		char msg[80];

		strcpy(msg, "\t");
		strcat(msg, ui_getstring(UI_commandmissing));
		strcat(msg, "\n\n\t");
		strcat(msg, ui_getstring(UI_lefthilight));
		strcat(msg, " ");
		if (shortcut)
			strcat(msg, ui_getstring(UI_returntogame));
		else
			strcat(msg, ui_getstring(UI_returntomain));
		strcat(msg, " ");
		strcat(msg, ui_getstring(UI_righthilight));
		ui_displaymessagewindow(bitmap, msg);
	}

	if ((command_scroll > 0) && input_ui_pressed_repeat(IPT_UI_UP,4))
	{
		if (command_scroll == 2) command_scroll = 0;	/* 1 would be the same as 0, but with arrow on top */
		else command_scroll--;
	}

	if (input_ui_pressed_repeat(IPT_UI_DOWN,4))
	{
		if (command_scroll == 0) command_scroll = 2;	/* 1 would be the same as 0, but with arrow on top */
		else command_scroll++;
	}

	if ((ret & SCR_PREV_PAGE) && input_ui_pressed_repeat(IPT_UI_LEFT,6))
	{
		if (command_scroll == maxrows - 1)
			command_scroll = 0;
		else if (command_scroll - visible < 0)
			command_scroll = 0;
		else
			command_scroll -= visible;
	}

	if ((ret & SCR_NEXT_PAGE) && input_ui_pressed_repeat(IPT_UI_RIGHT,6))
	{
		if (command_scroll >= buflines - visible)
			command_scroll = buflines - visible;
		else
			command_scroll += visible;
	}

	if (input_ui_pressed(IPT_UI_SELECT))
		sel = -1;

	if (input_ui_pressed(IPT_UI_CANCEL))
		sel = -1;

	if (input_ui_pressed(IPT_UI_CONFIGURE) && shortcut == 0)
		sel = -2;

	if (sel == -1 || sel == -2)
	{
		schedule_full_refresh();

		/* force buffer to be recreated */
		if (commandlist_buf)
		{
			MyFree(commandlist_buf);
			commandlist_buf = 0;
		}
	}

	return sel + 1;
}


static int displaycommand_ex(struct osd_bitmap *bitmap, int selected, int shortcut)
{
	static const char *menu_item[64];
	static int total = 0;
	int sel;

	sel = selected - 1;


	/* If a submenu has been selected, go there */
	if (command_lastselected)
	{
		command_lastselected = displaycommand(bitmap, (sel + 1), shortcut);

		if (command_lastselected == -1)
		{
			sel = -2;
			command_lastselected = 1;
			schedule_full_refresh();
		}

		return sel + 1;
	}

	/* No submenu active, display the main command index menu */
	if (total == -1)
	{
		char msg[80];

		strcpy(msg, "\t");
		strcat(msg, ui_getstring(UI_commandmissing));
		strcat(msg, "\n\n\t");
		strcat(msg, ui_getstring(UI_lefthilight));
		strcat(msg, " ");
		if (shortcut)
			strcat(msg, ui_getstring(UI_returntogame));
		else
			strcat(msg, ui_getstring(UI_returntomain));
		strcat(msg, " ");
		strcat(msg, ui_getstring(UI_righthilight));
		ui_displaymessagewindow(bitmap, msg);
	}
	else
	{
		if (total == 0 || command_load)
		{
			total = command_sub_menu(Machine->gamedrv, menu_item);
			command_load = 0;

			if (total > 0)
			{
				if (shortcut)
					menu_item[total++] = ui_getstring(UI_returntogame);
				else
					menu_item[total++] = ui_getstring(UI_returntomain);
				menu_item[total] = 0;
			}
			else
			{
				char msg[80];

				total = -1;
				strcpy(msg, "\t");
				strcat(msg, ui_getstring(UI_commandmissing));
				strcat(msg, "\n\n\t");
				strcat(msg, ui_getstring(UI_lefthilight));
				strcat(msg, " ");
				if (shortcut)
					strcat(msg, ui_getstring(UI_returntogame));
				else
					strcat(msg, ui_getstring(UI_returntomain));
				strcat(msg, " ");
				strcat(msg, ui_getstring(UI_righthilight));
				ui_displaymessagewindow(bitmap, msg);
			}
		}
		else
		{
			if (shortcut)
				menu_item[total - 1] = ui_getstring(UI_returntogame);
			else
				menu_item[total - 1] = ui_getstring(UI_returntomain);

			ui_displaymenu(bitmap,menu_item,0,0,sel,0);
		}

		if (total > 0)
		{
			if (input_ui_pressed_repeat(IPT_UI_DOWN,4))
				sel = (sel + 1) % total;

			if (input_ui_pressed_repeat(IPT_UI_UP,4))
				sel = (sel + total - 1) % total;
		}
	}

	if (input_ui_pressed(IPT_UI_SELECT))
	{
		if (sel == (total - 1) || total == -1) // return to main
		{
			command_lastselected = 0;
			sel = -1;
		}
		else
		{
			command_lastselected = 1;
			schedule_full_refresh();
		}
	}

	/* Cancel pops us up a menu level */
	if (input_ui_pressed(IPT_UI_CANCEL))
		sel = -1;

	/* The UI key takes us all the way back out */
	if (input_ui_pressed(IPT_UI_CONFIGURE) && shortcut == 0)
		sel = -2;

	if (sel == -1 || sel == -2)
	{
		schedule_full_refresh();
		total = 0;
	}

	return sel + 1;
}


static int memcard_menu(struct osd_bitmap *bitmap, int selection)
{
	int sel;
	int menutotal = 0;
	const char *menuitem[10];
	char buf[256];
	char buf2[256];

	sel = selection - 1;

	sprintf(buf, "%s %03d", ui_getstring (UI_loadcard), mcd_number);
	menuitem[menutotal++] = buf;
	menuitem[menutotal++] = ui_getstring (UI_ejectcard);
	menuitem[menutotal++] = ui_getstring (UI_createcard);
	menuitem[menutotal++] = ui_getstring (UI_resetcard);
	menuitem[menutotal++] = ui_getstring (UI_returntomain);
	menuitem[menutotal] = 0;

	if (mcd_action!=0)
	{
		switch(mcd_action)
		{
			case 1:
				strcpy (buf2, ui_getstring (UI_loadfailed));
				break;
			case 2:
				strcpy (buf2, ui_getstring (UI_loadok));
				break;
			case 3:
				strcpy (buf2, ui_getstring (UI_cardejected));
				break;
			case 4:
				strcpy (buf2, ui_getstring (UI_cardcreated));
				break;
			case 5:
				strcpy (buf2, ui_getstring (UI_cardcreatedfailed));
				strcat (buf2, "\n");
				strcat (buf2, ui_getstring (UI_cardcreatedfailed2));
				break;
			default:
				strcpy (buf2, ui_getstring (UI_carderror));
				break;
		}

		strcat (buf2, "\n");
		ui_displaymessagewindow(bitmap,buf2);
		if (input_ui_pressed(IPT_UI_SELECT))
			mcd_action = 0;
	}
	else
	{
		ui_displaymenu(bitmap,menuitem,0,0,sel,0);

		if (input_ui_pressed_repeat(IPT_UI_RIGHT,8))
			mcd_number = (mcd_number + 1) % 1000;

		if (input_ui_pressed_repeat(IPT_UI_LEFT,8))
			mcd_number = (mcd_number + 999) % 1000;

		if (input_ui_pressed_repeat(IPT_UI_DOWN,4))
			sel = (sel + 1) % menutotal;

		if (input_ui_pressed_repeat(IPT_UI_UP,4))
			sel = (sel + menutotal - 1) % menutotal;

		if (input_ui_pressed(IPT_UI_SELECT))
		{
			switch(sel)
			{
			case 0:
				neogeo_memcard_eject();
				if (neogeo_memcard_load(mcd_number))
				{
					memcard_status=1;
					memcard_number=mcd_number;
					mcd_action = 2;
				}
				else
					mcd_action = 1;
				break;
			case 1:
				neogeo_memcard_eject();
				mcd_action = 3;
				break;
			case 2:
				if (neogeo_memcard_create(mcd_number))
					mcd_action = 4;
				else
					mcd_action = 5;
				break;
			case 3:
				memcard_manager=1;
				sel=-2;
				machine_reset();
				break;
			case 4:
				sel=-1;
				break;
			}
		}

		if (input_ui_pressed(IPT_UI_CANCEL))
			sel = -1;

		if (input_ui_pressed(IPT_UI_CONFIGURE))
			sel = -2;

		if (sel == -1 || sel == -2)
		{
			schedule_full_refresh();
		}
	}

	return sel + 1;
}


static void save_load_state(struct osd_bitmap *bitmap, int pressed)
{
	int check = 0;
	char buf[100];
	const char *leftarrow = ui_getstring(UI_leftarrow);
	const char *rightarrow = ui_getstring(UI_rightarrow);
	const char *blank = " ";
	extern char date_str[];
	extern char time_str[];
	extern char info_str[];

	osd_sound_enable(0);

	neogeo_save_screen();
	paused = 1;

	while (pressed && !IsCloseButtonPressed())
	{
		check = check_state_data();

		sprintf(buf,
			"%s - No. %s%3d%s\n\n%s: %s\n%s: %s\n%s: %s",
			(pressed == 1) ? ui_getstring(UI_savestate) : ui_getstring(UI_loadstate),
			(saveslot == 0) ? blank : leftarrow,
			saveslot,
			(saveslot == 999) ? blank : rightarrow,
			ui_getstring(UI_date),
			date_str,
			ui_getstring(UI_time),
			time_str,
			ui_getstring(UI_stateinfo),
			info_str
		);

		erase_screen(bitmap);
		draw_screen();
		ui_displaymessagewindow(bitmap, buf);

		if (input_ui_pressed(IPT_UI_CANCEL))
			pressed = 0;

		if (input_ui_pressed(IPT_UI_SELECT))
		{
			if (pressed == 1)
			{
				if (save_contexts(OSD_FILETYPE_STATE, saveslot))
					usrintf_showmessage(ui_getstring(UI_savestateok));
				pressed = 0;
			}
			else if (pressed == 2 && check)
			{
				if (restore_contexts(OSD_FILETYPE_STATE, saveslot))
					usrintf_showmessage(ui_getstring(UI_loadstateok));
				pressed = 0;
			}
		}

		if (input_ui_pressed_repeat(IPT_UI_LEFT,8))
		{
			if (saveslot)
				saveslot--;
		}

		if (input_ui_pressed_repeat(IPT_UI_RIGHT,8))
		{
			if (saveslot <= 999)
				saveslot++;
		}

		update_video_and_audio();
	}

	paused = 0;

	osd_sound_enable(1);
}


static void save_load_combo(struct osd_bitmap *bitmap, int pressed)
{
	int check = 0;
	int i = 0;
	char buf[400];
	const char *leftarrow = ui_getstring(UI_leftarrow);
	const char *rightarrow = ui_getstring(UI_rightarrow);
	const char *blank = " ";
	char combo_name[8][MAX_COMBONAME_LEN];

	osd_sound_enable(0);

	neogeo_save_screen();
	paused = 1;

	while (pressed && !IsCloseButtonPressed())
	{
		check = check_combo_data(combo_name);

		sprintf(buf,
			"%s           %s%-8s.cb%d%s\n",
			(pressed == 1) ? ui_getstring(UI_savecombo) : ui_getstring(UI_loadcombo),
			(combosaveslot == 0) ? blank : leftarrow,
			Machine->gamedrv->name,
			combosaveslot,
			(combosaveslot == 9) ? blank : rightarrow
		);

		for (i = 0; i < 8; i++)
		{
			sprintf(buf + strlen(buf),
				"\n%d: %s",
				i + 1,
				combo_name[i]
			);
		}

		erase_screen(bitmap);
		draw_screen();
		ui_displaymessagewindow(bitmap, buf);

		if (input_ui_pressed(IPT_UI_CANCEL))
		{
			if(pressed == 1)
			{
				combo_record_cancel();
			}
			pressed = 0;
		}

		if (pressed == 1)
		{
			if (input_ui_pressed(IPT_UI_SELECT))
			{
				if (combo_record_end())
				{
					change_port_mask();
					usrintf_showmessage(ui_getstring(UI_savecombook));
				}
				pressed = 0;
			}
		}

		if (pressed == 2)
		{
			if (input_ui_pressed(IPT_UI_SELECT) && check)
			{
				if (check)
				{
					if (custom_button_init() && read_combo_ex())
					{
						change_port_mask();
						usrintf_showmessage(ui_getstring(UI_loadcombook));
					}
				}
				pressed = 0;
			}
		}

		if (input_ui_pressed_repeat(IPT_UI_LEFT,8))
		{
			if (combosaveslot)
				combosaveslot--;
		}

		if (input_ui_pressed_repeat(IPT_UI_RIGHT,8))
		{
			if (combosaveslot < 9)
				combosaveslot++;
		}

		update_video_and_audio();
	}

	paused = 0;

	osd_sound_enable(1);
}


enum { UI_SWITCH = 0,UI_DEFCODE,UI_CODE,UI_ANALOG,UI_CALIBRATE,
		UI_STATS,UI_GAMEINFO, UI_HISTORY, UI_CUSTOM, UI_AUTOFIRE,
		UI_CHEAT,UI_RESET,UI_MEMCARD,UI_COMMAND,UI_EXIT };


#define MAX_SETUPMENU_ITEMS 20
static const char *menu_item[MAX_SETUPMENU_ITEMS];
static int menu_action[MAX_SETUPMENU_ITEMS];
static int menu_total;


static void setup_menu_init(void)
{
	extern int neogeo_bios_type;

	menu_total = 0;

	menu_item[menu_total] = ui_getstring (UI_inputgeneral); menu_action[menu_total++] = UI_DEFCODE;
	menu_item[menu_total] = ui_getstring (UI_inputspecific); menu_action[menu_total++] = UI_CODE;
#ifdef KAILLERA
	if (!kPlay)
	{
#endif
	menu_item[menu_total] = ui_getstring (UI_dipswitches); menu_action[menu_total++] = UI_SWITCH;
#ifdef KAILLERA
	}
#endif
	/* Determine if there are any analog controls */
	{
		struct InputPort *in;
		int num;

		in = Machine->input_ports;

		num = 0;
		while (in->type != IPT_END)
		{
			if (((in->type & 0xff) > IPT_ANALOG_START) && ((in->type & 0xff) < IPT_ANALOG_END)
					&& !(!options.cheat && (in->type & IPF_CHEAT)))
				num++;
			in++;
		}

		if (num != 0)
		{
			menu_item[menu_total] = ui_getstring (UI_analogcontrols); menu_action[menu_total++] = UI_ANALOG;
		}
		else
		{
			menu_item[menu_total] = ui_getstring (UI_custombuttons); menu_action[menu_total++] = UI_CUSTOM;
			menu_item[menu_total] = ui_getstring (UI_autofiredelay); menu_action[menu_total++] = UI_AUTOFIRE;
		}
	}

	/* Joystick calibration possible? */
	if ((osd_joystick_needs_calibration()) != 0)
	{
		menu_item[menu_total] = ui_getstring (UI_calibrate); menu_action[menu_total++] = UI_CALIBRATE;
	}

	menu_item[menu_total] = ui_getstring (UI_bookkeeping); menu_action[menu_total++] = UI_STATS;
	menu_item[menu_total] = ui_getstring (UI_gameinfo); menu_action[menu_total++] = UI_GAMEINFO;
	menu_item[menu_total] = ui_getstring (UI_history); menu_action[menu_total++] = UI_HISTORY;
	menu_item[menu_total] = ui_getstring (UI_command); menu_action[menu_total++] = UI_COMMAND;

	if (options.cheat)
	{
#ifdef KAILLERA
		if (!kPlay)
		{
#endif
		menu_item[menu_total] = ui_getstring (UI_cheat); menu_action[menu_total++] = UI_CHEAT;
#ifdef KAILLERA
		}
#endif
	}

	if (neogeo_bios_type == BIOS_TYPE_STANDARD)
	{
#ifdef KAILLERA
		if (!kPlay)
		{
#endif
		menu_item[menu_total] = ui_getstring (UI_memorycard); menu_action[menu_total++] = UI_MEMCARD;
#ifdef KAILLERA
		}
#endif
	}

#ifdef KAILLERA
	if (!kPlay)
	{
#endif
	menu_item[menu_total] = ui_getstring (UI_resetgame); menu_action[menu_total++] = UI_RESET;
#ifdef KAILLERA
	}
#endif
	menu_item[menu_total] = ui_getstring (UI_returntogame); menu_action[menu_total++] = UI_EXIT;
	menu_item[menu_total] = 0; /* terminate array */
}


static int setup_menu(struct osd_bitmap *bitmap, int selected)
{
	int sel,res=-1;

	if (selected == -1)
		sel = menu_lastselected;
	else
		sel = selected - 1;

	if (sel > SEL_MASK)
	{
		switch (menu_action[sel & SEL_MASK])
		{
			case UI_SWITCH:
				res = setdipswitches(bitmap, sel >> SEL_BITS);
				break;
			case UI_DEFCODE:
				res = setdefcodesettings(bitmap, sel >> SEL_BITS);
				break;
			case UI_CODE:
				res = setcodesettings(bitmap, sel >> SEL_BITS);
				break;
			case UI_ANALOG:
				res = settraksettings(bitmap, sel >> SEL_BITS);
				break;
			case UI_CALIBRATE:
				res = calibratejoysticks(bitmap, sel >> SEL_BITS);
				break;
			case UI_STATS:
				res = mame_stats(bitmap, sel >> SEL_BITS);
				break;
			case UI_GAMEINFO:
				res = displaygameinfo(bitmap, sel >> SEL_BITS);
				break;
			case UI_HISTORY:
				res = displayhistory(bitmap, sel >> SEL_BITS);
				break;
			case UI_COMMAND:
				res = displaycommand_ex(bitmap, sel >> SEL_BITS, 0);
				break;
			case UI_CHEAT:
				res = cheat_menu(bitmap, sel >> SEL_BITS, 0);
				break;
			case UI_MEMCARD:
				res = memcard_menu(bitmap, sel >> SEL_BITS);
				break;
			case UI_CUSTOM:
				res = setcustombutton(bitmap, sel >> SEL_BITS);
				change_port_mask();
				break;
			case UI_AUTOFIRE:
				res = setautofire(bitmap, sel >> SEL_BITS);
				change_port_mask();
				break;
		}

		if (res == -1)
		{
			menu_lastselected = sel;
			sel = -1;
		}
		else
			sel = (sel & SEL_MASK) | (res << SEL_BITS);

		return sel + 1;
	}


	ui_displaymenu(bitmap,menu_item,0,0,sel,0);

	if (input_ui_pressed_repeat(IPT_UI_DOWN,4))
		sel = (sel + 1) % menu_total;

	if (input_ui_pressed_repeat(IPT_UI_UP,4))
		sel = (sel + menu_total - 1) % menu_total;

	if (input_ui_pressed(IPT_UI_SELECT))
	{
		switch (menu_action[sel])
		{
			case UI_SWITCH:
			case UI_DEFCODE:
			case UI_CODE:
			case UI_ANALOG:
			case UI_CALIBRATE:
			case UI_STATS:
			case UI_GAMEINFO:
			case UI_HISTORY:
			case UI_COMMAND:
			case UI_CHEAT:
			case UI_MEMCARD:
			case UI_CUSTOM:
			case UI_AUTOFIRE:
				sel |= 1 << SEL_BITS;
				schedule_full_refresh();
				break;

			case UI_RESET:
				machine_reset();
				break;

			case UI_EXIT:
				menu_lastselected = 0;
				sel = -1;
				break;
		}
	}

	if (input_ui_pressed(IPT_UI_CANCEL) ||
			input_ui_pressed(IPT_UI_CONFIGURE))
	{
		menu_lastselected = sel;
		sel = -1;
	}

	if (sel == -1)
	{
		schedule_full_refresh();
	}

	return sel + 1;
}



/*********************************************************************

  start of On Screen Display handling

*********************************************************************/

static void displayosd(struct osd_bitmap *bitmap,const char *text,int percentage,int default_percentage)
{
	struct DisplayText dt[2];
	int avail;


	avail = (Machine->uiwidth / Machine->uifontwidth) * 19 / 20;

	draw3dbox(bitmap,(Machine->uiwidth - Machine->uifontwidth * avail) / 2,
			(Machine->uiheight - 7*Machine->uifontheight/2),
			avail * Machine->uifontwidth,
			3*Machine->uifontheight, 1);

	avail--;

	drawbar(bitmap,(Machine->uiwidth - Machine->uifontwidth * avail) / 2,
			(Machine->uiheight - 3*Machine->uifontheight),
			avail * Machine->uifontwidth,
			Machine->uifontheight,
			percentage,default_percentage);

	dt[0].text = text;
	dt[0].color = UI_COLOR_NORMAL;
	dt[0].x = (Machine->uiwidth - Machine->uifontwidth * strlen(text)) / 2;
	dt[0].y = (Machine->uiheight - 2*Machine->uifontheight) + 2;
	dt[1].text = 0; /* terminate array */
	displaytext(bitmap,dt,FONT_NORMAL);
}



static void onscrd_volume(struct osd_bitmap *bitmap,int increment,int arg)
{
	char buf[20];
	int attenuation;

	if (increment)
	{
		attenuation = osd_get_mastervolume();
		attenuation += increment;
		if (attenuation > 0) attenuation = 0;
		if (attenuation < -32) attenuation = -32;
		osd_set_mastervolume(attenuation);
	}
	attenuation = osd_get_mastervolume();

	sprintf(buf,"%s %3ddB", ui_getstring (UI_volume), attenuation);
	displayosd(bitmap,buf,100 * (attenuation + 32) / 32,100);
}

static void onscrd_mixervol(struct osd_bitmap *bitmap,int increment,int arg)
{
	static void *driver = 0;
	char buf[40];
	int volume,ch;
	int doallchannels = 0;
	int proportional = 0;


	if (code_pressed(KEYCODE_LSHIFT) || code_pressed(KEYCODE_RSHIFT))
		doallchannels = 1;
	if (!code_pressed(KEYCODE_LCONTROL) && !code_pressed(KEYCODE_RCONTROL))
		increment *= 5;
	if (code_pressed(KEYCODE_LALT) || code_pressed(KEYCODE_RALT))
		proportional = 1;

	if (increment)
	{
		if (proportional)
		{
			static int old_vol[MIXER_MAX_CHANNELS];
			float ratio = 1.0;
			int overflow = 0;

			if (driver != Machine->drv)
			{
				driver = (void *)Machine->drv;
				for (ch = 0; ch < MIXER_MAX_CHANNELS; ch++)
					old_vol[ch] = mixer_get_mixing_level(ch);
			}

			volume = mixer_get_mixing_level(arg);
			if (old_vol[arg])
				ratio = (float)(volume + increment) / (float)old_vol[arg];

			for (ch = 0; ch < MIXER_MAX_CHANNELS; ch++)
			{
				if (mixer_get_name(ch) != 0)
				{
					volume = ratio * old_vol[ch];
					if (volume < 0 || volume > 100)
						overflow = 1;
				}
			}

			if (!overflow)
			{
				for (ch = 0; ch < MIXER_MAX_CHANNELS; ch++)
				{
					volume = ratio * old_vol[ch];
					mixer_set_mixing_level(ch,volume);
				}
			}
		}
		else
		{
			driver = 0; /* force reset of saved volumes */

			volume = mixer_get_mixing_level(arg);
			volume += increment;
			if (volume > 100) volume = 100;
			if (volume < 0) volume = 0;

			if (doallchannels)
			{
				for (ch = 0;ch < MIXER_MAX_CHANNELS;ch++)
					mixer_set_mixing_level(ch,volume);
			}
			else
				mixer_set_mixing_level(arg,volume);
		}
	}
	volume = mixer_get_mixing_level(arg);

	if (proportional)
		sprintf(buf,"%s%s %3d%%", ui_getstring (UI_allchannels), ui_getstring (UI_relative), volume);
	else if (doallchannels)
		sprintf(buf,"%s%s %3d%%", ui_getstring (UI_allchannels), ui_getstring (UI_volume), volume);
	else
		sprintf(buf,"%s%s %3d%%",mixer_get_name(arg), ui_getstring (UI_volume), volume);
	displayosd(bitmap,buf,volume,mixer_get_default_mixing_level(arg));
}

static void onscrd_brightness(struct osd_bitmap *bitmap,int increment,int arg)
{
	char buf[20];
	int brightness;


	if (increment)
	{
		brightness = osd_get_brightness();
		brightness += 5 * increment;
		if (brightness < 0) brightness = 0;
		if (brightness > 100) brightness = 100;
		osd_set_brightness(brightness);
	}
	brightness = osd_get_brightness();

	sprintf(buf,"%s %3d%%", ui_getstring (UI_brightness), brightness);
	displayosd(bitmap,buf,brightness,100);
}

static void onscrd_gamma(struct osd_bitmap *bitmap,int increment,int arg)
{
	char buf[20];
	float gamma_correction;

	if (increment)
	{
		gamma_correction = osd_get_gamma();

		gamma_correction += 0.05 * increment;
		if (gamma_correction < 0.5) gamma_correction = 0.5;
		if (gamma_correction > 2.0) gamma_correction = 2.0;

		osd_set_gamma(gamma_correction);
	}
	gamma_correction = osd_get_gamma();

	sprintf(buf,"%s %1.2f", ui_getstring (UI_gamma), gamma_correction);
	displayosd(bitmap,buf,100*(gamma_correction-0.5)/(2.0-0.5),100*(1.0-0.5)/(2.0-0.5));
}

static void onscrd_overclock(struct osd_bitmap *bitmap,int increment,int arg)
{
	char buf[30];
	double overclock;
	int cpu, doallcpus = 0, oc;
	extern short cpu_clocks[MAX_CPU];

	if (code_pressed(KEYCODE_LSHIFT) || code_pressed(KEYCODE_RSHIFT))
		doallcpus = 1;
	if (!code_pressed(KEYCODE_LCONTROL) && !code_pressed(KEYCODE_RCONTROL))
		increment *= 5;
	if( increment )
	{
		overclock = timer_get_overclock(arg);
		overclock += 0.01 * increment;
		if (overclock < 0.01) overclock = 0.01;
		if (overclock > 2.0) overclock = 2.0;
		if( doallcpus )
			for( cpu = 0; cpu < cpu_gettotalcpu(); cpu++ )
				timer_set_overclock(cpu, overclock);
		else
			timer_set_overclock(arg, overclock);
	}

	oc = 100 * timer_get_overclock(arg) + 0.5;

	if( doallcpus )
	{
		int i;

		sprintf(buf,"%s %s %3d%%", ui_getstring (UI_allcpus), ui_getstring (UI_overclock), oc);
		for (i = 0; i < MAX_CPU; i++)
			cpu_clocks[i] = oc;
	}
	else
	{
		sprintf(buf,"%s %s%d %3d%%", ui_getstring (UI_overclock), ui_getstring (UI_cpu), arg, oc);
		cpu_clocks[arg] = oc;
	}
	displayosd(bitmap,buf,oc/2,100/2);
}

static void onscrd_transparency(struct osd_bitmap *bitmap,int increment,int arg)
{
	char buf[50];
	int transparency;


	if (increment)
	{
		transparency = options.ui_transparency;
		transparency += 5 * increment;
		if (transparency < 0) transparency = 0;
		if (transparency > 100) transparency = 100;
		options.ui_transparency = transparency;
		osd_set_brightness(osd_get_brightness());
	}
	transparency = options.ui_transparency;

	sprintf(buf,"%s %3d%%", ui_getstring (UI_transparency), transparency);
	displayosd(bitmap,buf,transparency,100);
}

#define MAX_OSD_ITEMS 30
static void (*onscrd_fnc[MAX_OSD_ITEMS])(struct osd_bitmap *bitmap,int increment,int arg);
static int onscrd_arg[MAX_OSD_ITEMS];
static int onscrd_total_items;

static void onscrd_init(void)
{
	int item,ch;


	osd_selected = 0;
	item = 0;

	onscrd_fnc[item] = onscrd_volume;
	onscrd_arg[item] = 0;
	item++;

	for (ch = 0;ch < MIXER_MAX_CHANNELS;ch++)
	{
		if (mixer_get_name(ch) != 0)
		{
			onscrd_fnc[item] = onscrd_mixervol;
			onscrd_arg[item] = ch;
			item++;
		}
	}

	for (ch = 0;ch < cpu_gettotalcpu();ch++)
	{
		onscrd_fnc[item] = onscrd_overclock;
		onscrd_arg[item] = ch;
		item++;
	}

	onscrd_fnc[item] = onscrd_brightness;
	onscrd_arg[item] = 0;
	item++;

	onscrd_fnc[item] = onscrd_gamma;
	onscrd_arg[item] = 0;
	item++;

	onscrd_fnc[item] = onscrd_transparency;
	onscrd_arg[item] = 0;
	item++;

	onscrd_total_items = item;
}

static int on_screen_display(struct osd_bitmap *bitmap, int selected)
{
	int increment, sel;


	if (selected == -1)
		sel = osd_lastselected;
	else
		sel = selected - 1;

	increment = 0;
	if (input_ui_pressed_repeat(IPT_UI_LEFT,8))
		increment = -1;
	if (input_ui_pressed_repeat(IPT_UI_RIGHT,8))
		increment = 1;
	if (input_ui_pressed_repeat(IPT_UI_DOWN,8))
		sel = (sel + 1) % onscrd_total_items;
	if (input_ui_pressed_repeat(IPT_UI_UP,8))
		sel = (sel + onscrd_total_items - 1) % onscrd_total_items;

	(*onscrd_fnc[sel])(bitmap,increment,onscrd_arg[sel]);

	osd_lastselected = sel;

	if (input_ui_pressed(IPT_UI_ON_SCREEN_DISPLAY))
	{
		sel = -1;

		schedule_full_refresh();
	}

	return sel + 1;
}

/*********************************************************************

  end of On Screen Display handling

*********************************************************************/


static void displaymessage(struct osd_bitmap *bitmap,const char *text,int trans)
{
	struct DisplayText dt[2];
	int avail;


	if (Machine->uiwidth < Machine->uifontwidth * strlen(text))
	{
		ui_displaymessagewindow(bitmap,text);
		return;
	}

	avail = strlen(text)+2;

	draw3dbox(bitmap,(Machine->uiwidth - Machine->uifontwidth * avail) / 2,
			Machine->uiheight - 3*Machine->uifontheight,
			avail * Machine->uifontwidth,
			2*Machine->uifontheight,
			trans);

	dt[0].text = text;
	dt[0].color = UI_COLOR_NORMAL;
	dt[0].x = (Machine->uiwidth - Machine->uifontwidth * strlen(text)) / 2;
	dt[0].y = Machine->uiheight - 5*Machine->uifontheight/2;
	dt[1].text = 0; /* terminate array */
	displaytext(bitmap, dt, FONT_NORMAL);
}


void CLIB_DECL usrintf_showmessage(const char *text,...)
{
	va_list arg;
	va_start(arg,text);
	vsprintf(messagetext,text,arg);
	va_end(arg);
	messagecounter = 2 * Machine->drv->frames_per_second;
}

void CLIB_DECL usrintf_showmessage_secs(int seconds, const char *text,...)
{
	va_list arg;
	va_start(arg,text);
	vsprintf(messagetext,text,arg);
	va_end(arg);
	messagecounter = seconds * Machine->drv->frames_per_second;
}


static void flash_all_values_and_buffers(void)
{
	int i;

	menu_lastselected = 0;
	osd_lastselected = 0;
	setup_selected = 0;
	osd_selected = 0;
	single_step = 0;
	scroll_reset = 0;
	show_input = 0;
#ifdef MAME_DEBUG
	show_profiler = 0;
	show_total_colors = 0;
#endif

	saveslot = 0;
	combosaveslot = 0;
	if (record_combo)
	{
		combo_record_end();
		record_combo = 0;
	}

	if (wave_recording)
	{
		wave_record_end();
		wave_recording = 0;
	}
	wave_recording = 0;

	memset(&playlist, 0, sizeof(struct playlist_t) * MAX_PLAYLIST);
	playlist_selected = 0;

	memset(&neojuke, 0, sizeof(struct jukebox_t));
	neojuke.playlist_bak = -1;
	soundtest_selected = 1;
	jukebox = 0;
	line_volume[0] = 0;
	line_volume[1] = 0;
	line_counter[0] = 0;
	line_counter[1] = 0;
	volume_delay = 0;
	random_play = 0;
	playing_time = 0;
	playing_wait = 0;
	nosound_time = 0;
	randomplay_time = 0;
	fadeout_timing = 0;
	fadeout_frame = 600;
	fadeout_counter = 0;

	if (history_buf)
	{
		MyFree(history_buf);
		history_buf = 0;
	}
	history_scroll = 0;

	if (commandlist_buf)
	{
		MyFree(commandlist_buf);
		commandlist_buf = 0;
	}
	command_scroll = 0;
	command_lastselected = 0;

	free_datafile();

	messagecounter = 0;
	memset(messagetext, 0, 80);

	command_counter;
	memset(command_buf, 0, 80);

	for (i = 0 ; i < 0x300 >> 1 ; i++)
		skip_object[i] = 0;

	front_layer = 0;
	alt_backcolor = 0;
	object_no = 0;
	object_selected = 1;
	object = 0;
	object_cpu = 0;

	command_sc = 0;
	command_sel = 1;
	command_load = 1;

	cheat_sc = 0;
	cheat_sel = 1;

	show_time = 0;

	wave_minutes = 0;
	wave_seconds = 0;
}

/**************************************************************************/

static int atohex(char *s)
{
	int res = 0;

	while (*s)
	{
		res *= 0x10;
		if ((*s >= 'A' && *s <= 'F')
		||  (*s >= 'a' && *s <= 'f'))
		{
			res += (toupper(*s) - 'A') + 10;
		}
		else
		{
			res += *s - '0';
		}
		s++;
	}

	return res;
}


static int playlist_open(void)
{
	FILE *fp;
	char name[256];
	char tmpbuf[256], tmpnum[5] = { 0, 0, 0, 0, 0 };
	UINT16 tmpcode;
	int i, j, loaded_playlist, name_length;

	if (neojuke.playlist)
		return 1;

	if (neojuke.playlist_bak >= 0)
		return 0;

	name_length = strlen(Machine->gamedrv->name);

	sprintf(name, sound_filename);
	fp = fopen(name, "r");
	if (fp == NULL)
		return 0;

	loaded_playlist = 0;

	/* Ex.: kof98:0721:I[vjO*/
	while (1)
	{
		if (fgets(tmpbuf, 255, fp) == NULL) break;

		if (tmpbuf[name_length] != ':') continue;
		if (strncmp (tmpbuf, Machine->gamedrv->name, name_length) != 0) continue;
		if (tmpbuf[0] == ';') continue;

		for (i = 0 , j = name_length + 1 ; i <= 3 ; i++ , j++ )
		{
			tmpnum[i] = tmpbuf[j];
		}

		tmpcode = atohex(tmpnum);
		if (tmpcode == 0)
			break;

		i = 0;
		j = name_length + 6;

		while (tmpbuf[j] != '\n')
		{
			playlist[loaded_playlist].desc[i] = tmpbuf[j];
			i++;
			j++;
		}

		playlist[loaded_playlist].desc[++i] = '\n';

		playlist[loaded_playlist].code = tmpcode;

		loaded_playlist++;
		if (loaded_playlist == MAX_PLAYLIST)
			break;
	}

	fclose(fp);

	playlist_selected = 0;
	soundtest_selected = 1;

	return loaded_playlist;
}


static void jukebox_reset(void)
{
	if (neojuke.playlist)
	{
		neojuke.code[0] = playlist[playlist_selected].code >> 8;
		neojuke.code[1] = playlist[playlist_selected].code & 0xff;
	}
	else
	{
		neojuke.code[0]   = neojuke.bank[0] >> 8;
		neojuke.code[1]   = neojuke.bank[0] & 0xff;
	}

	neojuke.curr_bank = 0;
	neojuke.level[0]  = 0;
	neojuke.level[1]  = 0;
	neojuke.que[0]    = neojuke.start;
	neojuke.que[1]    = neojuke.stop2;
	neojuke.que[2]    = neojuke.stop1;
	neojuke.counter   = 3;
	neojuke.command   = 0;
	neojuke.status    = JUKE_STARTWAIT;
	line_volume[0]    = 0;
	line_volume[1]    = 0;
	line_counter[0]   = 0;
	line_counter[1]   = 0;
}


static void jukebox_start(void)
{
	int i;

	cpu_set_halt_line(0, 1);
	save_contexts(OSD_FILETYPE_STATE, TEMPORARY_FILE);
	neogeo_save_screen();

	in_jukebox = 1;
	volume_delay = 20;

	if (wave_recording)
	{
		wave_record_end();
		wave_recording = 0;
	}

	switch (Machine->gamedrv->sound_type)
	{
		case SOUND_SNK:
			neojuke.start = 0x07;
			neojuke.stop1 = 0x00;
			neojuke.stop2 = 0x03;
			break;

		case SOUND_ADK:
		case SOUND_DATAEAST:
		case SOUND_TECHNOS:
		case SOUND_TECMO:
			neojuke.start = 0x03;
			neojuke.stop1 = 0x00;
			neojuke.stop2 = 0x03;
			break;

		case SOUND_NAZKA:
			neojuke.start = 0x00;
			neojuke.stop1 = 0x00;
			neojuke.stop2 = 0x03;
			break;

		case SOUND_FACE:
			neojuke.start = 0x20;
			neojuke.stop1 = 0x00;
			neojuke.stop2 = 0x03;
			break;
	}

	memcpy(neojuke.bank, Machine->gamedrv->sound_bank, 8 * sizeof(UINT16));
	for (i = 0; i < 8; i++)
	{
		if (!neojuke.bank[i])
			break;
	}
	neojuke.num_bank  = i;

	if (neojuke.playlist_bak == -1)
		neojuke.playlist_bak = playlist_open();

	line_volume[0] = 0;
	line_volume[1] = 0;
	line_counter[0] = 0;
	line_counter[1] = 0;

	jukebox_reset();
}


static void jukebox_stop(void)
{
	neojuke.level[0] = 0;
	neojuke.level[1] = 0;
	neojuke.que[0]   = neojuke.start;
	neojuke.que[1]   = neojuke.stop2;
	neojuke.que[2]   = neojuke.stop1;
	neojuke.counter  = 3;

	if (wave_recording)
	{
		wave_record_end();
		wave_recording = 0;
	}
	neojuke.status = JUKE_STOPWAIT;

	in_jukebox = 0;
}


static int get_level(UINT16 val)
{
	int i;

	UINT16 level_table[64] =
	{
		  128,   138,   149,   161,   174,   188,   203,   219,
		  237,   256,   276,   299,   322,   348,   376,   406,
		  439,   474,   512,   553,   597,   645,   697,   753,
		  813,   878,   948,  1024,  1106,  1195,  1290,  1393,
		 1505,  1625,  1756,  1896,  2048,  2212,  2389,  2580,
		 2787,  3010,  3251,  3511,  3792,  4096,  4424,  4778,
		 5161,  5574,  6020,  6502,  7023,  7585,  8192,  8848,
		 9556, 10321, 11148, 12040, 13004, 14045, 15170, 16384
	};

	val >>= 1;

	for (i= 0; i < 64; i++)
	{
		if (val < level_table[i])
			break;
	}

	return i;
}


static void volume_window(struct osd_bitmap *bitmap)
{
	int i, j, x, y;
	static int delay[2] = { 0, 0 };
	static int counter[2];

	x = (Machine->uiwidth - 212) / 2 + 15;
	y = (Machine->uiheight - 7*Machine->uifontheight/2) + Machine->uifontheight / 2 + 2;

	draw3dbox(bitmap, (Machine->uiwidth - 212) / 2,
			(Machine->uiheight - 7*Machine->uifontheight/2) + 2,
			212,
			3*Machine->uifontheight - 4, 1);

	ui_text(bitmap, "L", x - 10, y - 2);
	ui_text(bitmap, "R", x - 10, y + 11);

	for (i = 0; i < 2; i++)
	{
		for (j = 0; j < 42; j++)
			ui_drawbox2(bitmap, x + j * 3, y + (i * 12), 2, 8, VOLUME_DARKGREEN);
		for (j = 42; j < 58; j++)
			ui_drawbox2(bitmap, x + j * 3, y + (i * 12), 2, 8, VOLUME_DARKYELLOW);
		for (j = 58; j < 64; j++)
			ui_drawbox2(bitmap, x + j * 3, y + (i * 12), 2, 8, VOLUME_DARKRED);

		if (volume_delay)
		{
			counter[0] = 0;
			counter[1] = 0;
			volume_delay--;
			ui_drawbox2(bitmap, x - 1, y + (i * 12), 1, 8, BUTTON_COLOR_BLUE);
		}
		else
		if (wave_recording == 0)
		{
			extern UINT16 current_level[2];
			int level;

			level = get_level(current_level[i]);

			if (delay[i]-- <= 0)
			{
				if (neojuke.level[i] > 1)
					neojuke.level[i] -= 2;
				else if (neojuke.level[i] == 1)
					neojuke.level[i] = 0;
				delay[i] = 1;
			}

			if (neojuke.level[i] < level)
			{
				neojuke.level[i] = level;
				delay[i] = 1;
			}

			if (line_volume[i] - counter[i] < neojuke.level[i])
			{
				line_volume[i] = neojuke.level[i];
				line_counter[i] = 0;
				counter[i] = 0;
			}
			else if (line_volume[i] - counter[i] > neojuke.level[i])
			{
				line_counter[i]++;
				counter[i] = (int)(line_counter[i] * line_counter[i] * 0.005);
				if (line_volume[i] - counter[i] < 0)
				{
					line_volume[i] = 0;
					line_counter[i] = 0;
					counter[i] = 0;
				}
			}

			ui_drawbox2(bitmap, x + (line_volume[i] - counter[i]) * 3 - 1, y + (i * 12), 1, 8, BUTTON_COLOR_BLUE);

			if (neojuke.level[i] <= 42)
			{
			for (j = 0; j < neojuke.level[i]; j++)
				ui_drawbox2(bitmap, x + j * 3, y + (i * 12), 2, 8, BUTTON_COLOR_GREEN);
			}
			else if (neojuke.level[i] <= 58)
			{
			for (j = 0; j < 42; j++)
				ui_drawbox2(bitmap, x + j * 3, y + (i * 12), 2, 8, BUTTON_COLOR_GREEN);
			for (j = 42; j < neojuke.level[i]; j++)
				ui_drawbox2(bitmap, x + j * 3, y + (i * 12), 2, 8, BUTTON_COLOR_YELLOW);
			}
			else 
			{
			for (j = 0; j < 42; j++)
				ui_drawbox2(bitmap, x + j * 3, y + (i * 12), 2, 8, BUTTON_COLOR_GREEN);
			for (j = 42; j < 58; j++)
				ui_drawbox2(bitmap, x + j * 3, y + (i * 12), 2, 8, BUTTON_COLOR_YELLOW);
			for (j = 58; j < neojuke.level[i]; j++)
				ui_drawbox2(bitmap, x + j * 3, y + (i * 12), 2, 8, BUTTON_COLOR_RED);
			}
		}
	}
}


static void jukebox_playlist(struct osd_bitmap *bitmap)
{
	static const char *menu_item[MAX_PLAYLIST + 3];
	const char *menu_subitem[MAX_PLAYLIST + 3];
	static int total;
	int i, sel, visible, arrowise = 0;
	float fadeout_second;
	char buf[256];

	sel = playlist_selected;
	visible = (Machine->uiheight / ROWHEIGHT) - 6;

	for (i = 0 ; i < MAX_PLAYLIST + 3 ; i++)
		menu_subitem[i] = 0;

	if (neojuke.menu_init == 0)
	{
		for (i = 0 ; i < MAX_PLAYLIST + 3 ; i++)
			menu_item[i] = 0;

		for (total = 0; total < neojuke.playlist; total++)
			menu_item[total] = playlist[total].desc;

		if (random_play == 0)
		{
			menu_item[total++]   = ui_getstring(UI_fadeouttimer);
			menu_item[total]     = ui_getstring(UI_gotomanualcode);
		}
		else
		{
			playing_time = 0;
			playing_wait = 0;
			nosound_time = 0;
			randomplay_time = options.random_time * 60;	// t[֕ύXB
			for (i = 0 ; i < neojuke.playlist ; i++)
				playlist[i].flag = 0;
			if (randomplay_time < 6000)
				fadeout_timing = randomplay_time * 4 / 5;
			else
				fadeout_timing = randomplay_time - 1200;	// 20b(1200t[)̃tF[hAEg
		}

		menu_item[total + 1] = 0;	/* terminate array */
		total++;
		neojuke.menu_init = 1;
		fadeout_counter = 0;
	}

	if (random_play == 0)
	{
		fadeout_second = (float)(fadeout_frame) / 60;

		sprintf(buf, "%2.1f%s", fadeout_second, ui_getstring(UI_fadeoutsecond));

		menu_subitem[total - 2] = (const char *)buf;

		if (fadeout_frame > 60)
			arrowise |= 1;
		if (fadeout_frame < 1200)
			arrowise |= 2;
	}

	neogeo_restore_screen();

	ui_displaymenu(bitmap, menu_item, menu_subitem, 0, sel, arrowise);

	volume_window(bitmap);

	if (fadeout_counter)
	{
		fadeout_counter++;

		if (fadeout_counter == fadeout_frame && random_play == 0)
		{
			neojuke.que[0]  = neojuke.start;
			neojuke.que[1]  = neojuke.stop2;
			neojuke.que[2]  = neojuke.stop1;
			neojuke.counter = 3;

			if (neojuke.status == JUKE_RECORDING)
			{
				wave_record_end();
				wave_recording = 0;
			}

			neojuke.status = JUKE_IDLE;
		}
	}
	
	if (random_play)
	{
		if (playing_time == 0)	// tJn
		{
			int flag = 0;

			for (i = 0; i < neojuke.playlist; i++)
			{
				if (playlist[i].flag == 0)
					flag = 1;
			}
			
			if (flag == 0)
			{
				for (i = 0; i < neojuke.playlist; i++)
				{
					playlist[i].flag = 0;
				}
			}

			while (1)
			{
				sel = rand();
				if (sel < neojuke.playlist && playlist[sel].flag == 0)
					break;
			}

			playlist_selected = sel;
			playlist[sel].flag = 1;

			neojuke.level[0] = 0;
			neojuke.level[1] = 0;
			line_volume[0] = 0;
			line_volume[1] = 0;
			line_counter[0] = 0;
			line_counter[1] = 0;

			neojuke.code[0] = playlist[sel].code >> 8;
			neojuke.code[1] = playlist[sel].code & 0xff;
			neojuke.que[0]  = neojuke.code[1];
			neojuke.que[1]  = neojuke.code[0];
			neojuke.counter = 2;
			neojuke.status = JUKE_PLAYING;
		}

		playing_time++;

		if (playing_time == fadeout_timing)	// tF[hAEgJn
		{
			fadeout_counter = 1;
			fadeout_frame = randomplay_time - fadeout_timing;
		}

		if (playing_time == randomplay_time)	// tI
		{
			if (playing_wait == 0)
			{
				neojuke.que[0]  = neojuke.start;
				neojuke.que[1]  = neojuke.stop2;
				neojuke.que[2]  = neojuke.stop1;
				neojuke.counter = 3;
				neojuke.status = JUKE_IDLE;
			}

			if (playing_wait == 120)
			{
				playing_time = 0;
				playing_wait = 0;
				nosound_time = 0;
			}
			else
			{
				playing_time--;
				playing_wait++;
			}
		}

		if (neojuke.level[0] == 0 && neojuke.level[1] == 0)	// ԂƂ̏B
		{
			nosound_time++;
			if (nosound_time == 300)
			{
				neojuke.que[0]  = neojuke.start;
				neojuke.que[1]  = neojuke.stop2;
				neojuke.que[2]  = neojuke.stop1;
				neojuke.counter = 3;
				neojuke.status = JUKE_IDLE;
				playing_time = 0;
				playing_wait = 0;
				nosound_time = 0;
			}
		}
		else
		{
			nosound_time = 0;
		}
		
		if (code_pressed_memory_repeat(KEYCODE_S, 8))
		{
			neojuke.level[0] = 0;
			neojuke.level[1] = 0;
			line_volume[0] = 0;
			line_volume[1] = 0;
			line_counter[0] = 0;
			line_counter[1] = 0;

			sel = 0;
			neojuke.playlist_bak = neojuke.playlist;
			neojuke.playlist = 0;
			playing_time = 0;
			playing_wait = 0;
			nosound_time = 0;
			random_play = 0;
			jukebox_reset();
			return;
		}

		if (code_pressed_memory_repeat(KEYCODE_N, 8))
		{
			neojuke.que[0]  = neojuke.start;
			neojuke.que[1]  = neojuke.stop2;
			neojuke.que[2]  = neojuke.stop1;
			neojuke.counter = 3;
			neojuke.status = JUKE_IDLE;
			playing_time = 0;
			playing_wait = 0;
			nosound_time = 0;
		}

		if (code_pressed(KEYCODE_SPACE))
		{
			sprintf(buf, "%s  %s  %s",
				ui_getstring(UI_jukerandomstop_exp),
				ui_getstring(UI_jukenext_exp),
				ui_getstring(UI_jukecode_exp));
			displaymessage(bitmap, buf, 0);
		}

		if (code_pressed(KEYCODE_C))
		{
			sprintf(buf, "%s : %02X  %s : %02X",
				ui_getstring(UI_code1),
				playlist[playlist_selected].code >> 8,
				ui_getstring(UI_code2),
				playlist[playlist_selected].code & 0xff);
			displaymessage(bitmap, buf, 0);
		}
	}
	else
	{
		if (code_pressed_memory_repeat(KEYCODE_S, 8))
		{
			neojuke.que[0]  = neojuke.start;
			neojuke.que[1]  = neojuke.stop2;
			neojuke.que[2]  = neojuke.stop1;
			neojuke.counter = 3;

			if (neojuke.status == JUKE_RECORDING)
			{
				wave_record_end();
				wave_recording = 0;
			}

			neojuke.status = JUKE_IDLE;
		}

		if (code_pressed_memory_repeat(KEYCODE_F, 8))
		{
			if (neojuke.status == JUKE_PLAYING || neojuke.status == JUKE_RECORDING)
			{
				fadeout_counter = 1;
			}
		}

		if (neojuke.status != JUKE_RECORDING && neojuke.status != JUKE_RECORDWAIT)
		{
			if (code_pressed(KEYCODE_SPACE))
			{
				sprintf(buf, "%s  %s  %s  %s",
					ui_getstring(UI_jukestop_exp2),
					ui_getstring(UI_jukerecord_exp2),
					ui_getstring(UI_jukecode_exp2),
					ui_getstring(UI_jukefadeout_exp2));
				displaymessage(bitmap, buf, 0);
			}

			if (code_pressed_memory_repeat(KEYCODE_R, 8))
			{
				if (sel < (total - 2))
				{
					neojuke.que[0]  = neojuke.code[1];
					neojuke.que[1]  = neojuke.code[0];
					neojuke.que[2]  = neojuke.start;
					neojuke.que[3]  = neojuke.stop2;
					neojuke.que[4]  = neojuke.stop1;
					neojuke.counter = 5;
					neojuke.status  = JUKE_RECORDWAIT;
				}
			}

			if (code_pressed(KEYCODE_C) && sel < total - 2)
			{
				sprintf(buf, "%s : %02X  %s : %02X",
					ui_getstring(UI_code1),
					playlist[sel].code >> 8,
					ui_getstring(UI_code2),
					playlist[sel].code & 0xff);
				displaymessage(bitmap, buf, 0);
			}

			if (input_ui_pressed_repeat(IPT_UI_SELECT, 8))
			{
				neojuke.level[0] = 0;
				neojuke.level[1] = 0;
				line_volume[0] = 0;
				line_volume[1] = 0;
				line_counter[0] = 0;
				line_counter[1] = 0;

				if (sel == (total - 1))
				{
					sel = 0;
					neojuke.playlist_bak = neojuke.playlist;
					neojuke.playlist = 0;
					jukebox_reset();
					return;
				}
				else if (sel == (total - 2) && (neojuke.status == JUKE_PLAYING || neojuke.status == JUKE_RECORDING))
				{
					fadeout_counter = 1;
				}
				else
				{
					neojuke.que[0]  = neojuke.code[1];
					neojuke.que[1]  = neojuke.code[0];
					neojuke.counter = 2;
					neojuke.status = JUKE_PLAYING;
				}
			}
		}
		else
		{
			sprintf(buf, "%s  %s",
				ui_getstring(UI_jukestop_exp),
				ui_getstring(UI_jukefadeout_exp));
			displaymessage(bitmap, buf, 0);
		}

		if (input_ui_pressed_repeat(IPT_UI_DOWN, 8))
		{
			sel = (sel + 1) % total;
			neojuke.code[0] = playlist[sel].code >> 8;
			neojuke.code[1] = playlist[sel].code & 0xff;
		}

		if (input_ui_pressed_repeat(IPT_UI_UP, 8))
		{
			sel = (sel + total - 1) % total;
			neojuke.code[0] = playlist[sel].code >> 8;
			neojuke.code[1] = playlist[sel].code & 0xff;
		}

		if (input_ui_pressed_repeat(IPT_UI_LEFT, 8))
		{
			sel -= (visible - 1);
			if (sel < 0)	sel = 0;
			neojuke.code[0] = playlist[sel].code >> 8;
			neojuke.code[1] = playlist[sel].code & 0xff;
		}

		if (input_ui_pressed_repeat(IPT_UI_RIGHT, 8))
		{
			sel += (visible - 1);
			if (sel >= total)	sel = total - 1;
			neojuke.code[0] = playlist[sel].code >> 8;
			neojuke.code[1] = playlist[sel].code & 0xff;
		}

		if (sel == total - 2)
		{
			if (code_pressed_memory_repeat(KEYCODE_PGDN, 8) && fadeout_counter == 0)
			{
				if (fadeout_frame > 60)
					fadeout_frame -= 30;
			}

			if (code_pressed_memory_repeat(KEYCODE_PGUP, 8) && fadeout_counter == 0)
			{
				if (fadeout_frame < 1200)
					fadeout_frame += 30;
			}
		}
	}

	playlist_selected = sel;

}


static void jukebox_playcommand(struct osd_bitmap *bitmap)
{
	int i, sel, menu_type, total, arrowise = 0;
	static const char *bank_name[8];
	static const char *command_name[8];
	static const char *menu_item[2][9];
	static const char *menu_subitem[10];
	char buf[2][4], buf2[80];

	if (neojuke.status == JUKE_STARTWAIT)
	{
		for (i = 0 ; i <= 7 ; i++)
		{
			bank_name[i] = ui_getstring(UI_bgm + i);
		}
	
		command_name[0] = ui_getstring(UI_jukeplay);
		command_name[1] = ui_getstring(UI_jukestop);
		command_name[2] = ui_getstring(UI_jukerecord);

		menu_item[0][0] = menu_item[1][0] = ui_getstring(UI_jukebox);
		menu_item[0][1] = ui_getstring(UI_bank);
		menu_item[0][2] = ui_getstring(UI_code1);
		menu_item[0][3] = ui_getstring(UI_code2);
		menu_item[0][4] = ui_getstring(UI_jukecommand);
		menu_item[0][5] = ui_getstring(UI_returntogame);
		menu_item[1][1] = ui_getstring(UI_jukestop);
		menu_item[0][6] = menu_item[0][7] = menu_item[0][8] = menu_item[1][2] = 0;	/* terminate array */
	}

	for (i = 0; i < 10; i++)
		menu_subitem[i] = 0;

	if (neojuke.status == JUKE_RECORDING || neojuke.status == JUKE_RECORDWAIT)
		menu_type = 1;
	else
		menu_type = 0;

	if (!menu_type)
	{
		sprintf(buf[0], "%02X", neojuke.code[0]);
		sprintf(buf[1], "%02X", neojuke.code[1]);

		menu_subitem[1] = bank_name[neojuke.curr_bank];
		menu_subitem[2] = (const char *)buf[0];
		menu_subitem[3] = (const char *)buf[1];
		menu_subitem[4] = command_name[neojuke.command];

		sel = soundtest_selected;

		if (neojuke.playlist_bak)
		{
			menu_item[0][5] = ui_getstring(UI_gotoplaylist);
			menu_item[0][6] = ui_getstring(UI_gotorandomplay);
			if (options.only_jukebox == 0)
			{
				menu_item[0][7] = ui_getstring(UI_returntogame);
				total = 8;
			}
			else
			{
				total = 7;
			}
		}
		else
		{
			if (options.only_jukebox == 0)
			{
				menu_item[0][5] = ui_getstring(UI_returntogame);
				menu_item[0][6] = 0;
				total = 6;
			}
			else
			{
				menu_item[0][5] = 0;
				total = 5;
			}
		}

		switch (sel)
		{
			case 1:
				if (neojuke.curr_bank > 0)
					arrowise |= 1;
				if (neojuke.curr_bank < neojuke.num_bank - 1)
					arrowise |= 2;
				break;

			case 2:
			case 3:
				arrowise = 3;
				break;

			case 4:
				if (neojuke.command > 0)
					arrowise |= 1;
				if (neojuke.command < 2)
					arrowise |= 2;
				break;
		}
	}
	else
	{
		sel = 1;
		total = 2;
	}

	neogeo_restore_screen();

	ui_displaymenu(bitmap, menu_item[menu_type], menu_subitem, 0, sel, arrowise);

	volume_window(bitmap);

	if (!menu_type)
	{
		if (input_ui_pressed_repeat(IPT_UI_SELECT, 8))
		{
			switch (sel)
			{
				case 1: // bank
				case 2: // code1
				case 3: // code2
					neojuke.que[0]  = neojuke.code[1];
					neojuke.que[1]  = neojuke.code[0];
					neojuke.counter = 2;
					neojuke.status = JUKE_PLAYING;
					break;

				case 4: // command
					neojuke.level[0] = 0;
					neojuke.level[1] = 0;
					line_volume[0] = 0;
					line_volume[1] = 0;
					line_counter[0] = 0;
					line_counter[1] = 0;
					switch (neojuke.command)
					{
						case 0:  // play
							neojuke.que[0]  = neojuke.code[1];
							neojuke.que[1]  = neojuke.code[0];
							neojuke.counter = 2;
							neojuke.status = JUKE_PLAYING;
							break;

						case 1: // stop
							neojuke.que[0]  = neojuke.start;
							neojuke.que[1]  = neojuke.stop2;
							neojuke.que[2]  = neojuke.stop1;
							neojuke.counter = 3;
							if (neojuke.status == JUKE_RECORDING)
							{
								wave_record_end();
								wave_recording = 0;
							}
							neojuke.status = JUKE_IDLE;
							break;

						case 2: // record
							neojuke.que[0]  = neojuke.code[1];
							neojuke.que[1]  = neojuke.code[0];
							neojuke.que[2]  = neojuke.start;
							neojuke.que[3]  = neojuke.stop2;
							neojuke.que[4]  = neojuke.stop1;
							neojuke.counter = 5;
							neojuke.status  = JUKE_RECORDWAIT;
							break;
					}
					break;

				case 5: // exit / playlist
					if (total == 6)
					{
						jukebox = 0;
						jukebox_stop();
					}
					else
					{
						neojuke.playlist = neojuke.playlist_bak;
						neojuke.playlist_bak = 0;
						neojuke.menu_init = 0;
						random_play = 0;
						jukebox_reset();
					}
					break;

				case 6: // random play
					neojuke.playlist = neojuke.playlist_bak;
					neojuke.playlist_bak = 0;
					neojuke.menu_init = 0;
					random_play = 1;
					jukebox_reset();
					break;

				case 7: // exit
					jukebox = 0;
					jukebox_stop();
					break;
			}
		}

		if (input_ui_pressed_repeat(IPT_UI_LEFT, 8))
		{
			switch (sel)
			{
				case 1:
					if (neojuke.curr_bank > 0)
					{
						neojuke.curr_bank--;
						neojuke.code[0] = neojuke.bank[neojuke.curr_bank] >> 8;
						neojuke.code[1] = neojuke.bank[neojuke.curr_bank] & 0xff;
					}
					break;

				case 2:
				case 3:
					neojuke.code[sel - 2]--;
					neojuke.code[sel - 2] &= 0xff;
					break;

				case 4:
					if (neojuke.command > 0)
						neojuke.command--;
					break;
			}
		}

		if (input_ui_pressed_repeat(IPT_UI_RIGHT, 8))
		{
			switch (sel)
			{
				case 1:
					if (neojuke.curr_bank < neojuke.num_bank - 1)
					{
						neojuke.curr_bank++;
						neojuke.code[0] = neojuke.bank[neojuke.curr_bank] >> 8;
						neojuke.code[1] = neojuke.bank[neojuke.curr_bank] & 0xff;
					}
					break;

				case 2:
				case 3:
					neojuke.code[sel - 2]++;
					neojuke.code[sel - 2] &= 0xff;
					break;

				case 4:
					if (neojuke.command < 2)
						neojuke.command++;
					break;
			}
		}

		if (sel == 2 || sel == 3)
		{
			if (code_pressed_memory_repeat(KEYCODE_PGDN, 8))
			{
				neojuke.code[sel - 2] -= 0x10;
				neojuke.code[sel - 2] &= 0xff;
			}

			if (code_pressed_memory_repeat(KEYCODE_PGUP, 8))
			{
				neojuke.code[sel - 2] += 0x10;
				neojuke.code[sel - 2] &= 0xff;
			}
		}

		if (input_ui_pressed_repeat(IPT_UI_UP, 8))
		{
			if (sel > 0) sel--;
			if (sel == 0) sel = total - 1;
		}

		if (input_ui_pressed_repeat(IPT_UI_DOWN, 8))
		{
			if (sel < total) sel++;
			if (sel == total) sel = 1;
		}

		if (code_pressed_memory_repeat(KEYCODE_S, 8))
		{
			neojuke.que[0]  = neojuke.start;
			neojuke.que[1]  = neojuke.stop2;
			neojuke.que[2]  = neojuke.stop1;
			neojuke.counter = 3;
			if (neojuke.status == JUKE_RECORDING)
			{
				wave_record_end();
				wave_recording = 0;
			}
			neojuke.status = JUKE_IDLE;
		}

		if (code_pressed_memory_repeat(KEYCODE_R, 8))
		{
			neojuke.que[0]  = neojuke.code[1];
			neojuke.que[1]  = neojuke.code[0];
			neojuke.que[2]  = neojuke.start;
			neojuke.que[3]  = neojuke.stop2;
			neojuke.que[4]  = neojuke.stop1;
			neojuke.counter = 5;
			neojuke.status  = JUKE_RECORDWAIT;
		}

		if (code_pressed(KEYCODE_SPACE))
		{
			sprintf(buf2, "%s  %s",
				ui_getstring(UI_jukestop_exp),
				ui_getstring(UI_jukerecord_exp));
			displaymessage(bitmap, buf2, 0);
		}
	}
	else
	{
		if (input_ui_pressed_repeat(IPT_UI_SELECT, 8) || code_pressed_memory_repeat(KEYCODE_S, 8))
		{
			neojuke.que[0]  = neojuke.start;
			neojuke.que[1]  = neojuke.stop2;
			neojuke.que[2]  = neojuke.stop1;
			neojuke.counter = 3;
			if (neojuke.status == JUKE_RECORDING)
			{
				wave_record_end();
				wave_recording = 0;
			}
			neojuke.status = JUKE_IDLE;
		}

		if (code_pressed(KEYCODE_SPACE))
		{
			sprintf(buf2, "%s", ui_getstring(UI_jukestop_exp));
			displaymessage(bitmap, buf2, 0);
		}
	}

	soundtest_selected = sel;

}

static void object_operation_start(void)
{
	save_contexts(OSD_FILETYPE_STATE, TEMPORARY_FILE);
//	neogeo_save_screen();
	object_selected = 1;
	if (paused == 0)
	{
		cpu_set_halt_line(0, 1);
		osd_sound_enable(0);
		object_cpu = 1;
	}
}

static void object_operation_end(void)
{
	if (object_cpu)
	{
		if (paused == 0)	osd_sound_enable(1);
		cpu_set_halt_line(0, 0);
		object_cpu = 0;
	}
	neogeo_restore_screen();
	restore_contexts(OSD_FILETYPE_STATE, TEMPORARY_FILE);
	object = 0;
}

static void object_operation(struct osd_bitmap *bitmap)
{
	int i, sel, object_number = 0, arrowise = 0;
	static const char *menu_item[8];
	static const char *menu_subitem[10];
	char buf[4][5];

	neogeo_save_screen();
	
	for (i = 0 ; i <= 6 ; i++)	menu_item[i] = ui_getstring(UI_object_operation + i);

	menu_item[7] = ui_getstring(UI_returntogame);
	menu_item[8] = 0;	/* terminate array */

	for (i = 0; i < 10; i++)
		menu_subitem[i] = 0;

	object_number = count_object_number();

	if (object_no > object_number)	object_no = object_number;

	sprintf(buf[0], "%d", object_no);
	menu_subitem[1] = (const char *)buf[0];

	sprintf(buf[1], "%s", skip_object[object_no] ? ui_getstring(UI_object_off) : ui_getstring(UI_object_on));
	menu_subitem[2] = (const char *)buf[1];

	sprintf(buf[2], "%s", front_layer ? ui_getstring(UI_object_off) : ui_getstring(UI_object_on));
	menu_subitem[3] = (const char *)buf[2];

	sprintf(buf[3], "%s", alt_backcolor ? ui_getstring(UI_object_on) : ui_getstring(UI_object_off));
	menu_subitem[4] = (const char *)buf[3];

	sel = object_selected;

	switch (sel)
	{
		case 1:
			if (object_no > 0)
				arrowise |= 1;
			if (object_no < object_number)
				arrowise |= 2;
			if (object_number == 0)
				arrowise = 0;
			break;

		case 2:
			if (skip_object[object_no] == 0)
				arrowise = 2;
			if (skip_object[object_no] == 1)
				arrowise = 1;
			if (object_number == 0)
				arrowise = 0;
			break;

		case 3:
			if (front_layer == 0)
				arrowise = 2;
			if (front_layer == 1)
				arrowise = 1;
			break;

		case 4:
			if (alt_backcolor == 0)
				arrowise = 2;
			if (alt_backcolor == 1)
				arrowise = 1;
			break;

		default:
			break;
	}

	neogeo_restore_screen();

	ui_displaymenu(bitmap, menu_item, menu_subitem, 0, sel, arrowise);

//	erase_screen(bitmap);

	if (input_ui_pressed_repeat(IPT_UI_SELECT, 8))
	{
		switch (sel)
		{
			case 1:
			case 2:
				if (object_number == 0)	break;
				skip_object[object_no] ^= 1;
				break;

			case 3:
				front_layer ^= 1;
				break;

			case 4:
				alt_backcolor ^= 1;
				alt_backcolor_exp();
				break;

			case 5:
				for (i = 0 ; i < 0x300 >> 1 ; i++)	skip_object[i] = 1;
				front_layer = 1;
				object_no = 0;
				break;

			case 6:
				for (i = 0 ; i < 0x300 >> 1 ; i++)	skip_object[i] = 0;
				front_layer = 0;
				alt_backcolor = 0;
				object_no = 0;
				alt_backcolor_exp();
				break;

			case 7:
				sel = 0;
				break;

			default:
				break;
		}
	}

	if (input_ui_pressed_repeat(IPT_UI_LEFT, 8))
	{
		switch (sel)
		{
			case 1:
				if (object_number == 0)	break;
				if (object_no > 0)	object_no--;
				break;

			case 2:
				if (object_number == 0)	break;
				if (skip_object[object_no] == 1)	skip_object[object_no] = 0;
				break;

			case 3:
				if (front_layer == 1)	front_layer = 0;
				break;

			case 4:
				if (alt_backcolor == 1)
				{
					alt_backcolor = 0;
					alt_backcolor_exp();
				}
				break;

			default:
				break;
		}
	}

	if (input_ui_pressed_repeat(IPT_UI_RIGHT, 8))
	{
		switch (sel)
		{
			case 1:
				if (object_number == 0)	break;
				if (object_no < object_number)	object_no++;
				break;

			case 2:
				if (object_number == 0)	break;
				if (skip_object[object_no] == 0)	skip_object[object_no] = 1;
				break;

			case 3:
				if (front_layer == 0)	front_layer = 1;
				break;

			case 4:
				if (alt_backcolor == 0)
				{
					alt_backcolor = 1;
					alt_backcolor_exp();
				}
				break;

			default:
				break;
		}
	}

	if (sel == 1 && object_number)
	{
		if (code_pressed_memory_repeat(KEYCODE_PGDN, 8))
		{
			object_no -= 10;
			if (object_no < 0)	object_no = 0;
		}

		if (code_pressed_memory_repeat(KEYCODE_PGUP, 8))
		{
			object_no += 10;
			if (object_no > object_number)	object_no = object_number;
		}
	}

	if (input_ui_pressed_repeat(IPT_UI_UP, 8))
	{
		sel--;
		if (sel == 0) sel = 7;
	}

	if (input_ui_pressed_repeat(IPT_UI_DOWN, 8))
	{
		sel++;
		if (sel == 8) sel = 1;
	}

	if (input_ui_pressed_repeat(IPT_UI_CANCEL, 8) || input_ui_pressed_repeat(IPT_UI_OPERATION_OBJ, 8))
	{
		sel = 0;
	}

	object_selected = sel;

	update_video_and_audio();

	if (sel == 0)	object_operation_end();

}

/**************************************************************************/

int handle_user_interface(struct osd_bitmap *bitmap)
{
	extern int usres;

#ifdef KAILLERA
	if (kPlay)
		KailleraChatUpdate(bitmap);

	if (KailleraChatIsActive())
	{
		/* This call is for the cheat, it must be called once a frame */
		if (options.cheat) DoCheat(bitmap);

		/* show popup message if any */
		if (messagecounter > 0)
		{
			displaymessage(bitmap, messagetext, FONT_NORMAL);

			if (--messagecounter == 0)
				schedule_full_refresh();
		}
	}
	else
	{
#endif
    /* if the user pressed F12, save the screen to a file */
	if (input_ui_pressed(IPT_UI_SNAPSHOT))
		osd_save_snapshot(bitmap);

	if (options.only_jukebox == 1)
	{
		options.only_jukebox++;
		jukebox = 1;
		jukebox_start();
	}
	else if (options.only_jukebox == 0)
	{
#ifdef KAILLERA
		if (!kPlay)
#endif
		if (setup_selected == 0 && input_ui_pressed_repeat(IPT_UI_SOUND_TEST, 8) && object == 0)
		{
			jukebox ^= 1;
			if (jukebox)
				jukebox_start();
			else
			{
				if (wave_recording)
				{
					neojuke.que[0]  = neojuke.start;
					neojuke.que[1]  = neojuke.stop2;
					neojuke.que[2]  = neojuke.stop1;
					neojuke.counter = 3;
					wave_record_end();
					wave_recording = 0;
					neojuke.status = JUKE_IDLE;
				}
				else
				{
					jukebox_stop();
				}
			}
		}
	}

	if (jukebox) 
	{
		if (options.only_jukebox == 0)
		{
			if (input_ui_pressed(IPT_UI_CANCEL))
			{
				if (wave_recording)
				{
					neojuke.que[0]  = neojuke.start;
					neojuke.que[1]  = neojuke.stop2;
					neojuke.que[2]  = neojuke.stop1;
					neojuke.counter = 3;
					wave_record_end();
					wave_recording = 0;
					neojuke.status = JUKE_IDLE;
				}
				else
				{
					jukebox_stop();
					jukebox = 0;
				}
			}
		}
		if (input_ui_pressed(IPT_UI_CONFIGURE) || input_ui_pressed(IPT_UI_OPERATION_OBJ))
		{
		}
	}

	if (neojuke.counter)
	{
		neojuke.counter--;
		neojuke_send_message(neojuke.que[neojuke.counter]);

		if (neojuke.counter == 0)
		{
			int plist_tmp, plistbak_tmp;
			fadeout_counter = 0;

		 	switch (neojuke.status)
		 	{
		 		case JUKE_RECORDWAIT:
					neojuke.status = JUKE_RECORDING;
					wave_counter = 0;
					wave_recording = 1;
					wave_record_start();
					break;

				case JUKE_STOPWAIT:
					plist_tmp = neojuke.playlist;
					plistbak_tmp = neojuke.playlist_bak;
					memset(&neojuke, 0, sizeof(struct jukebox_t));
					neojuke.playlist = plist_tmp;
					neojuke.playlist_bak = plistbak_tmp;
					neojuke.status = JUKE_STOP;
					restore_contexts(OSD_FILETYPE_STATE, TEMPORARY_FILE);
					cpu_set_halt_line(0, 0);
					break;

				case JUKE_STARTWAIT:
					neojuke.status = JUKE_IDLE;
					break;
			}
		}
	}

	if (neojuke.status != JUKE_STOP)
	{
		if (neojuke.playlist)
		{
			jukebox_playlist(bitmap);
		}
		else
		{
			jukebox_playcommand(bitmap);
		}
	}
	else
	{
#ifdef KAILLERA
		if (!kPlay)
#endif
		if (input_ui_pressed_repeat(IPT_UI_WAVE_RECORD, 8) && neojuke.status == JUKE_STOP && object == 0)
		{
			wave_recording ^= 1;

			if (wave_recording != 0)
				wave_record_start();
			else
				wave_record_end();

			wave_counter = 0;
		}
	}

#ifdef KAILLERA
	if (!kPlay)
#endif
	if (setup_selected == 0 && input_ui_pressed_repeat(IPT_UI_OPERATION_OBJ, 8) && neojuke.status == JUKE_STOP)
	{
		object ^= 1;
		if (object)
		{
			object_operation_start();
		}
		else
		{
			object_operation_end();
		}
	}

	if (object)
	{
		if (input_ui_pressed(IPT_UI_CONFIGURE) || input_ui_pressed(IPT_UI_SOUND_TEST))
		{
		}

		if (input_ui_pressed(IPT_UI_CANCEL))
		{
			object_operation_end();
		}
		else
		{
			object_operation(bitmap);
		}
	}
	
	if (wave_recording && (GetAviCapture() == 0))
	{
		char buf[20];
		int width = strwidth(ui_getstring(UI_waverecording), FONT_PROPORTIONAL);

        plot_box(bitmap, Machine->uiwidth - width + 8, Machine->uiheight + 4, width, Machine->uifontheight, Machine->uifont->colortable[0], 1);

		if (++wave_counter == 60)
		{
			wave_counter = 0;
			wave_seconds++;

			if (wave_seconds == 60)
			{
				wave_seconds = 0;
				wave_minutes++;
			}
		}

		if (wave_counter < 40)
		{
			ui_text(bitmap, ui_getstring(UI_waverecording), Machine->uiwidth - width, Machine->uiheight - Machine->uifontheight);
		}

		sprintf(buf, "%02d:%02d", wave_minutes, wave_seconds);
		width = strwidth(buf, FONT_PROPORTIONAL);
		ui_text(bitmap, buf, Machine->uiwidth - width, Machine->uiheight - Machine->uifontheight * 2);
		schedule_full_refresh();
	}

	if (options.only_jukebox)
	{
		if (setup_selected == 0 && input_ui_pressed(IPT_UI_CANCEL) || IsCloseButtonPressed())
		{
			flash_all_values_and_buffers();
			return 1;
		}
	}
	else if (neojuke.status == JUKE_STOP && object == 0)
	{
		if (input_ui_pressed_repeat(IPT_UI_COMMAND, 8))
		{
			command_sc ^= 1;
		}

		if (command_sc && setup_selected == 0)
		{
			static int res;
			res = displaycommand_ex(bitmap, command_sel, 1);
			command_sel = res++;

			if (command_sel == 0)
			{
				command_sc = 0;
				command_sel = 1;
			}
		}

		if (options.cheat && input_ui_pressed_repeat(IPT_UI_CHEAT, 8))
		{
			cheat_sc ^= 1;
		}

		if (cheat_sc && setup_selected == 0)
		{
			static int res;
			res = cheat_menu(bitmap, cheat_sel, 1);
			cheat_sel = res++;

			if (cheat_sel == 0)
			{
				cheat_sc = 0;
				cheat_sel = 1;
			}
		}

		if (input_ui_pressed_repeat(IPT_UI_TIME, 8))
		{
			show_time ^= 1;
		}

		if (show_time && wave_recording == 0)
		{
			char buf[20];
			int width;
			time_t ltime;
			struct tm *today;

			time(&ltime);
			today = localtime(&ltime);

			sprintf(buf, "%02d:%02d", today->tm_hour, today->tm_min);
			width = strwidth(buf, FONT_PROPORTIONAL);
			ui_text(bitmap, buf, Machine->uiwidth - width, Machine->uiheight - Machine->uifontheight);
		}
		
#ifdef KAILLERA
		if (!kPlay)
#endif
		if (input_ui_pressed_repeat(IPT_UI_SAVE_STATE, 8))
			save_load_state(bitmap, 1);

#ifdef KAILLERA
		if (!kPlay)
#endif
		if (input_ui_pressed_repeat(IPT_UI_LOAD_STATE, 8))
			save_load_state(bitmap, 2);

#ifdef KAILLERA
		if (!kPlay)
#endif
		{
			if (input_ui_pressed_repeat(IPT_UI_SAVE_COMBO, 8))
			{
				record_combo ^= 1;

				if (!record_combo)
				{
					save_load_combo(bitmap, 1);
				}
				else if (combo_record_start())
				{
					usrintf_showmessage(ui_getstring(UI_startrecordcombo));
				}
			}
			if (input_ui_pressed_repeat(IPT_UI_LOAD_COMBO, 8))
				save_load_combo(bitmap, 2);
		}

		/* This call is for the cheat, it must be called once a frame */
		if (options.cheat) DoCheat(bitmap);

		/* if the user pressed ESC, stop the emulation */
		/* but don't quit if the setup menu is on screen */
		if (setup_selected == 0 && input_ui_pressed(IPT_UI_CANCEL) || IsCloseButtonPressed())
		{
			flash_all_values_and_buffers();
			return 1;
		}

		if (setup_selected == 0 && input_ui_pressed(IPT_UI_CONFIGURE))
		{
			setup_selected = -1;
			if (osd_selected != 0)
			{
				osd_selected = 0;	/* disable on screen display */
				schedule_full_refresh();
			}
		}
		if (setup_selected != 0) setup_selected = setup_menu(bitmap, setup_selected);

#ifdef KAILLERA
		if (!kPlay)
#endif
		if (osd_selected == 0 && input_ui_pressed(IPT_UI_ON_SCREEN_DISPLAY))
		{
			osd_selected = -1;
			if (setup_selected != 0)
			{
				setup_selected = 0; /* disable setup menu */
				schedule_full_refresh();
			}
		}
		if (osd_selected != 0) osd_selected = on_screen_display(bitmap, osd_selected);


		/* if the user pressed F3, reset the emulation */
#ifdef KAILLERA
		if (!kPlay)
#endif
		if (input_ui_pressed(IPT_UI_RESET_MACHINE))
		{
			if (wave_recording)
				wave_record_end();
			machine_reset();
		}


#ifdef KAILLERA
		if (!kPlay)
#endif
		if (single_step || input_ui_pressed(IPT_UI_PAUSE)) /* pause the game */
		{
/*			osd_selected = 0;	   disable on screen display, since we are going   */
							/* to change parameters affected by it */

			neogeo_save_screen();
			paused = 1;
			single_step = 0;

			if (single_step == 0)
			{
				osd_sound_enable(0);
				osd_pause(1);
			}

			while (!input_ui_pressed(IPT_UI_PAUSE))
			{
				profiler_mark(PROFILER_VIDEO);
				if (osd_skip_this_frame() == 0)
				{
					/* keep calling vh_screenrefresh() while paused so we can stuff */
					/* debug code in there */
					draw_screen();
					neogeo_restore_screen();
				}
				profiler_mark(PROFILER_END);

				if (input_ui_pressed(IPT_UI_SNAPSHOT))
					osd_save_snapshot(bitmap);

				if (input_ui_pressed_repeat(IPT_UI_SINGLE_STEP, 8))
				{
					single_step = 1;
					break;
				}

				if (setup_selected == 0 && input_ui_pressed_repeat(IPT_UI_OPERATION_OBJ,8))
				{
					object ^= 1;
					if (object)
					{
						object_operation_start();
					}
					else
					{
						object_operation_end();
					}
				}

				if (object)
				{
					if (input_ui_pressed(IPT_UI_CONFIGURE))
					{
					}

					if (input_ui_pressed(IPT_UI_CANCEL))
					{
						object_operation_end();
					}
					else
					{
						object_operation(bitmap);
					}
				}

				if ((setup_selected == 0 && input_ui_pressed(IPT_UI_CANCEL)) || IsCloseButtonPressed())
				{
					flash_all_values_and_buffers();
					return 1;
				}

				if (setup_selected == 0 && input_ui_pressed(IPT_UI_CONFIGURE))
				{
					setup_selected = -1;
					if (osd_selected != 0)
					{
						osd_selected = 0;	/* disable on screen display */
						schedule_full_refresh();
					}
				}
				if (setup_selected != 0) setup_selected = setup_menu(bitmap, setup_selected);

				if (osd_selected == 0 && input_ui_pressed(IPT_UI_ON_SCREEN_DISPLAY))
				{
					osd_selected = -1;
					if (setup_selected != 0)
					{
						setup_selected = 0; /* disable setup menu */
						schedule_full_refresh();
					}
				}
				if (osd_selected != 0) osd_selected = on_screen_display(bitmap, osd_selected);

				if (options.cheat) DisplayWatches(bitmap);

				/* show popup message if any */
				if (messagecounter > 0) displaymessage(bitmap, messagetext, 1);

				update_video_and_audio();
			}

			neogeo_restore_screen();

			if (object)
			{
				cpu_set_halt_line(0, 1);
				object_cpu = 1;
			}
			else
			{
				osd_sound_enable(1);
			}

			osd_pause(0);

			paused = 0;
		}


		if (input_ui_pressed(IPT_UI_SHOW_INPUT_LOG))
		{
			show_input ^= 1;
			memset(command_buf, 0, COMMAND_LOG_BUFSIZE);
		}
		if (show_input && command_counter)
		{
			ui_text(bitmap,command_buf, 0, 213);
			if (--command_counter == 0)
				memset(command_buf, 0, COMMAND_LOG_BUFSIZE);
		}

		/* show popup message if any */
		if (messagecounter > 0)
		{
			displaymessage(bitmap, messagetext, 1);

			if (--messagecounter == 0)
				schedule_full_refresh();
		}


#ifdef MAME_DEBUG
		if (input_ui_pressed(IPT_UI_SHOW_PROFILER))
		{
			show_profiler ^= 1;
			if (show_profiler)
				profiler_start();
			else
			{
				profiler_stop();
				schedule_full_refresh();
			}
		}

		if (show_profiler) profiler_show(bitmap);
#endif

		/* if the user pressed F4, show the character set */
#ifdef KAILLERA
		if (!kPlay)
#endif
		if (input_ui_pressed(IPT_UI_SHOW_GFX))
		{
			osd_sound_enable(0);
			showcharset(bitmap);
			osd_sound_enable(1);
		}
	}

#ifdef KAILLERA
	}
#endif
	if (usres)
		flash_all_values_and_buffers();

	return usres;
}


void init_user_interface(void)
{
	extern int snapno;   /* in common.c */

	snapno = 0;

	setup_menu_init();
	onscrd_init();
	flash_all_values_and_buffers();

	lhw = strwidth(ui_getstring(UI_lefthilight), FONT_PROPORTIONAL);
	rhw = strwidth(ui_getstring(UI_righthilight), FONT_PROPORTIONAL);
	uaw = strwidth(ui_getstring(UI_uparrow), FONT_PROPORTIONAL);
	daw = strwidth(ui_getstring(UI_downarrow), FONT_PROPORTIONAL);
	law = strwidth(ui_getstring(UI_leftarrow), FONT_PROPORTIONAL);
	raw = strwidth(ui_getstring(UI_rightarrow), FONT_PROPORTIONAL);
}
