#include "driver.h"

#ifndef WAVE_FORMAT_PCM
#define WAVE_FORMAT_PCM  1
#endif


static struct RiffWaveHeader
{
	INT8   riff[4];         // 'R''I''F''F'
	INT32  cksize;          // `ÑTCY = t@CTCY - 8
	INT8   wave[4];         // 'W''A''V''E'
	INT8   fmt[4];          // 'f''m''t'' '
	INT32  fmtsize;         // wFormatTag ` nBitsPerSamplẽTCY = 16 bytes
	INT16  wFormatTag;      // WAVEtH[}bg(jAPCM(0x0001)̂)
	INT16  nChannels;       // `l(mono:1 stereo:2)
	INT32  nSamplesPerSec;  // TvO[g
	INT32  nBytesPerSec;    // oCg/b       = nSmaplesPerSec * nChannels * (nBitsPerSample / 8)
	INT16  nBytesPerSample; // oCg/Tv = nChannels * (nBitsPerSample / 8)
	UINT16 nBitsPerSample;  // rbg[g (16bit̂)
	INT8   data[4];         // 'd''a''t''a'
	INT32  datasize;        // f[^TCY = t@CTCY - wb_TCY(44 bytes)
} wh;

static void *wavfile = NULL;
static int wavno = 0;
static int is_stereo;
UINT16 current_level[2];


int wave_record_start(void)
{
	char name[256];
	extern int wave_minutes, wave_seconds;

	wavno = 0;
	wave_minutes = 0;
	wave_seconds = 0;
	/* avoid overwriting existing files */
	/* first of all try with "gamename.wav" */
	sprintf(name,"%s", Machine->gamedrv->name);
	if (osd_faccess(name,OSD_FILETYPE_WAVE))
	{
		do
		{
			/* otherwise use "name_NN.wav" */
			sprintf(name,"%s_%02d", Machine->gamedrv->name, wavno++);
		} while (osd_faccess(name, OSD_FILETYPE_WAVE));
	}

	if ((wavfile = osd_fopen(0, name, OSD_FILETYPE_WAVE, 1)) == NULL)
		return 0;

	memcpy(wh.riff, "RIFF", 4);
	memcpy(wh.wave, "WAVE", 4);
	memcpy(wh.fmt,  "fmt ", 4);
	memcpy(wh.data, "data", 4);

	wh.cksize          = 0;
	wh.fmtsize         = 16;
	wh.wFormatTag      = WAVE_FORMAT_PCM;
	wh.nChannels       = is_stereo;
	wh.nSamplesPerSec  = Machine->sample_rate;
	wh.nBytesPerSec    = wh.nSamplesPerSec * wh.nChannels * 2;
	wh.nBytesPerSample = 2 * wh.nChannels;
	wh.nBitsPerSample  = 16;
	wh.datasize        = 0;

	osd_fwrite(wavfile, &wh, sizeof(struct RiffWaveHeader));

	return 1;
}


void wave_record_end(void)
{
	long len;

	len = osd_ftell(wavfile);

	wh.cksize   = len - 8;
	wh.datasize = len - sizeof(struct RiffWaveHeader);

	osd_fseek(wavfile, 0, SEEK_SET);
	osd_fwrite(wavfile, &wh, sizeof(struct RiffWaveHeader));

	osd_fclose(wavfile);
	wavfile = NULL;
}


void wave_write(char *buf, int len)
{
	osd_fwrite(wavfile, buf, len);
}


void wave_set_stereo(int stereo)
{
	is_stereo = stereo + 1;
}

void wave_set_current_level(INT16 *buf, int len)
{
    int i;

    if (is_stereo == 2)
    {
        current_level[0] = buf[0];
        current_level[1] = buf[1];
        for (i = 1; i < len; i++)
        {
            current_level[0] += abs(buf[i * 2 + 0]);
            current_level[0] >>= 1;
            current_level[1] += abs(buf[i * 2 + 1]);
            current_level[1] >>= 1;
        }
    }
    else
    {
        current_level[0] = buf[0];
        for (i = 1; i < len; i++)
        {
            current_level[0] += abs(buf[i]);
            current_level[0] >>= 1;
        }
        current_level[1] = current_level[0];
    }
}
